#include "pproj.h"

/* =========================================================================
 * ================================= PPROJ =================================
 * =========================================================================
       ________________________________________________________________
      |              Project a point onto a polyhedron                 |
      |                                                                |
      |   min ||x - y|| subject to  lo <= x <= hi, bl <= Ax <= bu      |
      |                                                                |
      |   where || . || is the Euclidean norm and A is a matrix        |
      |                                                                |
      |                  Version 1.0  (January 9, 2015)                |
      |                                                                |
      |                 Copyright by William W. Hager                  |
      |                                                                |
      |        http://clas.ufl.edu/users/hager/papers/Software/        |
      |                                                                |
      |  Disclaimer: The views expressed are those of the authors and  |
      |              do not reflect the official policy or position of |
      |              the Department of Defense or the U.S. Government. |
      |                                                                |
      |      Approved for Public Release, Distribution Unlimited       |
      |________________________________________________________________|

      Reference:
         W. W. Hager and H. Zhang, Sparse Techniques for Polyhedral Projection

    If the constraints are rewritten as

              lo <= x <= hi,  bl <= b <= bu,  Ax = b

    then the solution algorithm is based on maximizing the
    regularized dual function

               L_R(lambda) = inf {.5||x - y||^2 + lambda'(b - Ax) :
                                lo <= x <= hi,  bl <= b <= bu}
                               -.5*sigma||lambda - shift_l||^2

    If the starting guess for lambda is given, then we take shift_l = lambda.
    Since b appears linearly in the minimization, the dual function can
    have a discontinuous derivative. The stopping condition is based on
    either the absolute gradient (Parm->stopabs = TRUE) or the relative
    gradient (Parm->stopabs = FALSE). The absolute stopping criterion is

        ||grad L (lambda)||_sup <= grad_tol,

    where the norm is the sup norm (maximum absolute component), while
    the relative stopping criterion is

        ||grad L (lambda)||_sup / absAx_sup <= grad_tol,

    where absAx_sup = max_i sum_j |a_{ij} * x_j| and x achieves minimum in the
    dual function.  Any scaling of the constraints should be done
    before calling pproj.
   ========================================================================== */

int pproj /* return status:
                 (0) PPROJ_SOLUTION_FOUND
                 (1) PPROJ_TOLERANCE_NOT_MET
                 (2) PPROJ_INVALID_BOUNDS
                 (3) PPROJ_OUT_OF_MEMORY
                 (4) PPROJ_SSOR_NONASCENT */
(
    double       *x, /* solution (size ncol, output) */
    double  *lambda, /* multiplier (size nrow, output), if Uparm->use_lambda
                        = TRUE => starting guess on input */
    void     *userI, /* Parm->freememory = F => return PPcom structure
                        Parm->providePPcom = T => userI = PPcom structure*/
    double grad_tol, /* relative error tolerance for dual function grad */
    PPparm   *UParm, /* NULL => use default parameters */
    PPstat   *UStat, /* NULL => do not return statistics */
    double       *y, /* project y onto polyhedron (size ncol) */
    INT        nrow, /* number of rows in A */
    INT        ncol, /* number of cols in A */
    INT         *Ap, /* size ncol+1, column pointers */
    INT         *Ai, /* size Ap [ncol], row indices for A in increasing
                        order in each column */
    double      *Ax, /* size Ap [ncol], numerical entries of A */
    double      *lo, /* size n, lower bounds for x */
    double      *hi, /* size n, upper bounds for x */
    double      *Bl, /* size n, lower bounds for b denoted bl above */
    double      *Bu  /* size n, upper bounds for b denoted bu above */
)
{
    int it, it_sub, status, PrintLevel ;
    INT i, j, ni, *ineq_row ;
    double bli, bui, tic, *b, *bl, *bu ;
    PPprob *Prob ;
    PPparm *Parm ;
    PPstat *Stat ;
    PPwork    *W ;
    PPcom     *I ;
    cholmod_common *cmm ;
    /* ====================================================================== */
    /* === initializations ================================================== */
    /* ====================================================================== */

    /* initialize the statistics structure */
    if ( UStat == NULL)
    {
        Stat = (PPstat *) malloc (sizeof (PPstat)) ;
    }
    else
    {
        Stat = UStat ;
    }

    /* initialize the parameters */
    if ( UParm == NULL )
    {
        Parm = (PPparm *) malloc (sizeof (PPparm)) ;
        pproj_default (Parm) ;
    }
    else
    {
        Parm = UParm ;
    }
    PrintLevel = Parm->PrintLevel ;

    if ( Parm->providePPcom ) /* extract work, problem and stat from userI */
    {
        I = userI ;
        W = I->Work ;
        W->UParm = UParm ;
        W->UStat = UStat ;
        Prob = I->Prob ;
        Stat = I->Stat ;
        I->Parm = Parm ;
        cmm = W->cmm ;
        pproj_copyx (Prob->lo, lo, ncol) ;
        pproj_copyx (Prob->hi, hi, ncol) ;
    }
    else                   /* allocate I */
    {
        I = (PPcom *) malloc (sizeof (PPcom)) ;
        I->Parm = Parm ;
        I->Stat = Stat ;
        Prob = (PPprob *) malloc (sizeof (PPprob)) ;
        I->Prob = Prob ;
        W = (PPwork *) malloc (sizeof (PPwork)) ;
        I->Work = W ;
        W->UParm = UParm ;
        W->UStat = UStat ;

        I->Check = (PPcheck *) malloc (sizeof (PPcheck)) ;
        if ( Parm->cholmod )
        {
            /* cholmod common initialization */
            cmm = (cholmod_common *) malloc (sizeof (cholmod_common)) ;
            W->cmm = cmm ;
            cholmod_start (cmm) ;
            cmm->error_handler = pproj_error ;
            cmm->final_asis = FALSE ;
            cmm->final_super = FALSE ;
            cmm->final_ll = FALSE ;
            cmm->final_pack = FALSE ;
            cmm->final_resymbol = TRUE ;
            cmm->final_monotonic = FALSE ;
            cmm->print = 0 ;
            /* previous 1.2, 1.2 10 */
            cmm->grow0 = 1.1 ; /* factor grows by this much (multiplicative) */
            cmm->grow1 = 1.2 ; /* column grows by this much (multiplicative) */
            cmm->grow2 = 100 ; /* pad this much space in a col (additive, int)*/

            /* col grows to grow1*need + grow2, if reallocated */
            /* factor grows to grow0*need

            defaults:
                grow0: 1.2 (used 2 in LPDASA)
                grow1: 1.2 (used 2 in LPDASA)
                grow2: 5   (used 5 in LPDASA) */

            cmm->nrealloc_col = 0 ;
            cmm->nrealloc_factor = 0 ;
            cmm->dbound = Parm->sigma ;
            /* initial problem allocations connected with cholmod */
            Prob->rowperm = (INT *) malloc (nrow*sizeof (INT)) ;
            Prob->colperm = (INT *) malloc (ncol*sizeof (INT)) ;
            Prob->Anz = (INT *) malloc (ncol*sizeof (INT)) ;
            /* will copy A so that user's A is not modified by permutations */
            Prob->Ap = (INT *) malloc ((ncol+1)*sizeof (INT)) ;
            pproj_copyI (Prob->Ap, Ap, ncol+1) ;
            Prob->Ai = (INT *) malloc (Ap [ncol]*sizeof (INT)) ;
            pproj_copyI (Prob->Ai, Ai, Ap [ncol]) ;
            Prob->Ax = (double *) malloc (Ap [ncol]*sizeof (double)) ;
            pproj_copyx (Prob->Ax, Ax, Ap [ncol]) ;
            Prob->y = (double *) malloc (ncol*sizeof (double)) ;
            pproj_copyx (Prob->y, y, ncol) ;
            Prob->lo = (double *) malloc (ncol*sizeof (double)) ;
            pproj_copyx (Prob->lo, lo, ncol) ;
            Prob->hi = (double *) malloc (ncol*sizeof (double)) ;
            pproj_copyx (Prob->hi, hi, ncol) ;
        }
        /* For iterative methods, there are no permutations, so no need to
           save a copy of the user's input arrays */
        else
        {
            W->cmm = NULL ;
            Prob->rowperm = NULL ;
            Prob->colperm = NULL ;
            Prob->Anz = NULL ;
            Prob->Ap = Ap ;
            Prob->Ai = Ai ;
            Prob->Ax = Ax ;
            Prob->y = y ;
            Prob->lo = lo ;
            Prob->hi = hi ;
        }
        Prob->nrow = nrow ;
        Prob->ncol = ncol ;
        Prob->ineq_row = NULL ;
        Prob->bl = NULL ;
        Prob->bu = NULL ;
        Prob->maxcol = -1 ; /* maximum number of nonzeros in a column */

        /* count the number of strict inequalities */
        ni = 0 ;
        for (i = 0; i < nrow; i++)
        {
            if ( Bl [i] < Bu [i] )
            {
                ni++ ;
            }
        }
        Prob->ni = ni ; /* number of inequalities that are strict */

        /* Separate the components of Bl that are strictly less than Bu to
           populate the arrays bl and bu of length ni.  The row numbers
           associated with the strict inequalities are stored in the array
           ineq_row where the first entry is in location 1 not 0. */
        ineq_row = (INT *) malloc ((ni+2)*sizeof (INT)) ;
        Prob->ineq_row = ineq_row ;
        bl = (double *) malloc ((ni+1)*sizeof (double)) ;
        Prob->bl = bl ;
        bu = (double *) malloc ((ni+1)*sizeof (double)) ;
        Prob->bu = bu ;

        pproj_copyx (Prob->lo, lo, ncol) ;
        pproj_copyx (Prob->hi, hi, ncol) ;

        /* The components of b are defined as follows:

               b_i = Bl_i if Bl_i = Bu_i
               b_i = 0  otherwise (a place holder) */

        b = (double *) malloc (nrow*sizeof (double)) ;
        Prob->b = b ;
        ni = 0 ;
        for (i = 0; i < nrow; i++)
        {
            bli = Bl [i] ;
            bui = Bu [i] ;
            if ( bli < bui )
            {
                b  [i] = ZERO ;
                ni++ ;
                bl [ni] = bli ;
                bu [ni] = bui ;
                ineq_row [ni] = i ;
            }
            else if ( bli == bui )
            {
                b [i] = bli ;
            }
            else
            {
                return (PPROJ_INVALID_BOUNDS) ; /* bl [i] > bu [i] */
            }
        }
        ineq_row [ni+1] = nrow ; /* stopper */
    }

    /* initialize work structure */
    if ( Parm->use_lambda == TRUE )
    {
        W->shiftl_is_zero = FALSE ; /* lambda will be stored in shift_l */
    }
    else
    {
        W->shiftl_is_zero = TRUE ;  /* shift_l will be set to 0 */
    }
    W->lambda = lambda ;
    W->x = x ;

    /* also store grad_tol in problem structure */
    Prob->grad_tol = grad_tol ;

    /* initialize work arrays and statistics structure */
    tic = pproj_timer () ;
    status = pproj_initwork (I) ;
    if ( status == PPROJ_OUT_OF_MEMORY ) return (status) ;

    /* copy y into work structure if memory was provided */
    if ( Parm->providePPcom )
    {
        if ( Parm->cholmod )
        for (j = 0; j < ncol; j++)
        {
            Prob->y [j] = y [Prob->colperm [j]] ;
        }
        else
        {
            pproj_copyx (Prob->y, y, ncol) ;
        }
        Stat->initialize += pproj_timer () - tic ;
    }
    /* initialize AT and possibly perform a multilevel partitioning */
    status = pproj_initlevels (I) ;
    Stat->initialize += pproj_timer () - tic ;
    if ( status == PPROJ_OUT_OF_MEMORY ) return (status) ;

#ifndef NDEBUG
    /* check that ineq_row corresponds to the strict inequalities in the
       original matrix */
    {
        INT k ;
        if ( Parm->cholmod == TRUE )
        {
            for (k = 1; k <= ni; k++)
            {
                i = Prob->rowperm [ineq_row [k]] ;
                if ( Bl [i] != Prob->bl [k] )
                {
                    pproj_error (-1, __FILE__, __LINE__,
                    "bl not permuted according to rowperm\n") ;
                }
                if ( Bu [i] != Prob->bu [k] )
                {
                    pproj_error (-1, __FILE__, __LINE__,
                    "bu not permuted according to rowperm\n") ;
                }
            }
        }
    }
#endif

    tic = pproj_timer () ;
    status = pproj_phase1 (I) ;
    Stat->phase1 += pproj_timer () - tic ;
    if ( status == PPROJ_SOLUTION_FOUND )
    {
        pproj_invert (I) ; /* invert row and column permutations */
        if ( Parm->freememory )
        {
            pproj_free (&I) ;
        }
        else /* return PPcom structure to user */
        {
            userI = I ;
        }
        return (status) ;
    }

    /* status = call_BAG (lambda, Prob) ; call Hongchao's code */

    /* Using the starting guess obtained in phase1, we now compute
       the projection of y onto the polyhedron. To facilitate the
       evaluation of the dual function, we make a change of variables
       x = x0 + z where x0 is a fixed point chosen to simplify the
       evaluation of the dual function.  After this translation, the
       dual function becomes

       (T)     L_R(lambda) = inf {.5||z + x0 - y||^2 + lambda'(b0 - Az) :
                                lo0 <= z <= hi0,  bl <= b <= bu}
                               -.5*sigma||lambda - shift_l||^2

       where b0 = b - A*x0, lo0 = lo - x0, and hi0 = hi - x0.

       The unconstrained minimizer x of the original Lagrangian is
       x(lambda) = y + A'lambda. x0 is chosen as follows:

           if x(lambda)_j < lo_j, then x0_j = lo_j,
           if x(lambda)_j > hi_j, then x0_j = hi_j,
           otherwise                   x0_j = (y + A'lambda)_j (j in F)

       After making this translation, the constrained minimizer of the
       Lagrangian (T) is z(lambda) = 0. Throughout the rest of the code,
       the variable x is really x0. z does not appear explicitly since
       it is readily computed. It is given by

           z(lambda) = mid (lo0, x(lambda) - x0, hi0),

       where mid is the componentwise operator which is the median (or middle
       of the 3 components).

       We also perform a translation in lambda. We express lambda as the
       sum of three vectors dlambda + lambda + shift_l, where shift_l is
       the proximal shift variable.  Both shift_l and lambda are fixed
       during the iteration of dasa, while dlambda is the variable that
       is being optimized over. Initially, dlambda = 0.  b0 above corresponds
       to the b array of the code.  The array c in the code stores
       the expression (y + A'lambda) - x0 which is contained in z(lambda).
       Thus cF = 0 initially by its definition above. If x(lambda)_j < lo_j,
       then c_j < 0, and when c_j reaches 0, the j-th constraint
       z_j >= 0 changes to free. By defining b and c in this way,
       the implementation of the dual active set algorithm is simplified.
       In particular, when we minimize the dual function for a given
       choice of F and B, we need to solve the linear system

       (AF*AF'+ sigma I)dlambda = b - sigma lambda - AB*xB - AF*cF

       where c = y + A'*(lambda + shift_l). Note that lambda and shift_l
       are fixed and xB in the translated problem is 0. Thus as indices
       move from B to F, the only thing that changes on the right side
       is cF, which is often a sparse change to the right side. Also,
       the directional derivative of the dual function in the direction d is

       d'*[b - sigma*(lambda + dlambda) - AB*xB -AF*cF].

       Again, xB = 0 in the translated problem so we only need to
       evaluate the product AF*cF since everything else in [ ... ]
       is fixed.  */
    it = 0 ;
    while ( status == PPROJ_TOLERANCE_NOT_MET )
    {
        it_sub = 0 ;
        while ( status == PPROJ_TOLERANCE_NOT_MET )
        {
            /* apply iteration of DASA, only free variables and drop rows */
            tic = pproj_timer () ;
            status = pproj_dasa (I) ;
            Stat->dasa += pproj_timer () - tic ;

            if ( (status == PPROJ_SSOR_NONASCENT) ||
                 (status == PPROJ_SSOR_MAX_ITS) )
            {
                break ;
            }

            /* check the error and either add rows or bind variables */
            tic = pproj_timer () ;
            status = pproj_check_error (I) ;
            Stat->checkerr += pproj_timer () - tic ;

            it_sub++ ;
            if ( PrintLevel > 0 )
            {
                printf ("dasa subiteration %i status = %i\n", it_sub, status) ;
            }
            /* if it does not appear that the stopping criteria can be
               reached, then perform a proximal update */
            if ( it_sub > 100 )
            {
                status = PPROJ_PROX_UPDATE ;
            }
        }

        if ( PrintLevel > 0 )
        {
            printf ("proximal iteration %i status = %i\n", it, status) ;
        }

        if ( (status == PPROJ_SSOR_NONASCENT) ||
             (status == PPROJ_SSOR_MAX_ITS) )
        {
            break ;
        }

        /* perform a proximal update, recompute b and c, and evaluate the
           error using the new b and c */
        tic = pproj_timer () ;
        status = pproj_prox_update (I) ;
        Stat->prox_update += pproj_timer () - tic ;
        it++ ;
        /* If nprox proximal updates have been performed without reaching the
           error tolerance, then exit */
        if ( it == Parm->nprox )
        {
            break ;
        }
    }
    if ( Parm->cholmod && (UStat != NULL) )
    {
        cholmod_factor *L ;
        INT lnnz, *ir, *Lnz ;
        L = W->L ;
        if ( L != NULL )
        {
            Lnz = L->nz ;
            ir = W->ir ;
            lnnz = 0 ;
            for (i = 0; i < nrow; i++)
            {
                if ( ir [i] <= ni ) /* row is active */
                {
                    lnnz += Lnz [i] ;
                }
            }
        }
        else
        {
            lnnz = 0 ;
        }
        Stat->lnnz = lnnz ;
    }
    pproj_invert (I) ; /* invert row and column permutations */
    if ( Parm->freememory )
    {
        pproj_free (&I) ;
    }
    else /* return PPcom structure to user */
    {
        userI = I ;
    }
    return (status) ;
}

/* =========================================================================
   ============================== pproj_defaults ===========================
   ========================================================================= */
void pproj_default
(
    PPparm *Parm /* Parameter structure */
)
{
    /* T = use the input argument lambda as the starting guess
       F = starting guess is lambda = 0 */
    Parm->use_lambda = FALSE ;

    /* T = use cholmod to update and downdate the linear equation
       F = solve linear equations by iterative method */
    Parm->cholmod = TRUE ;

    /* T = perform a multilevel partition of the problem (used if cholmod = T)
           and solve using the multilevel version of the algorithm
       F = use problem as given */
    Parm->multilevel = TRUE ;

    /* T = use absolute stopping criterion ||grad L (lambda)||_sup <= grad_tol
       F = use relative criterion ||grad L (lambda)||_sup/absAx_sup <= grad_tol
       where absAx_sup = max_i sum_j |a_{ij}x_j| where x achieves min in
       dual function */
    Parm->stopabs = FALSE ;

    /* T => no need to allocate memory in init_work and init_levels since
            the PPcom input argument has all the associated memory and
            initializations
       F => run init_work and init_levels to set up the PPcom structure */
    Parm->providePPcom = FALSE ;

    /* T => free memory before exiting
       F = return memory and return the PPcom structure */
    Parm->freememory = TRUE ;

    /* A regularization is added to the dual problem. Each iteration
       involves solving an equation with matrix (A_F A_F' + sigma I)
       where sigma is the regularization parameter. Parm->sigma is the
       starting value for sigma. */
    Parm->sigma = ldexp (1, -44) ;

    /* T => scale sigma by max |a_{ij}| */
    Parm->ScaleSigma = TRUE ;

    /* In each proximal update, sigma decays by the factor sigma_decay */
    Parm->sigma_decay = ldexp (1, -4) ;

    /* Upper bound on the maximum number of proximal updates. Assuming
       sigma is small, like 2^(-44), the number of proximal update should
       be small. */
    Parm->nprox = 100 ;

    /* Factor by which the stepsize parameter alpha grows when the Armijo
       condition in SpaRSA is not satisfied */
    Parm->armijo_grow = 4.0 ;

    /* Maximum number of times we try to satisfy the Armijo condition
       in SpaRSA */
    Parm->narmijo = 10 ;

    /* number of objective function values stored for the SpaRSA line search */
    Parm->mem = 8 ;

    /* Maximum number of SpaRSA iterations */
    Parm->nsparsa = 1000 ;

    /* The SpaRSA line search is nonmonotone. mem is the number of prior
       function values to store for computing the reference value in the
       line search */

    /* Stopping parameter for SpaRSA. Stop when a component of the gradient
       associated with equality constraints or bound inequality constraints
       for which |grad_i L_R| >= gamma * max_k |grad_k L_R| */
    Parm->gamma = 0.1 ;

    /* tau, beta, grad_decay, and gamma_decay are parameters that enter into
       the the definition of the undecided index set. An index j is undecided
       if sign (lambda_j) grad_j L_R <= -tau max_{k active} |grad_k L_R|^beta
       The active rows are the row in RLinkUp and RLinkDn. They correspond
       to equality constraints and active inequality constraints. Undecided
       indices are only monitored when ||current gradient|| <= grad_decay *
       ||initial gradient (after phase1)|| */
    Parm->tau = 0.1 ;
    Parm->beta = 0.5 ;
    Parm->grad_decay = 1.e-4 ;
    Parm->gamma_decay = 0.5 ;
    

    /* T = use CHOLMOD statistics to determine whether to perform a coordinate
           ascent iteration
       F = do not use coordinate ascent at the start of dasa */
    Parm->use_coor_ascent = TRUE ;

    /* The cost of a coordinate ascent iteration is proportional to Annz
       while the cost of an update/downdate iteration is proportional to Lnnz.
       We perform coordinate ascent iteration when Annz*coorcost <= estimate
       for the number of flops for an update */
    Parm->coorcost = 1. ;

    /* T = use ssor0 iteration
       F = do not use ssor0
       If T and CHOLMOD is used, then CHOLMOD statistics is used to decide
       whether the ssor0 iteration is worthwhile */
    Parm->use_ssor0 = TRUE ;

    /* T = use ssor1 iteration
       F = do not use ssor1
       If T and CHOLMOD is used, then CHOLMOD statistics is used to decide
       whether the ssor1 iteration is worthwhile */
    Parm->use_ssor1 = TRUE ;

    /* T = use SpaRSA iteration if appropriate
       F = do not use SpaRSA */
    Parm->use_sparsa = TRUE ;

    /* T = use phase1
       F = skip phase1 */
    Parm->use_phase1 = TRUE ;

    /* Stop ssor1 when current errls <= ssor_decay * errdual */
    Parm->ssordecay = 1.e-3 ;

    /* In the ssor1 iteration, we store ssormem prior search directions
       and optimize over the space spanned by these prior search directions
       when the active constraints change. */
    Parm->ssormem = 8 ; /* must be > 1 if ssor1 is used */

    /* The cost of an ssor iteration is proportional to Annz while the cost
       of an update/downdate iteration is proportional to Lnnz. We perform
       ssor iterations when Annz*ssorcost <= estimate for the number of
       flops for an update */
    Parm->ssorcost = 8. ;

    /* Upper bound on the number of ssor iterations */
    Parm->ssormaxits = INT_MAX ;

    /* If a row has been dropped, we do not immediately rebind it, but
       allow some slack before rebinding. If other optimality violations
       are large enough, we keep the row dropped or the column free
       until there is some benefit in the error from adding the row or
       binding the column */
    Parm->cutfactor = .0001 ;

    /* In the ssor iteration, we continue to ascend the dual function
       freeing variables until the residual error (err1) for the linear
       equation is small relative to the error (err) connected with free
       variables violating the bound constraints. Since the equation is
       positive definite, err1 tends to 0 as variables are freed and the
       dual function is ascended */
    Parm->tolssor = ldexp (1, -8) ; /* stop when err1 <= err*tol1 */

    /* To regularize the dual function, we add the proximal term
       sigma/2 ||lambda - shift_l|| to obtain the regularized dual function
       L_R (lambda) above. A proximal update is performed, replacing shift_l
       by the current lambda, when
       ||grad L_R (lambda)||_1 <= tolprox*||grad L (lambda)||_1 */
    Parm->tolprox = 0.01 ;

    /* Refactor A when
       ||b - Ax - sigma*(lamba-shift_l)||_1 >= tolrefactor*||b||_1
       The expression on the left is the norm of the residual of the
       linear system that is solved in dasa. */
    Parm->tolrefactor = .125 ;

    /* Phase1 controls the initial number of coordinate ascent iterations.
       The number of iterations is max {nrow^phase1, 5}. To skip phase1,
       set phase1 < 0 */
    Parm->phase1 = .333 ;

    /* PrintLevel = 0 (noprint)
                  = 1, 2, 3 (increasing levels of printing) */
    Parm->PrintLevel = 0 ;

    /* debug = 0 (no debugging)
       debug = 1 (debugging outside of main loops)
       debug = 2 (debugging inside of main loops) */
    Parm->debug = 1 ;

    /* relative tolerance for reporting an error in the debugging routines */
    Parm->checktol = 1.e-7 ;
}

/* ========================================================================== */
/* === pproj_print_stat ===================================================== */
/* ========================================================================== */
/* Print the data stored in the PPstat structure
   NOTE: If Parm->free is TRUE, then Stat->updowns and Stat->solves
         are freed at the end of this routine, after printing their results.  */
/* ========================================================================== */
void pproj_print_stat
(
    PPstat   *Stat, /* pointer to statistics structure */
    int freememory  /* T => free Stat->updowns and Stat->solves */
)
{
    int i, k ;
    double sparsity, n ;
    printf ("\n----------- Solver PPstat -----------\n") ;
    printf ("No. blocks in multilevel partition of A . %i\n", Stat->blks) ;
    printf ("Depth of multilevel partition tree ...... %i\n", Stat->maxdepth) ;
    printf ("Phase 1 iterations ...................... %i\n", Stat->phase1_its);
    printf ("Coordinate ascent iterations ............ %i\n",
            Stat->coor_ascent_its) ;
    printf ("    variables freed in coordinate ascent  %i\n",
            Stat->coor_ascent_free) ;
    printf ("    rows dropped in coordinate ascent ... %i\n",
            Stat->coor_ascent_drop) ;
    printf ("Gradient ascent iterations .............. %i\n",
            Stat->ssor0_its) ;
    printf ("    variables freed in gradient ascent .. %i\n",
            Stat->ssor0_free) ;
    printf ("    rows dropped in gradient ascent ..... %i\n",
            Stat->ssor0_drop) ;
    printf ("Preconditioned CG iterations ............ %i\n",
            Stat->ssor1_its) ;
    printf ("    variables freed in CG ............... %i\n",
            Stat->ssor1_free) ;
    printf ("    rows dropped in CG .................. %i\n",
            Stat->ssor1_drop) ;
    printf ("SpaRSA iterations ....................... %i\n",
            Stat->sparsa_its) ;
    printf ("    change in column activity ........... %i\n",
            Stat->sparsa_col) ;
    printf ("    change in row activity .............. %i\n",
            Stat->sparsa_row) ;
    printf ("    failures of Armijo step ............. %i\n",
            Stat->sparsa_step_fail) ;
    printf ("Proximal updates ........................ %i\n",
            Stat->nprox) ;
    printf ("Cholesky factorizations ................. %i\n", Stat->nchols) ;
    printf ("    nonzeros in final factor ............ %i", Stat->lnnz) ;
    n = (double) Stat->nrow ;
    sparsity = 100*(1. - 2.*(double) Stat->lnnz/(n*n + n)) ;
    printf (" %4.1f%% sparse\n", sparsity) ;
    printf ("    rows dropped from L ................. %i\n", Stat->rowdn) ;
    printf ("    rows added to L ..................... %i\n", Stat->rowup) ;
    printf ("    rank 1 downdates to L ............... %i\n", Stat->coldn) ;
    printf ("    rank 1 updates to L ................. %i\n", Stat->colup) ;
    printf ("    Size breakdown of the updates ([size]: "
            "number of this size):\n") ;
    k = Stat->size_updowns ;
    for (i = 1; i < k; i++)
    {
        if (Stat->updowns [i] > 0)
        {
            printf ("        updowns [%3d]: %d\n", i, Stat->updowns [i]) ;
        }
    }
    if ( Stat->updowns [k] > 0 )
    {
        printf ("        updowns [>=%3d]: %d\n", k, Stat->updowns [k]) ;
    }
    if ( Stat->blks == 1 )
    {
        printf ("    No. of solves:   %i\n", Stat->solves [0]) ;
    }
    else
    {
        printf ("    No. of solves by depth in the multilevel partition "
                "tree:\n") ;
        printf ("        (deeper <=> further from root of tree <=> "
                "fewer flops)\n");
        for (i = 0; i <= Stat->maxdepth; i++)
        {
            printf ("        depth [%2d]: %d\n", i, Stat->solves [i]) ;
        }
    }
    if ( freememory )
    {
        free (Stat->updowns) ;
        free (Stat->solves) ;
    }
    printf ("\n-------- Time spent in routines ---------\n") ;
    printf ("Multilevel partition and reorder A ...... %e\n", Stat->partition) ;
    printf ("Initialization (includes partition) ..... %e\n", Stat->initialize);
    printf ("Phase 1 ................................. %e\n", Stat->phase1) ;
    printf ("Coordinate ascent ....................... %e\n",Stat->coor_ascent);
    printf ("SSOR0 ................................... %e\n", Stat->ssor0);
    printf ("SSOR1 ................................... %e\n", Stat->ssor1);
    printf ("SpaRSA .................................. %e\n", Stat->sparsa);
    printf ("DASA .................................... %e\n", Stat->dasa) ;
    printf ("DASA line search ........................ %e\n", Stat->dasa_line) ;
    printf ("Check error ............................. %e\n", Stat->checkerr) ;
    printf ("Proximal update ......................... %e\n",Stat->prox_update);
    printf ("Invert permutation ...................... %e\n", Stat->invert) ;
    printf ("Row modifications of Cholesky factor .... %e\n", Stat->modrow) ;
    printf ("Column modifications of Cholesky factor . %e\n", Stat->modcol) ;
    printf ("Cholesky factorization .................. %e\n", Stat->chol) ;
    printf ("Partial Cholesky factorization .......... %e\n", Stat->cholinc) ;
    printf ("Back solves ............................. %e\n", Stat->dltsolve) ;
    printf ("Forward solves .......................... %e\n", Stat->lsolve) ;
}

/* ========================================================================== */
/* === pproj_KKTerror.c ===================================================== */
/* ========================================================================== */

/*  Determine KKT error in a solution to the polyhedral projection problem

      min ||x - y||^2 subject to  lo <= x <= hi, bl <= b <= bu, Ax = b

    where || . || is the Euclidean norm and A is a matrix.  Input is a
   computed projection x and a Lagrange multiplier lambda associated with
   the constraint b - Ax = 0. The KKT conditions can be expressed as follows:

        0 in partial L (lambda),

    where partial stands for subdifferential and L is the dual function
    defined by

        L (lambda) = inf { L(lambda, x, b) : lo <= x <= hi, bl <= b <= bu }, and

        L (lambda, x, b) = 0.5||x-y||^2 + lambda'(b - Ax).

    For a given lambda, the minimizer over x is given by

        x_i (lambda) = mid {lo_i, c_i, hi_i}, where c = y + A'lambda.

    Here mid {u, v, w}_i is the median (or middle) of u_i, v_i, and w_i.
    The subdifferential can be expressed as follows:

        partial L (lambda)_i = b_i + [Ax (lambda)]_i, where
               -
              |    bl_i      if lambda_i > 0,
        b_i = |    bu_i      if lambda_i < 0,
              | [bl_i, bu_i] if lambda_i = 0.
               -
    The error estimate is the sup norm of the perturbation in b needed
    to satisfy the KKT conditions. The code also returns the 1-norm
    relative difference between x and x(lambda) as well as the sup
    norm of absAx where absAx_i = sum_j |a_{ij}*x_j (lambda)|.  */

double pproj_KKTerror /* returns the largest of the primal and dual errors */
(
    double    *errb, /* sup norm error in b (only returned when not NULL) */
    double    *errx, /* 1 norm relative difference of x and x(lamba)
                        (only returned when not NULL) */
    double   *absAx, /* sup norm of absAx (only returned when not NULL) */
    double       *x, /* given projection (input, size ncol) */
    double  *lambda, /* given multiplier (input, size nrow) */
    double       *y, /* y, the point project onto the polyhedron (size ncol) */
    INT        nrow, /* number of rows in A */
    INT        ncol, /* number of cols in A */
    int  PrintLevel, /* 0 = no printout, 1 = dual function value,
                        2 = amount of violation of each constraint on b_i */
    INT         *Ap, /* column pointers (size ncol+1) */
    INT         *Ai, /* row indices for A in increasing order for each
                        column (size Ap [ncol]) */
    double      *Ax, /* numerical entries of A (size Ap [ncol]) */
    double      *lo, /* lower bounds for x (size ncol) */
    double      *hi, /* upper bounds for x (size ncol) */
    double      *bl, /* lower bounds for b (size nrow) */
    double      *bu, /* upper bounds for b (size nrow) */
    double    *work  /* work array (size 2*nrow) */
)
{
    INT i, j, p, p0, q ;
    double xj, t, s, err, pertx, pertb, normx, dualval,
           *AtimesX, *absAtimesX, *workd ;

    dualval = ZERO ;
    pertx = ZERO ;
    normx = ZERO ;
    workd = work ;
    AtimesX    = workd ; workd += nrow ;
    absAtimesX = workd ; workd += nrow ;
    pproj_initx (AtimesX, ZERO, nrow) ;
    pproj_initx (absAtimesX, ZERO, nrow) ;
    p = 0 ;
    for (j = 0; j < ncol; j++)
    {
        xj = y [j] ;
        p0 = p ;
        q = Ap [j+1] ;
        for (; p < q; p++)
        {
            xj += lambda [Ai [p]]*Ax [p] ;
        }
        if ( xj < lo [j] )
        {
            xj = lo [j] ;
        }
        else if ( xj > hi [j] )
        {
            xj = hi [j] ;
        }
        pertx += fabs (xj - x [j]) ;
        normx += fabs (x [j]) ;
        dualval += .5*(xj - y [j])*(xj - y [j]) ;
        p = p0 ;
        for (; p < q; p++)
        {
            INT ai = Ai [p] ;
            s = xj*Ax [p] ;
            AtimesX [ai] += s ;
            absAtimesX [ai] += fabs (s) ;
        }
    }
    if ( PrintLevel > 0 )
    {
        for (i = 0; i < nrow; i++)
        {
            if ( PrintLevel > 1 )
            {
                t = ZERO ;
                if ( AtimesX [i] < bl [i] )
                {
                    t = bl [i] - AtimesX [i] ;
                }
                else if ( AtimesX [i] > bu [i] )
                {
                    t = AtimesX [i] - bu [i] ;
                }
                printf ("constraint i: %i violation: %e\n", i, t) ;
            }
            if ( PrintLevel == 1 )
            {
                if ( lambda [i] < ZERO )
                {
                    dualval += lambda [i]*(bu [i]-AtimesX [i]) ;
                }
                else if ( lambda [i] > ZERO )
                {
                    dualval += lambda [i]*(bl [i]-AtimesX [i]) ;
                }
            }
        }
        printf ("dual function value ..................... %e\n", dualval) ;
    }
    s = ZERO ;
    pertb = ZERO ;
    for (i = 0; i < nrow; i++)
    {
        if ( lambda [i] < ZERO )
        {
            if ( pertb < fabs (bu [i] - AtimesX [i]) )
            {
                pertb = fabs (bu [i] - AtimesX [i]) ;
            }
            if ( s < absAtimesX [i] )
            {
                s = absAtimesX [i] ;
            }
        }
        else if ( lambda [i] > ZERO )
        {
            if ( pertb < fabs (bl [i] - AtimesX [i]) )
            {
                pertb = fabs (bl [i] - AtimesX [i]) ;
            }
            if ( s < absAtimesX [i] )
            {
                s = absAtimesX [i] ;
            }
        }
        else
        {
            t = bl [i] - AtimesX [i] ;
            if ( t > ZERO )
            {
                if ( pertb < t )
                {
                    pertb = t ;
                }
                if ( s < absAtimesX [i] )
                {
                    s = absAtimesX [i] ;
                }
            }
            else
            {
                t = bu [i] - AtimesX [i] ;
                if ( t < ZERO )
                {
                    if ( pertb < -t )
                    {
                        pertb = -t ;
                    }
                    if ( s < absAtimesX [i] )
                    {
                        s = absAtimesX [i] ;
                    }
                }
            }
        }
    }
    if ( normx != ZERO ) pertx /= normx ;
    if ( errb != NULL ) *errb = pertb ;
    if ( errx != NULL ) *errx = pertx ;
    if ( absAx != NULL ) *absAx = s ;

    /* return the largest error */
    err = MAX (pertx, pertb) ;
    return (err) ;
}

/* ========================================================================== */
/* ====== pproj_error ======================================================= */
/* ========================================================================== */
/* when -g compiler option is used, prints line number of error */
void pproj_error
(
    int status,
    const char *file,
    int line,
    const char *message
)
{
    if (status < 0)
    {
        PRINTF ("file: %s line: %d status: %d %s\n", file, line, status, message) ;
#ifdef MATLAB_MEX_FILE
        mexErrMsgTxt (message) ;
#else
        ASSERT (0) ;
        abort ( ) ;
#endif
    }
}

/* ========================================================================= */
/* === pproj_free ========================================================== */
/* ========================================================================= */
/* Free memory allocated for PPcom structure */
/* ========================================================================= */
void pproj_free
(
    PPcom **Ihandle
)
{
    PPcom     *I ;
    PPprob *Prob ;
    PPparm *Parm ;
    PPwork    *W ;

    I = *Ihandle ;
    Prob = I->Prob ;
    Parm = I->Parm ;
    W = I->Work ;
    free (I->Check) ; /* pproj */
    if ( W->UStat == NULL ) /* free statistics structure */
    {
        free (I->Stat->updowns) ; /* initwork */
        free (I->Stat->solves) ;  /* initlevels */
        free (I->Stat) ;          /* pproj */
    }
    Parm = I->Parm ;

    /* allocations specific to update/downdates */
    if ( Parm->cholmod )
    {
        /* pproj */
        cholmod_free_factor (&(W->L), W->cmm) ;
        cholmod_free_work (W->cmm) ;
        free (W->cmm) ;
        free (Prob->rowperm) ;
        free (Prob->colperm) ;
        free (Prob->Anz) ;
        free (Prob->Ap) ;
        free (Prob->Ai) ;
        free (Prob->Ax) ;
        free (Prob->y) ;
        free (Prob->lo) ;
        free (Prob->hi) ;

        /* initwork */
        free (W->changeRHS) ;
        free (W->Cp) ;
        free (W->Cnz) ;
        free (W->A) ;

        /* initlevels */
        free (W->Kids) ;
        free (W->nkids) ;
        free (W->parent) ;
        free (W->Kp) ;
        free (W->depth) ;
        free (W->leftdesc) ;
        free (W->col_start) ;
        free (W->sol_start) ;
        free (W->row_start) ;
        free (W->col_to_blk) ;
        free (W->sol_to_blk) ;
        free (W->joblist) ;
        free (W->kidsleft) ;
        free (W->jobcols) ;
        free (W->jobrows) ;
        free (W->Rstart) ;
        free (W->Rend) ;
    }
    else
    {
#ifndef NDEBUG
        free (W->leftdesc) ;
        free (W->col_start) ;
        free (W->row_start) ;
#endif
    }

    /* allocations applying to both iterative methods and update/downdates */

    /* pproj */
    free (Prob->ineq_row) ;
    free (Prob->bl) ;
    free (Prob->bu) ;
    free (Prob->b) ;
    free (Prob) ;

    /* initwork */
    free (W->ATp) ;
    free (W->ATi) ;
    free (W->ATx) ;
    free (W->AFTp) ;
    free (W->AFTnz) ;
    free (W->AFTi) ;
    free (W->AFTx) ;
    free (W->lambda_tot) ;
    free (W->dlambda) ;
    free (W->ns) ;
    free (W->RowmodList) ;
    free (W->RowmodFlag) ;
    free (W->ColmodList) ;
    free (W->ColmodFlag) ;
    free (W->dropped) ;
    free (W->F) ;
    free (W->ib) ;
    free (W->shift_l) ;
    free (W->b) ;
    free (W->c) ;
    free (W->cold) ;
    free (W->lo) ;
    free (W->hi) ;
    free (W->D) ;
    free (W->RLinkUp) ;
    free (W->RLinkDn) ;
    free (W->SLinkUp) ;
    free (W->SLinkDn) ;
    free (W->ir) ;

    /* initlevels */
    free (W->lstart) ;
    free (W->ustart) ;
    free (W->arrayi) ;
    free (W->arrayd) ;

    /* pproj */
    if ( W->UParm == NULL ) /* free parameter structure */
    {
        free (I->Parm) ;
    }
    free (W) ;
    free (I) ;
}

/* ========================================================================= */
/* === pproj_check_error =================================================== */
/* ========================================================================= */
/* Check the error at the current dual iterate:

       errdual = sup-norm of dual function gradient excluding the proximal
                 term.  If errdual is sufficiently small, then we are done.
                 The error is relative to ||absAx||_sup where absAx is
                 defined by absAx_i = sum_j |a_{ij}*x_j (lambda)|.

       errls   = sup-norm of the residual of the linear system. With
                 exact arithmetic, this should be zero. That is, at the final
                 iteration, after free variables and dropping rows, we
                 solve a linear system.  errls is the sup-norm of
                 the residual for this linear system. This basically measures
                 the accumulated error in the update/downdate process.
                 If errls is too big, then the matrix will be refactored.
                 The variable lhs is used to accumulate the left side term
                 AF*AF'*dlambda, which works out to be AF*cF. Both lhs and
                 sigma*dlambda are subtracted from b to obtain the residual
                 vector. The error is relative to ||absAx||_sup.

       errprox = sup-norm of the dual function gradient including the proximal
                 term.  If errprox is sufficiently small, then we perform a
                 proximal update.

       colerr  = Max violation of bounds (if a violation is small enough
                 relative to the max violation, we keep the variable free)

       rowerr  = Max violation of inequality constraints (if a violation
                 for a row is small enough relative to the max violation,
                 the row remains dropped) */
/* ========================================================================= */
int pproj_check_error /* return status:
                             (0) PPROJ_SOLUTION_FOUND
                             (1) PPROJ_TOLERANCE_NOT_MET
                             (5) PPROJ_PROX_UPDATE */
(
    PPcom *I
)
{
    INT    i, j, k, l, m, p, p0, q, ibj, blk, nrow, ncol,
           ni, ni1, ni2, nrowindex, ncolindex, row, Ll, Ul, Rl,
           prevblkL, prevblkU, Annz, ATnz, Lnnz, nactive,
           ncoldel, nrowdel, ncoladd, nrowadd, nf,
           *Ai, *Anz, *Ap, *ATi, *ATp, *AFTp, *AFTi, *AFTnz,
           *ir, *ineq_row, *ColmodFlag, *ColmodList, *RowmodFlag, *RowmodList,
           *ColIndex, *RowIndex, *F, *Lnz, *ns, *RLinkUp, *RLinkDn,
           *lstart, *ustart, *lLinkUp, *lLinkDn, *uLinkUp, *uLinkDn,
           *sol_to_blk, *worki ;
    int blks, return_nochol, status, *ib ;
    double s, t, errdual, errls, errprox, colerr, rowerr, normb, normx,
           ax, bi, newbi, cj, loj, hij, rowcutoff, colcutoff,
           sigma, z, tic, bi_noprox,
           *b, *c, *Ax, *ATx, *AFTx, *D, *lambda, *dlambda, *lhs,
           *bl, *bu, *lo, *hi, *RowErri, *x, *workd ;

    PPstat *Stat ;
    PPparm *Parm ;
    PPprob *Prob ;
    PPwork    *W ;

#ifndef NDEBUG
    char *where ;
    I->Check->location = 5 ;
    where = "at start of check_err" ;
    checkF (I, where) ;
    checkD (I, where) ;
    checkb (I, where) ;
    check_modlist (I, where) ;
#endif

    tic = pproj_timer () ;
    Stat = I->Stat ;
    Parm = I->Parm ;
    Prob = I->Prob ;
    W = I->Work ;
    /* return_nochol = TRUE if code returns from pproj_dasa before a chol */
    return_nochol = W->return_nochol ;

    /* PPprob */
    Ap   = Prob->Ap ;
    Ai   = Prob->Ai ;
    Anz  = Prob->Anz ;
    Ax   = Prob->Ax ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    ni1 = ni + 1 ;
    ni2 = ni + 2 ;
    ineq_row = Prob->ineq_row ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    lo = Prob->lo ;
    hi = Prob->hi ;

    /* Transpose of A */
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;

    /* Transpose of AF */
    AFTp = W->AFTp ;
    AFTx = W->AFTx ;
    AFTi = W->AFTi ;
    AFTnz = W->AFTnz ;

    lambda = W->lambda ;
    dlambda = W->dlambda ;
    x = W->x ;

    /* ib [col] = +1 if column at upper bound
                  -1 if column at lower bound
                   0 if column is free */
    ib = W->ib ;

    /* ir [row] = 0         for an equality constraint
                =  strict # for an active inequality at upper bound
                = -strict # for an active inequality at lower bound
                =  strict # + ni for a dropped constraint */
    ir = W->ir ;
    b = W->b ; /* the part of grad L (lambda) associated with bound variables */
    c = W->c ; /* y + A'lambda */
    F = W->F ; /* free indices */
    D = W->D ; /* diag of AF*AF' */
    ns = W->ns ; /* used in line search, points from index to break point # */
    lo = W->lo ; /* lower bounds on x */
    hi = W->hi ; /* upper bounds on x */

    /* ColmodFlag is empty if row not modified, otherwise points into
       ColmodList showing columns to modify*/
    ColmodFlag = W->ColmodFlag ;
    ColmodList = W->ColmodList ;
    RowmodFlag = W->RowmodFlag ;
    RowmodList = W->RowmodList ;

    /* Links for active inequalities. lLinkUp points to the strict
       inequalities with b_i = bl_i while uLinkUp points to the strict
       inequalities with b_i = bu_i. Since only one of these can hold for
       each i, lLinkUp and uLinkUp can be stored in the same array */
    lLinkUp = W->SLinkUp ;
    lLinkDn = W->SLinkDn ;
    uLinkUp = W->SLinkUp ;
    uLinkDn = W->SLinkDn ;
    sol_to_blk = W->sol_to_blk ;

    /* Links for the active rows */
    RLinkUp = W->RLinkUp ;
    RLinkDn = W->RLinkDn ;

    /* lstart and ustart are associated with the inequalities that are
       strict.  For each block in the multilevel decomposition,
       lstart points to the first singleton at its lower bound while ustart
       points to the first singleton at it upper bound */
     lstart = W->lstart ;
     ustart = W->ustart ;

    blks = W->blks ; /* number of nodes (blocks) in the multilevel tree */

    /* work arrays used in checking error:

       lhs        - double nrow, grad L (lambda) (same space as RowErri)
       RowErri    - double ni, amount that currently dropped inequalities
                    violate bounds
       ColIndex   - INT ncol, column indices of column violators
       RowIndex   - INT ni, inequality index of row violators */

    workd = W->arrayd ;
    lhs = workd ; workd += nrow ;
    RowErri = lhs ;

    worki = W->arrayi ;
    RowIndex = worki ; worki += ni ;
    ColIndex = worki ; worki += ncol ;

    sigma = W->sigma ;

    errdual = ZERO ;
    errls = ZERO ;
    errprox = ZERO ;
    colerr = ZERO ;
    rowerr = ZERO ;
    normb = ZERO ;  /* sup norm of b */
    normx = ZERO ;  /* L1 norm of x */
    nrowindex = 0 ; /* number of dropped rows to activate */
    ncolindex = 0 ; /* number of free columns to bind */

    /* If errls, the norm of the residual in the linear system is large
       enough relative to the norm of the right side, then we refactor
       the matrix. To estimate the norm of the right right, we evaluate
       the norm of b. This neglects the change that occurred to the
       right side due to bound variables becoming free. */
    for (row = RLinkUp [nrow]; row < nrow; row = RLinkUp [row])
    {
        if ( normb < fabs (b [row]) )
        {
            normb = fabs(b [row]) ;
        }
    }
    pproj_initx (lhs, ZERO, nrow) ; /* used to store AF*cF */
    nf = 0 ;
    p = 0 ;
    for (j = 0; j < ncol; j++)
    {
        ibj = ib [j] ;
        if ( ibj == 0 ) /* currently free variable */
        {
            p = Ap [j] ;
            q = Ap [j+1] ;
            cj = c [j] ;
            if ( cj < lo [j] ) /* lower bind */
            {
                z = lo [j] ;
                lo [j] = ZERO ;
                cj -= z ;
                c [j] = cj ;
                /* save cj in case we decide later the violation was too small
                   and it is better to keep the variable free */
                ColIndex [ncolindex] = j ;
                ncolindex++ ;
                hi [j] -= z ;
                x [j] += z ;
                colerr = MAX (colerr, -cj) ;   /* max constraint violation */
                if ( z == ZERO )
                {
                    for (; p < q; p++)
                    {
                        lhs [Ai [p]] += Ax [p]*cj ; /* AF*cF */
                    }
                }
                else
                {
                    for (; p < q; p++)
                    {
                        i = Ai [p] ;
                        ax = Ax [p] ;
                        b [i] -= ax*z ;    /* new right side b */
                        lhs [i] += ax*cj ; /* AF*cF */
                    }
                }
            }
            else if ( cj > hi [j] ) /* upper bind */
            {
                z = hi [j] ;
                hi [j] = ZERO ;
                cj -= z ;
                c [j] = cj ;
                /* save cj in case we decide later the violation was too small*/
                ColIndex [ncolindex] = j ;
                ncolindex++ ;
                lo [j] -= z ;
                x [j] += z ;
                colerr = MAX (colerr, cj) ;   /* max constraint violation */
                if ( z == ZERO )
                {
                    for (; p < q; p++)
                    {
                        lhs [Ai [p]] += Ax [p]*cj ; /* AF*cF */
                    }
                }
                else
                {
                    for (; p < q; p++)
                    {
                        i = Ai [p] ;
                        ax = Ax [p] ;
                        b [i] -= ax*z ;    /* new right side b */
                        lhs [i] += ax*cj ; /* AF*cF */
                    }
                }
            }
            else
            {
                F [nf++] = j ;
                lo [j] -= cj ;
                hi [j] -= cj ;
                x [j] += cj ;
                c [j] = ZERO ;
                for (; p < q; p++)
                {
                    b [Ai [p]] -= Ax [p]*cj ;
                }
            }
        }
        normx += fabs (x [j]) ;
    }

    ATnz = 0 ;
    Lnnz = 0 ;
    Annz = 0 ;
    W->cholflops = 0 ;
    nactive = 0 ;
    if ( Stat->nchols > 0 )
    {
        Lnz = W->L->nz ;
    }
    for (i = 0; i < nrow; i++)
    {
        bi = b [i] ; /* includes prox term -sigma*lambda [i] */
        bi_noprox = bi + sigma*lambda [i] ;
        /* b [i] includes the proximal term, newbi also needs dlambda part */
        newbi = bi - sigma*dlambda [i] ;
        b [i] = newbi ;
        k = ir [i] ;
        if ( k <= ni ) /* i is an active row */
        {
            nactive++ ;
            ATnz += ATp [i+1] - ATp [i] ;
            Annz += AFTnz [i] ;
            /* residual in equation is newbi minus lhs */
            if ( errls < fabs (newbi - lhs [i]) )
            {
                errls = fabs (newbi - lhs [i]) ;
            }

            /* gradient of dual function excluding proximal term. */
            if ( errdual < fabs (bi_noprox) )
            {
                errdual = fabs (bi_noprox) ;
            }

            /* gradient of dual function including proximal term */
            if ( errprox < fabs (newbi) )
            {
                errprox = fabs (newbi) ;
            }

            /* update lambda (lambda does not change for dropped equations) */
            lambda [i] += dlambda [i] ;
            dlambda [i] = ZERO ;
            Rl = i ;
            if ( Stat->nchols > 0 )
            {
                Lnnz += Lnz [i] ;
                W->cholflops += ((double) Lnz [i]) * ((double) Lnz [i]) ;
            }
        }
        else   /* i is a dropped row */
        {
            k -= ni ; /* k = ineqindex */
            loj = bl [k] ;
            hij = bu [k] ;
            s = newbi + loj ;
            if ( s > ZERO ) /* lambda_i > 0 increases L */
            {
                if ( rowerr < s )
                {
                    rowerr = s ; /* max violation in dropped rows */
                }
                if ( Parm->PrintLevel > 0 )
                {
                    printf ("activate lower bound in row: %i grad: %e\n", i, s);
                }
                RowErri [nrowindex] = s ;
                ns [i] = 1 ;     /* activate row at lower bound */
                RowIndex [nrowindex] = k ;
                nrowindex++ ;
            }
            else
            {
                s = newbi + hij ;
                if ( s < ZERO )       /* lambda_i < 0 increases dual function */
                {
                    if ( rowerr < -s )
                    {
                        rowerr = -s ; /* max violation in dropped rows */
                    }
                    if ( Parm->PrintLevel > 0 )
                    {
                        printf ("activate upper bound in row: %i grad: %e\n",
                                i, s);
                    }
                    RowErri [nrowindex] = s ;
                    ns [i] = 2 ;      /* active row at upper bound */
                    RowIndex [nrowindex] = k ;
                    nrowindex++ ;
                }
            }
            /* for dualerr, do not include the proximal term */
            s = bi_noprox + loj ;
            if ( s > ZERO ) /* lambda_i > 0 increases L */
            {
                if ( errdual < s )
                {
                    errdual = s ; /* max violation in dropped rows */
                }
            }
            else
            {
                s = bi_noprox + hij ;
                if ( s < ZERO )       /* lambda_i < 0 increases dual function */
                {
                    if ( errdual < -s )
                    {
                        errdual = -s ; /* max violation in dropped rows */
                    }
                }
            }
        }
    }
    if ( errprox >= W->gamma*rowerr ) /* sparsa condition does not hold */
    {
        W->sparsaOK = FALSE ;
    }
    else                              /* condition for using sparsa holds */
    {
        W->sparsaOK = TRUE ;
    }
    errprox = MAX (errprox, rowerr) ; /* update errprox to include rowerr */

    /* if the stopping criterion is relative, not absolute, divide by absAx */
    if ( !Parm->stopabs )
    {
        errprox /= W->absAx ;
        errdual /= W->absAx ;
        errls   /= W->absAx ;
        normb   /= W->absAx ;
    }

    if ( Parm->PrintLevel > 0 )
    {
        printf ("\ndual error:            %e\n", errdual) ;
        printf ("linear system error:   %e\n", errls) ;
        printf ("proximal error:        %e\n", errprox) ;
        printf ("column error:          %e\n", colerr) ;
        printf ("row error:             %e\n", rowerr) ;
        printf ("1-norm of x:           %e\n", normx) ;
        printf ("absAx:                 %e\n", W->absAx) ;
        printf ("sup-norm of b:         %e\n", normb) ;
        printf ("# new bound columns:   %i\n", ncolindex) ;
        printf ("# new active rows:     %i\n", nrowindex) ;
        printf ("# current active rows: %i\n\n", nactive) ;
    }
    /* If the error in the residual for the linear system is sufficiently large,
       then refactor the matrix when possible */
    if ( errls >= normb*Parm->tolrefactor )
    {
        W->fac = FALSE ; /* matrix is not factored and needs to be factored */
        if ( Parm->PrintLevel > 0 )
        {
            printf ("Refactor, large relative error in solution\n") ;
        }
    }

    /* By default, convergence tolerance has not been met */
    status = PPROJ_TOLERANCE_NOT_MET ;

    /* If proximal error is sufficiently small, perform a proximal update */
    if ( errprox <= Parm->tolprox*errdual )
    {
        status = PPROJ_PROX_UPDATE ;
    }

    /* If the dual error is sufficiently small, the solution was found */
    if ( errdual <= Prob->grad_tol )
    {
        status = PPROJ_SOLUTION_FOUND ;
    }

    if ( Stat->nchols > 0 )
    {
        W->Lnnz = Lnnz ;            /* number of nonzero in L */
        W->npup_old = W->npup_cur ; /* number of partial updates */
    }

    /* If the solution was found or if a proximal update is to be performed,
       then we return to the main program. */

    if ( (status == PPROJ_SOLUTION_FOUND) || (status == PPROJ_PROX_UPDATE) )
    {
        /* restore ns to be all EMPTY */
        for (k = 0; k < nrowindex; k++)
        {
            i = ineq_row [RowIndex [k]] ;
            ns [i] = EMPTY ;
        }

        /* restore F since it was overwritten near the start of check_error */
        nf = 0 ;
        for (j = 0; j < ncol; j++)
        {
            if ( !ib [j] )
            {
                F [nf++] = j ;
            }
        }
        Stat->checkerr += pproj_timer () - tic ;
        W->nactive = nactive ;
        W->ATnz = ATnz ;
        W->Annz = Annz ;
        return (status) ;
    }

    /* Otherwise, get set for another iteration. Add dropped rows and
       bind columns for which the row error or the column error are
       sufficiently large. */
    if ( Parm->use_sparsa )
    {
        rowcutoff = ZERO ;
        colcutoff = ZERO ;
    }
    else
    {
        rowcutoff = Parm->cutfactor*rowerr ;
        colcutoff = Parm->cutfactor*colerr ;
    }

    ncoldel = W->ncoldel ;
    ncoladd = W->ncoladd ;
    nrowdel = W->nrowdel ;
    nrowadd = W->nrowadd ;

    /* ------------------------------------------------------------------ */
    /* check whether to activate query columns */
    /* ------------------------------------------------------------------ */
    for (k = 0; k < ncolindex; k++)
    {
        j = ColIndex [k] ;
        cj = c [j] ;
        if ( cj < 0 ) /* check lower bound */
        {
            if ( -cj > colcutoff )  /* lower bind j */
            {
                if ( Parm->PrintLevel > 1 )
                {
                    printf ("j: %i free -> lower (%e)\n", j, -cj) ;
                }
                ib [j] = -1 ;
                /* add column to the delete list if not in the add list */
                if ( ColmodFlag [j] == EMPTY )
                {
                    ncoldel++ ;
                    W->ColmodList [ncol-ncoldel] = j ;
                    ColmodFlag [j] = ncol-ncoldel ;
                }
                else /* column was in the add list, remove it */
                {
                    l = ColmodFlag [j] ;
                    ncoladd-- ;
                    m = ColmodList [ncoladd] ;
                    ColmodList [l] = m ;
                    ColmodFlag [m] = l ;
                    ColmodFlag [j] = EMPTY ;
                }
                if ( return_nochol == FALSE )
                {
                    q = Ap [j] + Anz [j] ;
                }
                else
                {
                    q = Ap [j+1] ;
                }
                for (p = Ap [j]; p < q; p++)
                {
                    i = Ai [p] ;
                    if ( ir [i] <= ni )
                    {
                        Annz-- ;
                        ax = Ax [p] ;
                        D [i] -= ax*ax ;
                        ns [i] = 0 ;
                    }
                }
            }
            else        /* although it is bound, keep it free */
            {
                if ( Parm->PrintLevel > 1 )
                {
                    printf ("j: %i keep free (%e)\n", j, -cj) ;
                }
                F [nf++] = j ;
                lo [j] -= cj ;
                hi [j] -= cj ;
                c [j] = ZERO ;
                x [j] += cj ;
                q = Ap [j+1] ;
                for (p = Ap [j]; p < q; p++)
                {
                    b [Ai [p]] -= Ax [p]*cj ; /* fix b */
                }
            }
        }
        else
        {
            if ( cj > colcutoff )   /* upper bind j */
            {
                if ( Parm->PrintLevel > 1 )
                {
                    printf ("j: %i free -> upper (%e)\n", j, cj) ;
                }
                ib [j] = +1 ;
                /* add column to the delete list if not in the add list */
                if ( ColmodFlag [j] == EMPTY )
                {
                    ncoldel++ ;
                    W->ColmodList [ncol-ncoldel] = j ;
                    ColmodFlag [j] = ncol-ncoldel ;
                }
                else /* column was in the add list, remove it */
                {
                    l = ColmodFlag [j] ;
                    ncoladd-- ;
                    m = ColmodList [ncoladd] ;
                    ColmodList [l] = m ;
                    ColmodFlag [m] = l ;
                    ColmodFlag [j] = EMPTY ;
                }
                if ( return_nochol == FALSE )
                {
                    q = Ap [j] + Anz [j] ;
                }
                else
                {
                    q = Ap [j+1] ;
                }
                for (p = Ap [j]; p < q; p++)
                {
                    i = Ai [p] ;
                    if ( ir [i] <= ni )
                    {
                        Annz-- ;
                        ax = Ax [p] ;
                        D [i] -= ax*ax ;
                        ns [i] = 0 ;
                    }
                }
            }
            else        /* although it is bound, keep it free */
            {
                if ( Parm->PrintLevel > 1 )
                {
                    printf ("j: %i keep free (%e)\n", j, cj) ;
                }
                F [nf++] = j ;
                lo [j] -= cj ;
                hi [j] -= cj ;
                c [j] = ZERO ;
                x [j] += cj ;
                q = Ap [j+1] ;
                for (p = Ap [j]; p < q; p++)
                {
                    b [Ai [p]] -= Ax [p]*cj ; /* fix b */
                }
            }
        }
    }


    /* ------------------------------------------------------------------ */
    /* check whether to activate query rows */
    /* ------------------------------------------------------------------ */
    /* 1. remove bound indices from AFT
       2. check to see if inactive row can be activated
       3. if inactive row is activated, then build row of AFT
       4. if inactive row remains inactive, then remove from link */

    if ( Parm->cholmod )
    {
        for (k = 0; k < blks; k++)
        {
            lstart [k] = ni1 ;
            ustart [k] = ni2 ;
        }
    }
    nrowindex = 0 ;
    Ll = ni1 ;
    Ul = ni2 ;
    Rl = nrow ;
    prevblkL = prevblkU = EMPTY ;
    for (i = 0; i < nrow; i++)
    {
        j = ns [i] ;
        k = ir [i] ;
        if ( j <= 0 )
        {
            if ( k <= ni ) /* currently active row */
            {
                Rl = i ;
                if ( k < 0 ) /* the row is active and at lower bound */
                {
                    k = -k ;
                    Ll = k ;
                    if ( Parm->cholmod == TRUE )
                    {
                        blk = sol_to_blk [k] ;
                        if ( blk > prevblkL )
                        {
                            lstart [blk] = k ;
                            prevblkL = blk ;
                        }
                    }
                }
                else if ( k > 0 )  /* the row is active and at upper bound */
                {
                    Ul = k ;
                    if ( Parm->cholmod == TRUE )
                    {
                        blk = sol_to_blk [k] ;
                        if ( blk > prevblkU )
                        {
                            ustart [blk] = k ;
                            prevblkU = blk ;
                        }
                    }
                }
            }
            if ( j == 0 ) /* this is an active row with new bound variables */
            {
                if ( D [i] < sigma )
                {
                    D [i] = sigma ;
                }
                p0 = l = p = AFTp [i] ;
                q = p + AFTnz [i] ;
                ASSERT (ir [i] <= ni) ;
                for (; p < q; p++)
                {
                    if ( ib [AFTi [p]] == 0 ) /* column is free */
                    {
                        AFTi [l] = AFTi [p] ;
                        AFTx [l] = AFTx [p] ;
                        l++ ;
                    }
                }
                AFTnz [i] = l - p0 ;
                ns [i] = EMPTY ;
            }
            continue ;
        }
        if ( j == 1 ) /* inactive row that may activate at lower bound */
        {
            k -= ni ;
            s = RowErri [nrowindex] ;
            if ( s > rowcutoff ) /* activate row */
            {
                if ( Parm->PrintLevel > 1 )
                {
                    printf ("i: %i inactive -> lo active (err: %e)\n",
                             i, RowErri [nrowindex]) ;
                }
                nactive++ ;
                dlambda [i] = ZERO ;
                ir [i] = -k ;   /* unDrop row */
                b [i] += bl [k] ;
                ATnz += ATp [i+1] - ATp [i] ;
                /* if the row not in the delete list, then add it */
                if ( RowmodFlag [k] == EMPTY )
                {
                    nrowadd++ ;
                    RowmodList [ni-nrowadd] = k ;
                    RowmodFlag [k] = ni-nrowadd ;
                }
                else /* remove row from the delete list */
                {
                    l = RowmodFlag [k] ;
                    nrowdel-- ;
                    m = RowmodList [nrowdel] ;
                    RowmodList [l] = m ;
                    RowmodFlag [m] = l ;
                    RowmodFlag [k] = EMPTY ;
                }

                /* add inequality to active list */
                l = lLinkUp [Ll] ;
                lLinkUp [k] = l ;
                lLinkUp [Ll] = k ;
                lLinkDn [l] = k ;
                lLinkDn [k] = Ll ;
                Ll = k ;

                /* adjust start of the block */
                if ( Parm->cholmod )
                {
                    blk = sol_to_blk [k] ;
                    if ( blk > prevblkL )
                    {
                        lstart [blk] = k ;
                        prevblkL = blk ;
                    }
                }
            }
            nrowindex++ ;
        }
        else if ( j == 2 ) /* inactive row may activate at upper bound */
        {
            k -= ni ;
            s = RowErri [nrowindex] ;
            if ( -s > rowcutoff ) /* activate row */
            {
                if ( Parm->PrintLevel > 1 )
                {
                    printf ("i: %i inactive -> hi active (err: %e)\n",
                    i, RowErri [nrowindex]) ;
                }
                nactive++ ;
                dlambda [i] = ZERO ;
                ir [i] = k ;   /* activate row */
                b [i] += bu [k] ;
                ATnz += ATp [i+1] - ATp [i] ;
                /* if the row not in the delete list, then add it */
                if ( RowmodFlag [k] == EMPTY )
                {
                    nrowadd++ ;
                    RowmodList [ni-nrowadd] = k ;
                    RowmodFlag [k] = ni-nrowadd ;
                }
                else /* remove row from the delete list */
                {
                    l = RowmodFlag [k] ;
                    nrowdel-- ;
                    m = RowmodList [nrowdel] ;
                    RowmodList [l] = m ;
                    RowmodFlag [m] = l ;
                    RowmodFlag [k] = EMPTY ;
                }

                /* add inequality to active list */
                l = uLinkUp [Ul] ;
                uLinkUp [k] = l ;
                uLinkUp [Ul] = k ;
                uLinkDn [l] = k ;
                uLinkDn [k] = Ul ;
                Ul = k ;

                /* adjust start of the block */
                if ( Parm->cholmod )
                {
                    blk = sol_to_blk [k] ;
                    if ( blk > prevblkU )
                    {
                        ustart [blk] = k ;
                        prevblkU = blk ;
                    }
                }

            }
            nrowindex++ ;
        }
        ns [i] = EMPTY ;
        /* if the row is newly added, update AFT and RLink */
        if ( ((j == 1)  || (j == 2)) && (ir [i] <= ni) )
        {
            /* add row to the list of active rows */
            l = RLinkUp [Rl]  ;
            RLinkUp [i] = l  ;
            RLinkUp [Rl] = i ;
            RLinkDn [l] = i ;
            RLinkDn [i] = Rl ;
            Rl = i ;

            /* insert row in AFT */
            t = sigma ;
            q = ATp [i+1] ;
            p = ATp [i] ;
            l = AFTp [i] ;
            for (; p < q; p++)
            {
                j = ATi [p] ;
                if ( !ib [j] ) /* the column is free */
                {
                    ax = ATx [p] ;
                    t += ax*ax ;
                    AFTx [l] = ax ;
                    AFTi [l] = j ;
                    l++ ;
                }
            }
            D [i] = t ;            /* new diagonal element for SSOR */
            m =  l - AFTp [i] ;
            AFTnz [i] = m ;
            Annz += m ;
            ASSERT (ir [i] <= ni) ;
        }
    }
    W->nrowadd = nrowadd ;
    W->ncoladd = ncoladd ;
    W->nrowdel = nrowdel ;
    W->ncoldel = ncoldel ;
    W->ATnz = ATnz ;
    W->Annz = Annz ;
    W->nactive = nactive ;
    W->nf = nf ;

    /* if relative change in L1 norm of x >= 1.5, recompute absAx */
    if ( normx != ZERO )
    {
        t = W->normx/normx ;
        W->normx = normx ;
        if ( (t >= 1.5) || (t <= .667) ) /* compute new absAx */
        {
            s = ZERO ;
            for (row = RLinkUp [nrow]; row < nrow; row = RLinkUp [row])
            {
                t = ZERO ;
                q = ATp [row+1] ;
                for (p = ATp [row]; p < q; p++)
                {
                    t += fabs (ATx [p]*x [ATi [p]]) ;
                }
                s = MAX (s, t) ;
            }
            W->absAx = s ;
        }
    }
    W->nf = nf ;

#ifndef NDEBUG
    where = "before exiting check_error" ;
    check_modlist (I, where) ;
    k = MAX (ncol+ni+1, nrow) ;
    check_const (NULL, 0, ns, EMPTY, k, where) ;
    checkF (I, where) ;
    if ( status == PPROJ_TOLERANCE_NOT_MET )
    {
        checkb (I, where) ;
        checkD (I, where) ;
    }
    check_dual (I, NULL, where, TRUE, TRUE) ;
#endif

    if ( Parm->PrintLevel > 0 )
    {
        printf ("final # active rows:   %i\n", nactive) ;
        printf ("final # free columns:  %i\n", nf) ;
        printf ("# added rows:          %i\n", nrowadd) ;
        printf ("# deleted columns:     %i\n", ncoldel) ;
        printf ("rowcutoff:             %e\n", rowcutoff) ;
        printf ("colcutoff:             %e\n", colcutoff) ;
        printf ("Lnnz:                  %i\n", Lnnz) ;
        printf ("Annz:                  %i\n", Annz) ;
        printf ("nchol:                 %i\n", W->nchols) ;
        printf ("cholflops:             %e\n", W->cholflops) ;
        printf ("partial updates:       %e\n", W->npup) ;
        printf ("average flops:         %e\n", W->pupflops/MAX (1, W->npup)) ;
        printf ("rooted  updates:       %e\n", W->nrup) ;
        printf ("average flops:         %e\n", W->rupflops/MAX (1, W->nrup)) ;
        fflush(stdout) ;
    }
    Stat->checkerr += pproj_timer () - tic ;
    return (status) ;
}

/* ========================================================================= */
/* === pproj_prox_update =================================================== */
/* ========================================================================= */
/* Perform a proximal update, recompute b and c, and compute grad L to
   determine whether the stopping criterion is satisfied. */
/* ========================================================================= */
int pproj_prox_update
(
    PPcom *I
)
{
    INT    i, j, k, l, p, q, nrow, ncol, ncoladd, ineqindex, ni, nf, nxmod,
           *Ai, *Ap, *ATi, *ATp, *AFTp, *AFTi, *AFTnz, *xmod_index,
           *ir, *ColmodFlag, *ColmodList, *F, *worki ;
    int    compute_rhsmod, set_to_bound, status, *ib ;
    double s, t, errdual, gradi, cj, xj, loj, hij, tic,
           *c, *Ax, *ATx, *AFTx, *D, *lambda, *shift_l,
           *lo, *hi, *x, *absAx, *rhsmod, *workb, *worklo, *workhi, *workd ;

    PPstat *Stat ;
    PPparm *Parm ;
    PPprob *Prob ;
    PPwork    *W ;

    tic = pproj_timer () ;
    Stat = I->Stat ;
    Parm = I->Parm ;
    Prob = I->Prob ;
    W = I->Work ;

    /* update sigma */
    W->sigma *= Parm->sigma_decay ;

    if ( Parm->cholmod )
    {
        W->cmm->dbound = W->sigma ;
        W->fac = FALSE ;              /* need to refactor the matrix */
    }

    /* PPprob */
    Ap   = Prob->Ap ;
    Ai   = Prob->Ai ;
    Ax   = Prob->Ax ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    lo = Prob->lo ;
    hi = Prob->hi ;

    c = W->c ;
    workb = W->b ;
    worklo = W->lo ;
    workhi = W->hi ;

    /* Transpose of A */
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;

    workd = W->arrayd ;
    rhsmod = workd ; workd += nrow ;
    absAx  = workd ; workd += nrow ;

    worki = W->arrayi ;
    xmod_index = worki ; worki += ncol ;

    lambda = W->lambda ;
    shift_l = W->shift_l ;
    x = W->x ;
    D = W->D ;

    /* ib [col] = +1 if column at upper bound
                  -1 if column at lower bound
                   0 if column is free */
    ib = W->ib ;

    /* ir [row] = 0         for an equality constraint
                =  strict # for an active inequality at upper bound
                = -strict # for an active inequality at lower bound
                =  strict # + ni for a dropped constraint */
    ir = W->ir ;

    if ( W->shiftl_is_zero )
    {
        pproj_copyx (shift_l, lambda, nrow) ;
    }
    else
    {
        pproj_saxpy (shift_l, lambda, ONE, nrow) ;
    }
    /* starting lambda = 0 */
    pproj_initx (lambda, ZERO, nrow) ;
    pproj_initx (absAx, ZERO, nrow) ;

    /* shift_l != 0 */
    W->shiftl_is_zero = FALSE ;
    /* recompute c */
    pproj_copyx (c, Prob->y, ncol) ;
    /* initialize diagonal of D to be the new sigma */
    pproj_initx (D, W->sigma, nrow) ;
    /* compute c by rows of A, set shift_l to zero for dropped rows */
    /* set b_i to bl, bu, or 0 */
    for (i = 0; i < nrow; i++)
    {
        k = ir [i] ;
        if ( k <= ni ) /* row is active */
        {
            if ( k == 0 )
            {
                s = Prob->b [i] ;
            }
            else if ( k < 0 )
            {
                s = Prob->bl [-k] ;
            }
            else
            {
                s = Prob->bu [k] ;
            }
            t = shift_l [i] ;
            if ( t != ZERO )
            {
                q = ATp [i+1] ;
                for (p = ATp [i]; p < q; p++)
                {
                    c [ATi [p]] += t*ATx [p] ;
                }
            }
        }
        else               /* row is inactive */
        {
            shift_l [i] = ZERO ;
            s = ZERO ;
        }
        workb [i] = s ;
    }

    /* Compute the gradient of the dual function, absAx_sup, and check to
       see if the convergence tolerance has been satisfied.  Due to
       rounding errors, the ib flag may indicate that a component
       of x is bound, while the value of c may imply that the minimizer
       of the dual function is free. In this case, we need to change
       the ib value to 0 to free the variable. We compute both b,
       which corresponds to b0 in the pproj code, and a correction
       rhsmod that is added to b to obtain the dual gradient. */
    nf = W->nf ;
    AFTp = W->AFTp ;
    AFTi = W->AFTi ;
    AFTnz = W->AFTnz ;
    AFTx = W->AFTx ;
    F = W->F ;
    pproj_initx (rhsmod, ZERO, nrow) ;
    ncoladd = W->ncoladd ;
    /* ColmodFlag is empty if row not modified, otherwise points into
       ColmodList showing columns to modify*/
    ColmodList = W->ColmodList ;
    ColmodFlag = W->ColmodFlag ;
    /* At this point, we do not know if we will terminate. Save a copy
       in xmod_index of the indices of components of x that may need to
       be adjusted if we have reached the stopping criterion. */
    nxmod = 0 ;
    p = 0 ;
    for (j = 0; j < ncol; j++)
    {
        k = ib [j] ;
        cj = c [j] ;
        q = Ap [j+1] ;
        if ( k == 0 ) /* treat variable as free */
        {
            if ( cj < lo [j] )
            {
                xmod_index [nxmod] = j ;
                nxmod++ ;
                xj = lo [j] ;
                compute_rhsmod = TRUE ;
            }
            else if ( cj > hi [j] )
            {
                xmod_index [nxmod] = j ;
                nxmod++ ;
                xj = hi [j] ;
                compute_rhsmod = TRUE ;
            }
            else
            {
                compute_rhsmod = FALSE ;
            }
            worklo [j] = lo [j] - cj ;
            workhi [j] = hi [j] - cj ;
            x [j] = cj ;
            c [j] = ZERO ;
            /* If ib [j] = 0, we treat the variable as free, but in
               grad L, the variable could be bound. Hence, we compute
               rhsmod, which is the correction that is added to b
               to obtain grad L */
            if ( compute_rhsmod )
            {
                t = cj - xj ;
                for (; p < q; p++)
                {
                    INT row = Ai [p] ;
                    double ax = Ax [p] ;
                    D [row] += ax*ax ;
                    absAx [row] += fabs (cj*ax) ;
                    workb [row] -= cj*ax ;
                    rhsmod [row] += t*ax ;
                }
            }
            else
            {
                for (; p < q; p++)
                {
                    INT row = Ai [p] ;
                    double ax = Ax [p] ;
                    D [row] += ax*ax ;
                    absAx [row] += fabs (cj*ax) ;
                    workb [row] -= cj*ax ;
                }
            }
        }
        else /* currently bound variable */
        {
            if ( cj >= hi [j] ) /* at upper bound */
            {
                set_to_bound = TRUE ;
                ib [j] = +1 ;
                t = hi [j] ;
                worklo [j] = lo [j] - t ;
                workhi [j] = ZERO ;
            }
            else if ( cj <= lo [j] ) /* at lower bound */
            {
                set_to_bound = TRUE ;
                ib [j] = -1 ;
                t = lo [j] ;
                workhi [j] = hi [j] - t ;
                worklo [j] = ZERO ;
            }
            else /* this bound variable does not satisfy a bound, hence,
                    we must make it free */
            {
                set_to_bound = FALSE ;
            }
            if ( set_to_bound ) /* the variable remains bound*/
            {
                x [j] = t ;
                c [j] -= t ;
                for (; p < q; p++)
                {
                    INT row = Ai [p] ;
                    double ax = Ax [p] ;
                    absAx [row] += fabs (t*ax) ;
                    workb [row] -= t*ax ;
                }
            }
            else /* the variable is bound but set free */
            {
                if ( Parm->PrintLevel > 0 )
                {
                    printf ("col %i is bound (%i) but set to free in "
                            "pproj based on lo: %e cj: %e hi: %e\n",
                             j, ib [j], lo [j], cj, hi [j]) ;
                }
                F [nf] = j ;
                nf++ ;
                ib [j] = 0 ;
                ColmodList [ncoladd] = j ; /* use 1st part for add */
                ASSERT (ColmodFlag [j] == EMPTY) ;
                ColmodFlag [j] = ncoladd ;
                ncoladd++ ;
                x [j] = cj ;
                workhi [j] = hi [j] - cj ;
                worklo [j] = lo [j] - cj ;
                c [j] = ZERO ;
                for (; p < q; p++)
                {
                    INT row = Ai [p] ;
                    double ax = Ax [p] ;
                    absAx [row] += fabs (cj*ax) ;
                    workb [row] -= cj*ax ;
                    if ( ir [row] <= ni )
                    {
                        D [row] += ax*ax ;
                        l = AFTp [row] + AFTnz [row]++ ;
                        AFTx [l] = ax ;
                        AFTi [l] = j ;
                    }
                }
            }
        }
    }
    W->ncoladd = ncoladd ;
    W->nf = nf ;
    errdual = ZERO ;
    t = ZERO ;
    for (i = 0; i < nrow; i++)
    {
        /* add rhsmod to b to obtain gradient of dual function */
        gradi = workb [i] + rhsmod [i] ;
        k = ir [i] ;
        if ( k <= ni ) /* i is an active row */
        {
            if ( t < absAx [i] )
            {
                t = absAx [i] ;
            }
            if ( errdual < fabs (gradi) )
            {
                errdual = fabs (gradi) ;
            }
        }
        else /* row is inactive, subgradient = [gradi+loj, gradi+hij] */
        {
            ineqindex = k - ni ;
            loj = Prob->bl [ineqindex] ;
            s = gradi + loj ;
            if ( s > ZERO ) /* lambda_i > 0 increases L */
            {
                if ( errdual < s )
                {
                    errdual = s ; /* max violation in dropped rows */
                }
            }
            else
            {
                hij = Prob->bu [ineqindex] ;
                s = gradi + hij ;
                if ( s < ZERO )   /* lambda_i < 0 increases dual function */
                {
                    if ( errdual < -s )
                    {
                        errdual = -s ; /* max violation in dropped rows */
                    }
                }
            }
        }
    }
    if ( t == ZERO )
    {
        t = ONE ;
    }
    W->absAx = t ;
    if ( !Parm->stopabs )
    {
        errdual /= t ;
    }
    if ( (errdual <= Prob->grad_tol) )
    {
        Stat->errdual = errdual ;
        status = PPROJ_SOLUTION_FOUND ;
        /* Adjust the xmod components of x so that they satisfy the bound
           constraints. */
        for (k = 0; k < nxmod; k++)
        {
            j = xmod_index [k] ;
            if ( x [j] < lo [j] )
            {
                x [j] = lo [j] ;
            }
            else
            {
                x [j] = hi [j] ;
            }
        }
    }
    else
    {
        status = PPROJ_TOLERANCE_NOT_MET ;
        if ( Parm->PrintLevel > 0 )
        {
            printf ("\nprox update dual error %e\n", errdual) ;
        }
    }
#ifndef NDEBUG
    /* record new dual value but do not insist on objective increase */
    check_dual (I, NULL, "in pproj after proximal step", TRUE, FALSE) ;
#endif
    Stat->nprox++ ;
    Stat->prox_update += pproj_timer () - tic ;
    return (status) ;
}

/* ==========================================================================
   ====== pproj_invert ======================================================
   ==========================================================================
   Invert any row permutations applied to lambda and any column permutations
   applied to x to obtain the solution relative to the starting problem.
   ========================================================================== */

void pproj_invert
(
    PPcom *I
)
{
    double xj, tic, *lambda, *newlambda, *x, *newx, *workd, *lo, *hi ;
    INT    i, j, nrow, ncol, *colperm, *rowperm ;
    PPwork    *W ;
    PPprob *Prob ;
    PPparm *Parm ;

    tic = pproj_timer () ;
    W = I->Work ;
    Prob = I->Prob ;
    Parm = I->Parm ;
    nrow = Prob->nrow ;
    ncol = Prob->ncol ;
    rowperm = Prob->rowperm ;
    colperm = Prob->colperm ;
    lo = Prob->lo ;
    hi = Prob->hi ;
    lambda = W->lambda ;

    if ( Parm->cholmod == TRUE ) /* invert the row and column permutations */
    {
        /* allocate the working arrays */
        workd = W->arrayd ;
        newlambda = workd ; workd += nrow ;
        newx      = workd ; workd += ncol ;
    
        /* rowperm [i] is the row number of the original matrix corresponding
           to row i of the permuted matrix */
        if ( !W->shiftl_is_zero )
        {
            /* if shift_l is nonzero, then add it to lambda */
            for (i = 0; i < nrow; i++)
            {
                newlambda [rowperm [i]] = lambda [i] + W->shift_l [i] ;
            }
        }
        else
        {
            for (i = 0; i < nrow; i++)
            {
                newlambda [rowperm [i]] = lambda [i] ;
            }
        }
        pproj_copyx (lambda, newlambda, nrow) ; /* copy newlambda to lambda */
    
        /* colperm [j] = column of original matrix corresponding to column j in
                         permuted matrix */
        x = W->x ;
        for (j = 0; j < ncol; j++)
        {
            xj = x [j] ;
            if ( xj < lo [j] )
            {
                xj = lo [j] ;
            }
            else if ( xj > hi [j] )
            {
                xj = hi [j] ;
            }
            newx [colperm [j]] = xj ;
        }
        pproj_copyx (x, newx, ncol) ; /* copy newx to x */
    }
    else /* iterative methods do not permute */
    {
        if ( W->shiftl_is_zero == FALSE )
        {
            /* if shift_l is nonzero, then add it to lambda */
            for (i = 0; i < nrow; i++)
            {
                lambda [i] += W->shift_l [i] ;
            }
        }
        for (j = 0; j < ncol; j++)
        {
            xj = W->x [j] ;
            if ( xj < lo [j] )
            {
                xj = lo [j] ;
            }
            else if ( xj > hi [j] )
            {
                xj = hi [j] ;
            }
            W->x [j] = xj ;
        }
    }
    I->Stat->invert += pproj_timer () - tic ;
}

/* =========================================================================
   =========================== pproj_cholmod_sparse ========================
   =========================================================================
    Create a cholmod sparse matrix.
   ========================================================================== */

void pproj_cholmod_sparse
(
    cholmod_sparse *A, /* pointer to a cholmod sparse matrix */
    INT          nrow, /* number of rows */
    INT          ncol, /* number of columns */
    INT         nzmax, /* max number of nonzeros */
    INT            *p, /* size ncol + 1, the column pointers */
    INT           *nz, /* size ncol, # nonzeros in each column, could be NULL */
    INT            *i, /* size Ap [ncol], the row indices */
    double         *x, /* size Ap [ncol], the numerical values in A */
    int        sorted, /* TRUE if columns of A are sorted, FALSE otherwise */
    int        packed, /* TRUE if A is packed, FALSE otherwise */
    int         xtype  /* CHOLMOD_PATTERN or CHOLMOD_REAL */
)
{
    A->nrow = nrow ;
    A->ncol = ncol ;
    A->p = p ;
    A->nz = nz ;
    A->i = i ;
    A->x = x ;
    A->packed = packed ;
    A->sorted = sorted ;
    A->xtype = xtype ;
    A->nzmax = nzmax ;
    A->z = NULL ;
    A->stype = 0 ; /* unsymmetric */
    /* CAUTION: need to sink these with cholmod */
    A->itype = CHOLMOD_INT ;
    A->dtype = CHOLMOD_DOUBLE ;
}

/* =========================================================================
   =========================== pproj_cholmod_dense =========================
   =========================================================================
    Create a cholmod dense vector.
   ========================================================================== */

void pproj_cholmod_dense
(
    cholmod_dense *A, /* pointer to a cholmod dense matrix */
    INT         nrow, /* number of rows */
    double        *x  /* size nrow, the numerical values in A */
)
{
    A->nrow = nrow ;
    A->ncol = 1 ;
    A->x = x ;
    A->nzmax = nrow ;
    /* CAUTION: need to sink these with cholmod */
    A->xtype = CHOLMOD_REAL ;
    A->dtype = CHOLMOD_DOUBLE ;
    A->z = NULL ;
    A->d = nrow ;
}

/* ========================================================================== */
/* === pproj_transpose ====================================================== */
/* ========================================================================== */
/*    Transpose a sparse matrix: B = A' */
/* ========================================================================== */

void pproj_transpose
(
    INT    *Bp,   /* size nrow+1, column pointers (output) */
    INT    *Bi,   /* size Ap [ncol], row indices of B (output) */
    double *Bx,   /* size Ap [ncol], numerical entries of B (output) */
    INT    *Ap,   /* size ncol+1, column pointers */
    INT    *Ai,   /* size Ap [ncol], row indices for A */
    double *Ax,   /* size Ap [ncol], numerical entries of A */
    INT   nrow,   /* number of rows in A */
    INT   ncol,   /* number of cols in A */
    INT     *W    /* work array of size nrow */
)
{
    INT i, j, p, q, pp ;

    /* ====================================================================== */
    /* === compute row counts of A ========================================== */
    /* ====================================================================== */

    for (i = 0; i < nrow; i++)
    {
        W [i] = 0 ;
    }
    p = 0 ;
    for (j = 1; j <= ncol; j++)
    {
        q = Ap [j] ;
        for (; p < q; p++)
        {
            W [Ai [p]]++ ;
        }
    }

    /* ====================================================================== */
    /* === compute column pointers of B given the counts ==================== */
    /* ====================================================================== */


    Bp [0] = 0 ;
    for (i = 0; i < nrow; i++)
    {
        Bp [i+1] = Bp [i] + W [i] ;
        W [i] = Bp [i] ;
    }

    /* ====================================================================== */
    /* === B = A' =========================================================== */
    /* ====================================================================== */

    p = 0 ;
    for (j = 0 ; j < ncol ; j++)
    {
        q = Ap [j+1] ;
        for (; p < q; p++)
        {
            pp = W [Ai [p]]++ ;
            Bi [pp] = j ;
            Bx [pp] = Ax [p] ;
        }
    }
}

    /* ====================================================================== */
    /* === pproj_iminsort =================================================== */
    /* ======================================================================
       ________________________________________________________
      |                                                        |
      |       sort an INT array in increasing order            |
      |                                                        |
      |    input:                                              |
      |                                                        |
      |         x  --array of numbers (INT *) of length n      |
      |         w  --working array (INT *) of length n         |
      |         n  --number of array elements to sort          |
      |                                                        |
      |    output:                                             |
      |                                                        |
      |         x  --original array (INT *) of length n        |
      |         y  --indices of x giving increasing order      |
      |              (INT *) of length n                       |
      |________________________________________________________| */

void pproj_iminsort
(
    INT         *y,  /* n-by-1 (output) */
    INT         *x , /* n-by-1 (input not modified) */
    INT         *w,  /* n-by-1, (input, working array) */
    INT          n   /* number of elements to sort */
)
{
    INT *yi, *wi, i, j, k, l, m, p, q ;
    INT s, t ;

    y [0] = 0 ;
    if ( n < 2 ) return ;

    j = k = 0 ;
    for (i = 1; i < n; i++)
    {
        if ( x [i] < x [j] )
        {
            w [k] = i ;
            k = i ;
        }
        y [i] = j = i ;
    }

    w [k] = n ;
    while ( k > 0 )
    {
        l = m = 0 ;
        while ( l < n )
        {
            i = l ;
            p = y [i] ;
            s = x [p] ;
            j = w [i] ;
            k = j ;
            if ( j == n )
            {
                y [i] = j ;
                l = j ;
                w [m] = p ;
                k += m - i ;
                yi = y+(i-m) ;
                for (m++; m < k; m++)
                {
                    w [m] = yi [m] ;
                }
            }
            else
            {
                q = y [j] ;
                t = x [q] ;
                l = w [j] ;
                y [i] = l ;
                while ( 1 )
                {
                    if ( s > t )
                    {
                        w [m] = q ;
                        m++ ;
                        j++ ;
                        if ( j == l )
                        {
                            w [m] = p ;
                            k += m - i ;
                            yi = y+(i-m) ;
                            for (m++; m < k; m++)
                            {
                                w [m] = yi [m] ;
                            }
                            break ;
                        }
                        q = y [j] ;
                        t = x [q] ;
                    }
                    else
                    {
                        w [m] = p ;
                        m++ ;
                        i++ ;
                        if ( i == k )
                        {
                            w [m] = q ;
                            k = m + l - j ;
                            yi = y+(j-m) ;
                            for (m++; m < k; m++)
                            {
                                w [m] = yi [m] ;
                            }
                            break ;
                        }
                        p = y [i] ;
                        s = x [p] ;
                    }
                }
            }
        }
        if ( y [0] == n )
        {
            for (i = 0; i < n; i++)
            {
                y [i] = w [i] ;
            }
            return ;
        }

        l = m = 0 ;
        while ( l < n )
        {
            i = l ;
            p = w [i] ;
            s = x [p] ;
            j = y [i] ;
            k = j ;
            if ( j == n )
            {
                w [i] = j ;
                l = j ;
                y [m] = p ;
                k += m - i ;
                wi = w+(i-m) ;
                for (m++; m < k; m++)
                {
                    y [m] = wi [m] ;
                }
            }
            else
            {
                q = w [j] ;
                t = x [q] ;
                l = y [j] ;
                w [i] = l ;
                while ( 1 )
                {
                    if ( s > t )
                    {
                        y [m] = q ;
                        m++ ;
                        j++ ;
                        if ( j == l )
                        {
                            y [m] = p ;
                            k += m - i ;
                            wi = w+(i-m) ;
                            for (m++; m < k; m++)
                            {
                                y [m] = wi [m] ;
                            }
                            break ;
                        }
                        q = w [j] ;
                        t = x [q] ;
                    }
                    else
                    {
                        y [m] = p ;
                        m++ ;
                        i++ ;
                        if ( i == k )
                        {
                            y [m] = q ;
                            k = m + l - j ;
                            wi = w+(j-m) ;
                            for (m++; m < k; m++)
                            {
                                y [m] = wi [m] ;
                            }
                            break ;
                        }
                        p = w [i] ;
                        s = x [p] ;
                    }
                }
            }
        }
        if ( y [0] == n ) return ;
    }
}

    /* ====================================================================== */
    /* === pproj_xminsort =================================================== */
    /* ======================================================================
       ________________________________________________________
      |                                                        |
      |       sort a double array in increasing order          |
      |                                                        |
      |    input:                                              |
      |                                                        |
      |         x     --array of numbers (double *) of length n|
      |         w     --working array (INT *) of length n      |
      |         n     --number of array elements to sort       |
      |                                                        |
      |    output:                                             |
      |                                                        |
      |         x     --original array (double *) of length n  |
      |         y     --indices of x giving increasing order   |
      |                 (INT *) of length n                    |
      |________________________________________________________| */

void pproj_xminsort
(
    INT    *y,  /* n-by-1 (output) */
    double *x , /* n-by-1 (input not modified) */
    INT    *w,  /* n-by-1, (input, working array) */
    INT     n   /* number of elements to sort */
)
{
    INT *yi, *wi, i, j, k, l, m, p, q ;
    double s, t ;

    y [0] = 0 ;
    if ( n < 2 ) return ;
    if ( n < 3 )
    {
        if ( x [0] > x [1] )
        {
            y [0] = 1 ;
            y [1] = 0 ;
        }
        else y [1] = 1 ;
        return ;
    }

    j = k = 0 ;
    for (i = 1; i < n; i++)
    {
        if ( x [i] < x [j] )
        {
            w [k] = i ;
            k = i ;
        }
        y [i] = j = i ;
    }

    w [k] = n ;
    while ( k > 0 )
    {
        l = m = 0 ;
        while ( l < n )
        {
            i = l ;
            p = y [i] ;
            s = x [p] ;
            j = w [i] ;
            k = j ;
            if ( j == n )
            {
                y [i] = j ;
                l = j ;
                w [m] = p ;
                k += m - i ;
                yi = y+(i-m) ;
                for (m++; m < k; m++) w [m] = yi [m] ;
            }
            else
            {
                q = y [j] ;
                t = x [q] ;
                l = w [j] ;
                y [i] = l ;
                while ( 1 )
                {
                    if ( s > t )
                    {
                        w [m] = q ;
                        m++ ;
                        j++ ;
                        if ( j == l )
                        {
                            w [m] = p ;
                            k += m - i ;
                            yi = y+(i-m) ;
                            for (m++; m < k; m++) w [m] = yi [m] ;
                            break ;
                        }
                        q = y [j] ;
                        t = x [q] ;
                    }
                    else
                    {
                        w [m] = p ;
                        m++ ;
                        i++ ;
                        if ( i == k )
                        {
                            w [m] = q ;
                            k = m + l - j ;
                            yi = y+(j-m) ;
                            for (m++; m < k; m++) w [m] = yi [m] ;
                            break ;
                        }
                        p = y [i] ;
                        s = x [p] ;
                    }
                }
            }
        }
        if ( y [0] == n )
        {
            for (i = 0; i < n; i++) y [i] = w [i] ;
            return ;
        }

        l = m = 0 ;
        while ( l < n )
        {
            i = l ;
            p = w [i] ;
            s = x [p] ;
            j = y [i] ;
            k = j ;
            if ( j == n )
            {
                w [i] = j ;
                l = j ;
                y [m] = p ;
                k += m - i ;
                wi = w+(i-m) ;
                for (m++; m < k; m++) y [m] = wi [m] ;
            }
            else
            {
                q = w [j] ;
                t = x [q] ;
                l = y [j] ;
                w [i] = l ;
                while ( 1 )
                {
                    if ( s > t )
                    {
                        y [m] = q ;
                        m++ ;
                        j++ ;
                        if ( j == l )
                        {
                            y [m] = p ;
                            k += m - i ;
                            wi = w+(i-m) ;
                            for (m++; m < k; m++) y [m] = wi [m] ;
                            break ;
                        }
                        q = w [j] ;
                        t = x [q] ;
                    }
                    else
                    {
                        y [m] = p ;
                        m++ ;
                        i++ ;
                        if ( i == k )
                        {
                            y [m] = q ;
                            k = m + l - j ;
                            wi = w+(j-m) ;
                            for (m++; m < k; m++) y [m] = wi [m] ;
                            break ;
                        }
                        p = w [i] ;
                        s = x [p] ;
                    }
                }
            }
        }
        if ( y [0] == n ) return ;
    }
}

/* =========================================================================
   ======================== pproj_copyI ====================================
   =========================================================================
   Copy INT vector x into vector y
   ========================================================================= */
void pproj_copyI
(
    INT *y, /* output of copy */
    INT *x, /* input of copy */
    INT  n  /* length of vectors */
)
{
    INT i, n5 ;
    n5 = n % 5 ;
    for (i = 0; i < n5; i++) y [i] = x [i] ;
    for (; i < n; )
    {
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
    }
}

/* =========================================================================
   ======================== pproj_copyi ====================================
   =========================================================================
   Copy int vector x into vector y
   ========================================================================= */
void pproj_copyi
(
    int *y, /* output of copy */
    int *x, /* input of copy */
    INT  n  /* length of vectors */
)
{
    INT i, n5 ;
    n5 = n % 5 ;
    for (i = 0; i < n5; i++) y [i] = x [i] ;
    for (; i < n; )
    {
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
    }
}

/* =========================================================================
   ======================== pproj_copyx ====================================
   =========================================================================
   Copy double vector x into vector y
   ========================================================================= */
void pproj_copyx
(
    double *y, /* output of copy */
    double *x, /* input of copy */
    INT     n  /* length of vectors */
)
{
    INT i, n5 ;
    n5 = n % 5 ;
    for (i = 0; i < n5; i++) y [i] = x [i] ;
    for (; i < n; )
    {
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
        y [i] = x [i] ;
        i++ ;
    }
}

/* =========================================================================
   ===================== proj_initI ========================================
   =========================================================================
   Initialize an INT array
   ========================================================================= */
void pproj_initI
(
    INT *x,  /* array to be initialized */
    INT  s,  /* scalar */
    INT  n   /* length of x */
)
{
    INT j, n5 ;
    INT *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

/* =========================================================================
   ===================== proj_initi ========================================
   =========================================================================
   Initialize an int array
   ========================================================================= */
void pproj_initi
(
    int *x,  /* array to be initialized */
    int  s,  /* scalar */
    INT  n   /* length of x */
)
{
    INT j, n5 ;
    int *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

/* =========================================================================
   ===================== proj_initx ========================================
   =========================================================================
   Initialize a double array
   ========================================================================= */
void pproj_initx
(
    double *x,  /* array to be initialized */
    double  s,  /* scalar */
    INT     n   /* length of x */
)
{
    INT j, n5 ;
    double *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

/* =========================================================================
   ===================== proj_initFx =======================================
   =========================================================================
   Initialize a double array at indices in F
   ========================================================================= */
void pproj_initFx
(
    double *x,  /* array to be initialized */
    double  s,  /* scalar */
    INT    *F,  /* indices to be initialized */
    INT     n   /* length of F */
)
{
    INT j, n5, *fj ;
    fj = F ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [F [j]] = s ;
    fj = F+j ;
    for (; j < n; j += 5)
    {
        *(x+(*fj++)) = s ;
        *(x+(*fj++)) = s ;
        *(x+(*fj++)) = s ;
        *(x+(*fj++)) = s ;
        *(x+(*fj++)) = s ;
    }
}

/* =========================================================================
   ===================== proj_scale ========================================
   =========================================================================
   Scale a double array
   ========================================================================= */
void pproj_scale
(
    double *x,  /* array to be scaled */
    double *y,  /* array used for the scaling */
    double  s,  /* scale */
    INT     n   /* length of x */
)
{
    INT j, n5 ;
    double *xj, *yj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s*y [j] ;
    xj = x+j ;
    yj = y+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s*(*yj++) ;
        *(xj++) = s*(*yj++) ;
        *(xj++) = s*(*yj++) ;
        *(xj++) = s*(*yj++) ;
        *(xj++) = s*(*yj++) ;
    }
}

/* =========================================================================
   ===================== proj_dot ==========================================
   =========================================================================
   Dot product between double arrays
   ========================================================================= */
double pproj_dot
(
    double *x,
    double *y,
    INT     n  /* length of x */
)
{
    INT j, n5 ;
    double *xj, *yj, t ;
    n5 = n % 5 ;
    t = ZERO ;
    for (j = 0; j < n5; j++) t += x [j]*y [j] ;
    xj = x+j ;
    yj = y+j ;
    for (; j < n; j += 5)
    {
        t += (*xj++)*(*yj++) ;
        t += (*xj++)*(*yj++) ;
        t += (*xj++)*(*yj++) ;
        t += (*xj++)*(*yj++) ;
        t += (*xj++)*(*yj++) ;
    }
    return (t) ;
}

/* ==========================================================================
   ======================== pproj_minheap_build =============================
   ==========================================================================
   build a max heap in heap [1..nheap]
   ========================================================================== */

void pproj_minheap_build
(
    INT   *heap, /* on input, an unsorted set of element numbers */
    double   *x, /* the numerical values to be ordered */
    INT   nheap  /* number of elements to build into the heap */
)
{
    INT p ;
    for (p = nheap/2; p >= 1; p--)
    {
        pproj_minheap_ify (p, heap, x, nheap) ;
    }
}

/* ========================================================================== */
/* ===================== pproj_minheapify =================================== */
/* ========================================================================== */
/* heapify starting at node p.  On input, the heap below node p satisfies the */
/* heap property, except for heap [p] itself.  On output, the whole heap */
/* from p and below satisfies the heap property. */
/* ========================================================================== */

void pproj_minheap_ify
(
    INT      p, /* start at node p in the heap */
    INT  *heap, /* size n, containing indices into x */
    double  *x, /* not modified */
    INT  nheap  /* heap [1 ... nheap] is in use */
)
{
    INT left, right, e, hleft, hright ;
    double xleft, xright, xe ;

    e = heap [p] ;
    xe = x [e] ;

    while ( 1 )
    {
        left = p * 2 ;
        right = left + 1 ;

        if (right <= nheap)
        {
            hleft  = heap [left] ;
            hright = heap [right] ;
            xleft  = x [hleft] ;
            xright = x [hright] ;
            if (xleft < xright)
            {
                if (xe > xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
            else
            {
                if (xe > xright)
                {
                    heap [p] = hright ;
                    p = right ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
        }
        else
        {
            if (left <= nheap)
            {
                hleft = heap [left] ;
                xleft = x [hleft] ;
                if (xe > xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
            }
            heap [p] = e ;
            return ;
        }
    }
}

/* ========================================================================== */
/* ======================== pproj_minheap_add =============================== */
/* ========================================================================== */
/* add a new leaf to a min-heap */
/* ========================================================================== */

void pproj_minheap_add
(
    INT     leaf , /* the new leaf */
    INT    *heap , /* size n, containing indices into x */
    INT      *ns , /* pointer from node to store in heap */
    double    *x , /* not modified */
    INT   *nheap   /* number of elements in heap including new one */
)
{
    INT l, new, old ;
    double xold, xnew ;

    (*nheap)++ ;
    old = *nheap ;
    heap [old] = leaf ;
    xold = x [leaf] ;
    ns [leaf] = old ;
    while ( old > 1 )
    {
        new = old / 2 ;
        l = heap [new] ;
        xnew = x [l] ;
        if ( xnew > xold )
        {

/* swap new and old */

            ns [l] = old ;
            ns [leaf] = new ;
            heap [new] = leaf ;
            heap [old] = l ;
        }
        else return ;
        old = new ;
    }
}

/* ========================================================================== */
/* ======================== pproj_minheap_delete ============================ */
/* ========================================================================== */
/* delete the first element from a min-heap */
/* ========================================================================== */

void pproj_minheap_delete
(
    INT  *heap , /* containing indices into x, 1..n on input */
    INT    *ns , /* pointer from node to store */
    double  *x , /* not modified */
    INT *nheap , /* number of items in heap */
    INT      p   /* element to delete from the heap */
)
{

    INT j ;

/* Move element from the end of the heap to position p */

    j = heap [p] ;
    if ( *nheap == p )
    {
        ns [j] = EMPTY ;
        (*nheap)-- ;
        return ;
    }
    heap [p] = heap [(*nheap)--] ;
    ns [heap [p]] = p ;
    ns [j] = EMPTY ;

/* Update the heap at position p */

    pproj_minheap_update (heap, ns, x, *nheap, p) ;
}

/* ========================================================================== */
/* ======================== pproj_minheap_update ============================ */
/* ========================================================================== */
/* update the location of an element in a min-heap */
/* ========================================================================== */

void pproj_minheap_update
(
    INT   *heap , /* size n, containing indices into x */
    INT     *ns , /* pointer from node to store in heap */
    double   *x , /* not modified */
    INT   nheap , /* number of elements in the heap */
    INT       p   /* location of element to update */
)
{
    INT l, leaf, new, old;
    double xnew, xold ;

/* Bubble up the heap */

    old = p ;
    leaf = heap [old] ;
    xold = x [leaf] ;
    while ( old > 1 )
    {
        new = old / 2 ;
        l = heap [new] ;
        xnew = x [l] ;
        if ( xnew > xold )
        {

/* swap new and old */

            ns [l] = old ;
            ns [leaf] = new ;
            heap [new] = leaf ;
            heap [old] = l ;
        }
        else if ( old == p ) break ;
        else                 return ;
        old = new ;
    }

/* Fix the heap below position p */

    pproj_minheap_ns (p, heap, ns, x, nheap) ;
}

/* ========================================================================== */
/* ======================== pproj_minheap_ns ================================ */
/* ========================================================================== */
/* heapify starting at node p.  On input, the heap at node p satisfies the */
/* heap property, except for heap [p] itself.  On output, the whole heap */
/* satisfies the heap property. Also, record pointers from nodes to store */
/* ========================================================================== */

void pproj_minheap_ns
(
    INT       p , /* start at node p in the heap */
    INT   *heap , /* size n, containing indices into x */
    INT     *ns , /* pointer from node to store */
    double   *x , /* not modified */
    INT   nheap   /* heap [1 ... nheap] is in use */
)
{
    INT left, right, e, hleft, hright ;
    double xleft, xright, xe ;

    e = heap [p] ;
    ns [e] = p ;
    xe = x [e] ;

    while ( 1 )
    {
        left = p * 2 ;
        right = left + 1 ;

        if (right <= nheap)
        {
            hleft  = heap [left] ;
            hright = heap [right] ;
            xleft  = x [hleft] ;
            xright = x [hright] ;
            if (xleft < xright)
            {
                if (xe > xleft)
                {
                    heap [p] = hleft ;
                    ns [hleft] = p ;
                    p = left ;
                }
                else
                {
                    heap [p] = e ;
                    ns [e] = p ;
                    return ;
                }
            }
            else
            {
                if (xe > xright)
                {
                    heap [p] = hright ;
                    ns [hright] = p ;
                    p = right ;
                }
                else
                {
                    heap [p] = e ;
                    ns [e] = p ;
                    return ;
                }
            }
        }
        else
        {
            if (left <= nheap)
            {
                hleft = heap [left] ;
                xleft = x [hleft] ;
                if (xe > xleft)
                {
                    heap [p] = hleft ;
                    ns [hleft] = p ;
                    p = left ;
                }
            }
            heap [p] = e ;
            ns [e] = p ;
            return ;
        }
    }
}

/* ========================================================================== */
/* ====== pproj_saxpy ======================================================= */
/* ========================================================================== */
/* Perform the operation x <- x + s*y */
/* ========================================================================== */

void pproj_saxpy
(
    double *x,
    double *y,
    double  s,
    INT     n  /* dimension of the vectors */
)
{
    INT n5, i ;
    double *X, *Y ;
    n5 = n % 5 ;
    X = x+n5 ;
    Y = y+n5 ;
    if ( s == ONE )
    {
        for (i = 0; i < n5; i++) x [i] += y [i] ;
        for (; i < n; i += 5)
        {
            *X++ += *Y++ ;
            *X++ += *Y++ ;
            *X++ += *Y++ ;
            *X++ += *Y++ ;
            *X++ += *Y++ ;
        }
    }
    else
    {
        for (i = 0; i < n5; i++) x [i] += s*y [i] ;
        for (; i < n; i += 5)
        {
            *X++ += s*(*Y++) ;
            *X++ += s*(*Y++) ;
            *X++ += s*(*Y++) ;
            *X++ += s*(*Y++) ;
            *X++ += s*(*Y++) ;
        }
    }
}

/* ========================================================================== */
/* ====== pproj_max ========================================================= */
/* ========================================================================== */
/* Evaluate the absolute maximum element in an array */
/* ========================================================================== */

double pproj_max
(
    double *x,
    INT     n  /* dimension of x */
)
{
    INT n5, i ;
    double t ;
    n5 = n % 5 ;
    t = ZERO ;
    for (i = 0; i < n5; i++)
    {
        if ( fabs (x [i]) > t ) t = fabs (x [i]) ;
    }
    for (; i < n; i += 5)
    {
        if ( fabs (x [i]) > t ) t = fabs (x [i]) ;
        if ( fabs (x [i+1]) > t ) t = fabs (x [i+1]) ;
        if ( fabs (x [i+2]) > t ) t = fabs (x [i+2]) ;
        if ( fabs (x [i+3]) > t ) t = fabs (x [i+3]) ;
        if ( fabs (x [i+4]) > t ) t = fabs (x [i+4]) ;
    }
    return (t) ;
}

/* ========================================================================== */
/* ====== pproj_timer ======================================================= */
/* ========================================================================== */
/* Returns the time in seconds used by the process.  */
/* Used in mexFunctions only (not a part of the C-callable UMFPACK) */
/* ========================================================================== */

double pproj_timer ( void )
{
#if 0
    struct rusage ru ;
    double user, tsys ;
    (void) getrusage (RUSAGE_SELF, &ru) ;
    user =
    ru.ru_utime.tv_sec                  /* user time (seconds) */
    + 1e-6 * ru.ru_utime.tv_usec ;      /* user time (microseconds) */

    tsys =
    ru.ru_stime.tv_sec                  /* tsys time (seconds) */
    + 1e-6 * ru.ru_stime.tv_usec ;      /* tsys time (microseconds) */

    return (user + tsys) ;
#endif

    struct timeval tv ;
    double walltime ;

    (void) gettimeofday (&tv, NULL) ;
    walltime = tv.tv_sec + (double) (tv.tv_usec)/1.e6 ;
    return (walltime) ;
}
