#ifndef _PPROJ_H_
#define _PPROJ_H_
#include <sys/resource.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include "SuiteSparse_config.h"
#include "cholmod.h"
#include "ccolamd.h"
#include "pproj_user.h"

#define MAXLINE 256

#ifndef PRINTF
#define PRINTF printf
#endif

/* #define PRIVATE static
   #define PUBLIC */
#define EMPTY -1

#define ZERO ((double) 0)
#define ONE ((double) 1)
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

#ifdef MATLAB_MEX_FILE
#define ASSERT(expression) (mxAssert ((expression), ""))
#else
#define ASSERT(expression) (assert (expression))
#endif

/* ANSI C99 has a clean definition of IEEE infinity, defined in <math.h>.
   MATLAB has this as well.  With the ANSI (C90) version of C, there is no
   well-defined infinity, so DBL_MAX is used instead.

   You can override these defaults and define your own version of infinity
   with (for example):

   cc -ansi -DNAPHEAP_INF=1e200 pproj.c ...

   To do this, edit the OPTFLAGS string in Lib/pproj.mk (editting this
   file is not recommended).
*/

#ifndef PPROJINF
#ifdef INFINITY
/* ANSI C99 (gcc -std=c99) */
#define PPROJINF INFINITY
#else
/* ANSI C90 (gcc -ansi) */
#define PPROJINF DBL_MAX
#endif
#endif

/* -------------------------------------------------------------------------- */
/* status returned by pproj */
/* -------------------------------------------------------------------------- */
#define PPROJ_SOLUTION_FOUND    (0)
#define PPROJ_TOLERANCE_NOT_MET (1)
#define PPROJ_INVALID_BOUNDS    (2)
#define PPROJ_OUT_OF_MEMORY     (3)
#define PPROJ_SSOR_NONASCENT    (4)
#define PPROJ_SSOR_MAX_ITS      (5)
/* the following are used for internal control and are not returned to user */
#define PPROJ_SWITCH_TO_UPDOWN  (6)
#define PPROJ_ALL_ROWS_DROPPED  (7)
#define PPROJ_PROX_UPDATE       (8)
#define PPROJ_STATUS_OK         (9)

/* ========================================================================== */
/* === PPprob =============================================================== */
/* ========================================================================== */

typedef struct PPprob_struct
{
    INT           nrow ; /* number of rows in A */
    INT           ncol ; /* number of cols in A */
    INT             ni ; /* number of i such that bl_i < bu_i */
    INT            *Ap ; /* size ncol+1, the column pointers */
    INT           *Anz ; /* If cholmod is used, then the row indices in each
                            column are reordered so that the active rows are
                            first, followed by the inactive rows. Anz [j]
                            is the number of active rows in column j */
    INT            *Ai ; /* size Ap [ncol], the row indices */
    double         *Ax ; /* size Ap [ncol], the numerical values in A */
    double          *y ; /* project y onto polyhedron */
    double          *b ; /* size nrow, right-hand-side */
    INT      *ineq_row ; /* size ni+2 (stopper in ineq_row [ni+1])
                            ineq_row [1], ... ineq_row [ni] are the row
                            numbers where bl_i < bu_i */
    double         *bl ; /* size ni+1, bl_i with bl_i < bu_i */
    double         *bu ; /* size ni+1, bu_i with bl_i < bu_i */

    double         *hi ; /* size ncol, Upper bounds for x */
    double         *lo ; /* size ncol, Lower bounds for x */
    INT         maxrow ; /* maximum number of nonzeros in a row including
                            any singleton. Used in coordinate ascent steps */
    INT         maxcol ; /* maximum number of nonzero in a column. Used in
                            initlevels */
    INT       *colperm ; /* if nonnull, permutation applied to columns of A */
    INT       *rowperm ; /* if nonnull, permutation applied to rows of A */
    double    grad_tol ; /* relative error tolerance for dual function grad */
} PPprob ;

/* ========================================================================== */
/* === PPwork =============================================================== */
/* ========================================================================== */

typedef struct PPwork_struct
{
    PPparm      *UParm ; /* user's Parameter input */
    PPstat      *UStat ; /* user's Statistics input */
    int      ssorquery ; /* TRUE if code should try ssor, FALSE = no ssor*/
    int  return_nochol ; /* TRUE if code returns from pproj_dasa before a chol*/
    int            fac ; /* 0 (not factored), 1 (factored, not updated) */
    int           Exit ; /* 1 (terminated in ssor1), 2 (terminated in dasa)*/
    int shiftl_is_zero ; /* TRUE when the proximal shift_l is zero */

    INT           ATnz ; /* # nonzeros in active rows of full matrix */
    INT           Annz ; /* # nonzeros in active rows and free columns of A */
    INT           Lnnz ; /* # nonzeros in chol (AF * AF') */

    cholmod_sparse  *A ; /* A matrix in cholmod sparse format */

    /* AT is the transpose of the A matrix, compact format */
    INT           *ATp ; /* size nrow + 1, the column pointers */
    INT           *ATi ; /* size Ap [ncol], the column indices */
    double        *ATx ; /* size Ap [ncol], the numerical values in A */

    /* AFT is the transpose of the AF matrix, the submatrix of A associated
       with the free variables */
    INT          *AFTp ; /* size nrow+1, the column pointers */
    INT         *AFTnz ; /* size nrow, number of nonzeros in each row */
    INT          *AFTi ; /* size Ap [ncol], the column indices */
    double       *AFTx ; /* size Ap [ncol], the numerical values in A */

    double          *x ; /* size: ncol */
    double    *shift_l ; /* The proximal shift in lambda. When a nonzero
                            starting guess is given, set shift_l = lambda */
    double       sigma ; /* current value for prox regularization parameter */
    double     *lambda ; /* size: nrow */
    double *lambda_tot ; /* size: nrow, lambda + shift_l */
    double    *dlambda ; /* size: nrow, the change in lambda */

    INT            *ns ; /* size: ntot, EMPTY always, used in line search */
    INT             *F ; /* size: ncol, free columns */
    int            *ib ; /* size: ntot, ib [j] defines the status of a column:
                            0 : column j is free
                            1 : column j is at the upper bound
                           -1 : column j is at the lower bound */
    INT            *ir ; /* size: nrow, ir [i] defines status of row i:
                            0  : equation i is an equality
                          > ni : equation i is strict inequality, currently
                                 treated as inactive (dropped). In the list
                                 of strict inequalities, it is
                                 # ineq_row [ir[i] - ni]
                          < 0  : equation i is a strict inequality, currently
                                 treated as active and at its lower bound.
                                 In the list of strict inequalities, it is
                                 # ineq_row [-ir [i]]
                          else : equation i is a strict inequality, currently
                                 treated as active and at its upper bound.
                                 In the list of strict inequalities, it is
                                 # ineq_row [ir [i]] */

    INT    *RowmodFlag ; /* size: ni+1, empty if inequality not modified,
                            otherwise points into RowmodList showing rows
                            to modify*/
    INT    *RowmodList ; /* size: ni, first part of this list contains the
                            inequalities to delete, the tail contains
                            inequalities to add */

    INT    *ColmodFlag ; /* size: ncol, same as row flag and list, but
                            delete and add indices are flipped */
    INT    *ColmodList ; /* size: ncol */

    double          *b ; /* size: nrow, shifted b vector for current level */
    double          *c ; /* size: ntot, shifted cost vector for curr. level */
    double       *cold ; /* size: ncol, c at start of the ascent iteration */
    double         *lo ; /* size: ncol, Lower bounds in translated problem */
    double         *hi ; /* size: ncol, Upper bounds in translated problem */
    double          *D ; /* size: nrow, diagonal of AF*AF' */

    INT        *arrayi ; /* INT work array (see initlevels) */
    double     *arrayd ; /* double work array (see initlevels) */
    INT       *RLinkUp ; /* size nrow+1 linked list for active equations */
    INT       *RLinkDn ; /* size nrow+1 linked list for active equations */
                         /* numbering of strict inequalities (also referred to
                            as column singletons) starts at 1 instead of 0 */
    INT       *SLinkUp ; /* size nrow+2 linked list for active singletons */
    INT       *SLinkDn ; /* size nrow+2 linked list for active singletons */
    int        ssormem ; /* number of vector in ssor memory */
    INT      ssor1_its ; /* stop ssor1 if no change in active set in ssor1_its*/

    /* ---------------------------------------------------------------------- */
    /* scalars */
    /* ---------------------------------------------------------------------- */

    INT        nrowadd ; /* number of rows to add to AF */
    INT        nrowdel ; /* number of rows to delete from AF */
    INT        ncoladd ; /* number of columns to add to AF */
    INT        ncoldel ; /* number of columns to delete from AF */
    INT        nactive ; /* number of active rows in upcoming L */
    INT             nf ; /* number of currently free variables (zeros in ib)*/
    int        do_coor ; /* T => perform coordinate ascent */
    int        do_ssor ; /* T => perform ssor ascent */
    INT       chg_coor ; /* # freed variables and dropped rows in coor_ascent */
    INT      chg_ssor0 ; /* # freed variables and dropped rows in ssor0 */
    INT      chg_ssor1 ; /* # freed variables and dropped rows in ssor1 */
    INT     chg_sparsa ; /* # variables or rows that change status in sparsa */
    double       normx ; /* 1-norm of x */
    double       absAx ; /* stores max_i sum_j |A_{ij}x_j| */
    double       grad0 ; /* SpaRSA monitors undecided set when grad <= grad0 */
    double       gamma ; /* SpaRSA stopping parameter */
    int       sparsaOK ; /* T => condition for using sparsa satisfied */

    /* ========== arrays and variable used in update/downdate ========== */
    cholmod_common *cmm; /* Common block for CHOLMOD */
    INT       *dropped ; /* list of rows dropped in modrow since last
                            updateAnz */

    /* C and L allocated in lpdasa_initupdate and lpdasa_initrun */
    cholmod_factor  *L ; /* Cholesky factor */

    double  *changeRHS ; /* size: nrow, right side change for forward
                            solve update */
    INT            *Cp ; /* size ncol, pointers into Ai or Ax of columns
                            to add or delete */
    INT           *Cnz ; /* size ncol, number of nonzeros in columns to
                            add or delete */
    INT       *colmark ; /* size: ncol, gives botrow for each column */

    /* ---------------------------------------------------------------------- */
    /* scalars */
    /* ---------------------------------------------------------------------- */

    int             nd ; /* total number of dropped rows since updataAnz */
    double        npup ; /* total number of partial updates/downdates */
    double        nrup ; /* total number of rooted updates/downdates */
    int       npup_old ; /* number of partial updates in previous ascent */
    int       npup_cur ; /* number of partial updates in current ascent */
    double    rupflops ; /* total number of flops in a rooted update */
    double    pupflops ; /* total number of flops in a partial update */
    int         nchols ; /* total number of chols */
    double   cholflops ; /* flops in last chol (excl AA') */
    double   cholaatflops ;/* total flops in chols (incl AA'
                              and incremental rowfac) */

    /* ==== arrays and variables used in multilevel implementation ==== */
    int           blks ; /* number of blocks in the multilevel partition */
    int       maxdepth ; /* number of levels in the partition */
    int          *Kids ; /* size: blks, list of children */
    int         *nkids ; /* size: blks, number of children of a node */
    int        *parent ; /* size: blks, defines tree in multilevel partition*/
    int            *Kp ; /* size: blks, points into array Kids. kids of node
                            are in list Kids [Kp [node] ... nkids [node]]*/
    int         *depth ; /* size: blks, depth of each node in tree */
    int      *leftdesc ; /* size: blks, for each block, left descendant */
    INT     *col_start ; /* size: blks+1, for each block, first column */
    INT     *sol_start ; /* size: blks+1, for each block, first singleton */
    INT     *row_start ; /* size: blks+1, for each block, 1st row in matrix */
    int    *col_to_blk ; /* size: ncol, for each col, 1st block in matrix */
    int    *sol_to_blk ; /* size: ni+2, for each singleton, 1st block in
                            matrix, set sol_to_blk [ni+1] = blks */

    /* lstart and ustart are associated with the inequalities that are
       strict.  For each block in the multilevel decomposition,
       lstart points to the first singleton at its lower bound while ustart
       points to the first singleton at it upper bound. Rstart and Rend
       contain the first and last active row of each block.
       Arrays joblist and kidsleft are used in the multilevel implementation.
       joblist contain the blocks that are solved to generate the current
       search direction. For each block in the list, we treat all the
       variables as being fixed except for the variables associated with
       the block. The variables in the blocks, with the variables
       outside these blocks fixed, are completely uncoupled in the
       dual problem, so we can perform a DASA step by solving for these
       variables in an uncoupled fashion. We start at the leaves of the
       tree and work up to the root. Once all the childen of a node
       are optimized, we move up to the parent. jobcols is the number
       of columns associated with this each block that have been added
       while jobrows is the number of rows in this block that have
       been deleted.  */
    INT        *lstart ; /* size blks + 1 */
    INT        *ustart ; /* size blks + 1 */

    int       *joblist ; /* size blks */
    int      *kidsleft ; /* size blks */
    INT       *jobcols ; /* size blks + 1 */
    INT       *jobrows ; /* size blks + 1 */
    INT        *Rstart ; /* size blks + 1 */
    INT          *Rend ; /* size blks + 1 */
} PPwork ;

/* ========================================================================== */
/* === PPcheck ============================================================== */
/* ========================================================================== */

typedef struct PPcheck_struct
{

    /* ---------------------------------------------------------------------- */
    /* sizes of allocated blocks */
    /* ---------------------------------------------------------------------- */

    double       prior ; /* prior objective value */
    double        mark ; /* current objective value (not used for comparisons)*/
    double        zmax ; /* max component of the x that minimizes Lagrangian */
    double        lmax ; /* max component of the current lambda */
    int       location ; /* 1 = phase1, 2 = dasa, 3 = ssor0, 4 = ssor1,
                            5 = check_error, 6 = sparsa */
} PPcheck ;

/* ========================================================================== */
/* === PPcom ================================================================ */
/* ========================================================================== */

typedef struct PPcom_struct
{
    PPprob   *Prob ; /* Problem */
    PPparm   *Parm ; /* Parameters */
    PPstat   *Stat ; /* Statistics */
    PPwork   *Work ; /* Working arrays and parameters that change */
    PPcheck *Check ; /* Check structure */
} PPcom ;

/* ========================================================================== */
/* === Prototypes of PPROJ routines ========================================= */
/* ========================================================================== */

#include "Check/check.h"

void pproj_cholmod_sparse
(
    cholmod_sparse *A, /* pointer to a cholmod sparse matrix */
    INT          nrow, /* number of rows */
    INT          ncol, /* number of columns */
    INT         nzmax, /* max number of nonzeros */
    INT            *p, /* size ncol + 1, the column pointers */
    INT           *nz, /* size ncol, # nonzeros in each column, could be NULL */
    INT            *i, /* size Ap [ncol], the row indices */
    double         *x, /* size Ap [ncol], the numerical values in A */
    int        sorted, /* TRUE if columns of A are sorted, FALSE otherwise */
    int        packed, /* TRUE if A is packed, FALSE otherwise */
    int         xtype  /* CHOLMOD_PATTERN or CHOLMOD_REAL */
) ;

void pproj_cholmod_dense
(
    cholmod_dense *A, /* pointer to a cholmod dense matrix */
    INT         nrow, /* number of rows */
    double        *x  /* size nrow, the numerical values in A */
) ;

void pproj_transpose
(
    INT    *Bp,   /* size nrow+1, column pointers (output) */
    INT    *Bi,   /* size Ap [ncol], row indices of B (output) */
    double *Bx,   /* size Ap [ncol], numerical entries of B (output) */
    INT    *Ap,   /* size ncol+1, column pointers */
    INT    *Ai,   /* size Ap [ncol], row indices for A */
    double *Ax,   /* size Ap [ncol], numerical entries of A */
    INT   nrow,   /* number of rows in A */
    INT   ncol,   /* number of cols in A */
    INT     *W    /* work array of size nrow */
) ;

void pproj_iminsort
(
    INT         *y,  /* n-by-1 (output) */
    INT         *x , /* n-by-1 (input not modified) */
    INT         *w,  /* n-by-1, (input, working array) */
    INT          n   /* number of elements to sort */
) ;

void pproj_xminsort
(
    INT    *y,  /* n-by-1 (output) */
    double *x , /* n-by-1 (input not modified) */
    INT    *w,  /* n-by-1, (input, working array) */
    INT     n   /* number of elements to sort */
) ;

void pproj_copyI
(
    INT *y, /* output of copy */
    INT *x, /* input of copy */
    INT  n  /* length of vectors */
) ;

void pproj_copyi
(
    int *y, /* output of copy */
    int *x, /* input of copy */
    INT  n  /* length of vectors */
) ;

void pproj_copyx
(
    double *y, /* output of copy */
    double *x, /* input of copy */
    INT     n  /* length of vectors */
) ;

void pproj_initI
(
    INT *x,  /* array to be initialized */
    INT  s,  /* scalar */
    INT  n   /* length of x */
) ;

void pproj_initi
(
    int *x,  /* array to be initialized */
    int  s,  /* scalar */
    INT  n   /* length of x */
) ;

void pproj_initx
(
    double *x,  /* array to be initialized */
    double  s,  /* scalar */
    INT     n   /* length of x */
) ;

void pproj_initFx
(
    double *x,  /* array to be initialized */
    double  s,  /* scalar */
    INT    *F,  /* indices to be initialized */
    INT     n   /* length of F */
) ;

void pproj_scale
(
    double *x,  /* array to be scaled */
    double *y,  /* array used for the scaling */
    double  s,  /* scale */
    INT     n   /* length of x */
) ;

double pproj_dot
(
    double *x,  
    double *y,  
    INT     n  /* length of x */
) ;

void pproj_minheap_build
(
    INT   *heap, /* on input, an unsorted set of element numbers */
    double   *x, /* the numerical values to be ordered */
    INT   nheap  /* number of elements to build into the heap */
) ;

void pproj_minheap_ify
(
    INT      p, /* start at node p in the heap */
    INT  *heap, /* size n, containing indices into x */
    double  *x, /* not modified */
    INT  nheap  /* heap [1 ... nheap] is in use */
) ;

void pproj_minheap_add
(
    INT     leaf , /* the new leaf */
    INT    *heap , /* size n, containing indices into x */
    INT      *ns , /* pointer from node to store in heap */
    double    *x , /* not modified */
    INT   *nheap   /* number of elements in heap including new one */
) ;

void pproj_minheap_delete
(
    INT  *heap , /* containing indices into x, 1..n on input */
    INT    *ns , /* pointer from node to store */
    double  *x , /* not modified */
    INT *nheap , /* number of items in heap */
    INT      p   /* element to delete from the heap */
) ;

void pproj_minheap_update
(
    INT   *heap , /* size n, containing indices into x */
    INT     *ns , /* pointer from node to store in heap */
    double   *x , /* not modified */
    INT   nheap , /* number of elements in the heap */
    INT       p   /* location of element to update */
) ;

void pproj_minheap_ns
(
    INT       p , /* start at node p in the heap */
    INT   *heap , /* size n, containing indices into x */
    INT     *ns , /* pointer from node to store */
    double   *x , /* not modified */
    INT   nheap   /* heap [1 ... nheap] is in use */
) ;

void pproj_saxpy
(
    double *x,
    double *y,
    double  s,
    INT     n  /* dimension of the vectors */
) ;

double pproj_max
(
    double *x,
    INT     n  /* dimension of x */
) ;

double pproj_timer ( void ) ;

void pproj_error
(
    int status,
    const char *file,
    int line,
    const char *message
) ;

void pproj_free
(
    PPcom **Ihandle
) ;

int pproj_check_error
(
    PPcom *I
) ;

int pproj_prox_update
(
    PPcom *I
) ;

void pproj_invert
(
    PPcom *I
) ;

int pproj_initwork
(
    PPcom *I
) ;

int pproj_initlevels
(
    PPcom *I
) ;

int pproj_phase1
(             
    PPcom *I
) ;

int pproj_hongchao
(
    PPcom *I
) ;

int pproj_sparsa
(
    PPcom *I
) ;

void pproj_coor_ascent
(          
    PPcom *I
) ;

void pproj_ssor0
(
    PPcom *I
) ;

int pproj_ssor1
(
    PPcom *I
) ;

void pproj_updown_dense
(
    double    *L, /* lower triangle of factorization (beneath diagonal) */
    double *diag, /* D in the LDL' factorization */
    double    *w, /* update/downdate matrix */
    int     info, /* +1 = update, -1 = downdate */
    int        n  /* rank of the matrix */
) ;

int pproj_dasa
(
    PPcom *I
) ;

cholmod_factor *pproj_rechol
(
    cholmod_sparse   *A,
    cholmod_sparse *AFT,    /* A(:,f)' */
    INT              *F,    /* the free set f = columns F [0..nf-1] */
    INT              nf,
    double     beta [2],
    INT        *RLinkUp,
    PPwork             *W,
    cholmod_common *cmm
) ;

void pproj_modrow
(
    PPcom          *I,
    int            nj, /* number of jobs in multilevel update,
                          ignored if no solve update */
    int MarkedForward, /* = TRUE (use botrow to update forward solve
                          = FALSE (botrow not needed in forward solve) */
    int   UpdateSolve, /* = TRUE (forward solve update)
                          = FALSE (only update factorization) */
    int         RowOp, /* = +1 (row is added)
                          = -1 (row is deleted) */
    double   *forward, /* forward solve */
    double *changeRHS, /* change in right hand side */
    double        *dl, /* change in lambda */
    INT   *updatework  /* nrow INT work space needed when UpdateSolve TRUE */
) ;

void pproj_modcol
(
    PPcom          *I,
    int MarkedForward, /* = 0 (botrow not needed forward solve)
                          = 1 (use botrow to update forward solve) */
    int   UpdateSolve, /* = 0 (only update factorization)
                          = 1 (forward solve update) */
    int         ColOp, /* = +1 (cols are added)
                          = -1 (cols are deleted) */
    INT      *Colmark, /* bottom row of block associated with added columns */
    INT           *Cp, /* pointers into Ai or Ax of columns to add or delete */
    INT          *Cnz, /* number of nonzeros in columns to add or delete */
    double   *forward, /* forward solve */
    double *changeRHS, /* change in right hand side */
    INT            nf  /* number of columns to add */
) ;

void pproj_lsol_partial
(
    cholmod_factor *L,
    double         *X,
    INT        jstart,
    INT        botrow,
    INT       *RLinkUp
) ;

void pproj_lsol
(
    cholmod_factor *L,
    double         *X,
    INT        jstart,
    INT        botrow,
    INT       *RLinkUp
) ;

void pproj_lsolve
(
    PPwork      *W,
    double    *X, /* X [0 .. (L->ncol)-1], right hand side, on input.
                           solution to Lx=b in locations jstart to botrow
                           on output */
    INT *RLinkUp,
    INT   jstart, /* either toprow or seprow */
    INT   botrow,
    int   update
) ;

void pproj_dltsol
(
    cholmod_factor *L,
    double *X,
    double *Xforward,
    INT iend,
    INT toprow,
    INT *RLinkDn
) ;

void pproj_dltsol_partial
(
    cholmod_factor *L,
    double *X,
    double *Xforward,
    INT iend,
    INT toprow,
    INT *RLinkDn,
    INT botrow
) ;

void pproj_dltsolve
(
    PPwork         *W,
    double       *X, /* X [0 .. (L->ncol)-1], right hand side, on input. */
    double *forward, /* solution to DL'x=b, on output */
    INT    *RLinkDn,
    INT      toprow,
    INT        iend,
    INT      update,
    INT      botrow
) ;

void pproj_lsolve0
(
    PPwork       *W,
    double     *X, /* X [0 .. (L->ncol)-1], right hand side, on input.
                      solution to Lx=b in locations jstart to botrow
                      on output */
    INT  *RLinkUp,
    INT  *RLinkDn,
    INT   *Rstart, /* return the new Rstart for botblk */
    INT    jstart, /* first row in seperator block */
    INT    toprow,
    INT    seprow,
    INT    botrow,
    int    update
) ;

double pproj_upspeed
(
    double fl, /* flops to perform a chol (excludes AF*AF') */
    double lnz /* nonzeros in cholesky factor */
) ;

int pproj_cholquery
(
    PPcom *I
) ;

void pproj_iterquery
(
    PPcom *I
) ;

void pproj_updateAnz
(
    PPcom     *I,
    int location    /* = 0 means code called at the top of ascent
                           before adding the new rows to L
                        = 1 means at the factorization at the 
                            top of ascent
                        = 2 means inside the ascent iteration, it
                            was cheaper to refactor than update */
) ;

void pproj_rowadd_prep
(
    INT               k, /* number of the row to add */
    PPcom            *I,
    cholmod_sparse   *A, /* the matrix */
    INT            *ati, /* column numbers in row */
    double         *atx, /* nonzeros in row */
    INT              nz, /* number of nonzeros */
    cholmod_sparse   *R, /* new row of AF*AF' */
    INT           ndead, /* number of currently dead rows still to be added */
    INT      *dead_rows, /* row numbers of currently dead rows */
    cholmod_common *cmm
) ;

#endif
