/* =========================================================================
   ============================== pproj_coor_ascent ========================
   =========================================================================
    Apply one iteration of coordinate ascent to the dual function.
    This is mainly used to handle the situation where all the variables
    associated with nonzeros in a row of A are bound. In this case, the
    corresponding row of AF*AF' vanishes and the dual function depends
    linearly on the multiplier associated with this row. The code is
    very similar to the dual coordinate ascent scheme appearing in the
    top half of phase1. The main differences are the following:

        1. In coor_ascent, we determine the number of active rows in A
           (this is used to decide whether to perform ssor0 or ssor1).
        2. In coor_ascent, we save the rowmod and colmod list for the
           update/downdate routines.
        3. In coor_ascent, the bounds are always 0 due to the x-translation
           described in the code pproj.c
        4. In coor_ascent, we only evaluate elements of D or
           AFT for active rows.
   ========================================================================== */
#include "pproj.h"

void pproj_coor_ascent
(
    PPcom *I
)
{
    int     blk, debug, PrintLevel, *sol_to_blk, *ib ;
    INT     ir_row, i, j, k, l, kk, p, p1, q, row, nrow, ncol, ni,
            ncoladd, ncoldel, nrowadd, nrowdel, coor_ascent_drop,
            maxrow, brk, nbrk, nf, ineqindex, Rl,
           *Ai, *Ap, *ATp, *ATi, *AFTi, *AFTnz, *AFTp,
           *Br_index, *Br_order, *F, *RLinkDn, *RLinkUp,
           *uLinkUp, *uLinkDn, *lLinkUp, *lLinkDn, *lstart, *ustart,
           *ir, *ns, *ineq_row, *worki, *iwork,
           *RowmodFlag, *RowmodList, *ColmodFlag, *ColmodList ;

    double  Br, ax, sigma,
            s, t, st, st0, fd, fd0, fn, sd, tic,
           *Ax, *ATx, *AFTx, *b, *bl, *bu, *c, *hi, *lo, *lambda,
           *shift_l, *Br_value, *D, *workd ;

    PPparm *Parm ;
    PPprob *Prob ;
    PPstat *Stat ;
    PPwork    *W ;

    tic = pproj_timer () ;

    /* extract the problem, statistics, and work structures from I */
    Parm = I->Parm ;
    Prob = I->Prob ;
    Stat = I->Stat ;
    W = I->Work ;
    PrintLevel = Parm->PrintLevel ;
    sigma = W->sigma ;
    debug = Parm->debug ;

    /* Problem matrix */
    Ap   = Prob->Ap ;
    Ai   = Prob->Ai ;
    Ax   = Prob->Ax ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    ineq_row = Prob->ineq_row ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    lo = W->lo ;
    hi = W->hi ;

    /* Transpose of A */
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;

    /* Transpose of AF */
    AFTp = W->AFTp ;
    AFTnz = W->AFTnz ;
    AFTi = W->AFTi ;
    AFTx = W->AFTx ;

    /* Links point to active rows */
    RLinkUp = W->RLinkUp ;
    RLinkDn = W->RLinkDn ;

    /* Links point to active singletons. lLinkUp points to the strict
       inequalities with b_i = bl_i while uLinkUp points to the strict
       inequalities with b_i = bu_i. Since only one of these can hold for
       each i, lLinkUp and uLinkUp can be stored in the same array */
    lLinkUp = W->SLinkUp ; 
    lLinkDn = W->SLinkDn ;
    uLinkUp = W->SLinkUp ;
    uLinkDn = W->SLinkDn ;

    sol_to_blk = W->sol_to_blk ; /* block associated with each inequality */
    lstart = W->lstart ;
    ustart = W->ustart ;

    /* working arrays */
    lambda = W->lambda ;
    shift_l = W->shift_l ;
    ib = W->ib ;
    ir = W->ir ;
    b = W->b ; /* the part of grad L (lambda) associated with bound variables */
    c = W->c ; /* y + A'lambda */
    maxrow = Prob->maxrow ; /* max number of nonzeros in a row of A */
    F = W->F ;
    nf = W->nf ;
    D = W->D ;
    ns = W->ns ;
   
    /* ======== start allocations ========
    -> work arrays in coordinate ascent step
 
       Br_index  - INT maxrow
       Br_order  - INT maxrow
       iwork     - INT maxrow
       Br_value  - double maxrow (maxrow <= ncol) */

    /* coordinate ascent step before ssor step */
    worki = W->arrayi ;
    Br_index  = worki ; worki += maxrow ;
    Br_order  = worki ; worki += maxrow ;
    iwork     = worki ; worki += maxrow ;

    workd = W->arrayd ;
    Br_value = workd ; workd += maxrow ;
    /* ======== end of allocations ======== */

    /* some parameters */
    ncoladd = W->ncoladd ;
    ncoldel = W->ncoldel ;
    nrowadd = W->nrowadd ;
    nrowdel = W->nrowdel ;
    RowmodFlag = W->RowmodFlag ;
    RowmodList = W->RowmodList ;
    ColmodFlag = W->ColmodFlag ;
    ColmodList = W->ColmodList ;
    coor_ascent_drop = Stat->coor_ascent_drop ;

    for (row = RLinkUp [nrow]; row < nrow; row = RLinkUp [row])
    {
        fd = b [row] ; /* b includes the sigma*lambda term */
        ASSERT (ir [row] <= ni) ;
        ir_row = ir [row] ;

        p = AFTp [row] ;
        p1 = p + AFTnz [row] ;
        for (; p < p1; p++)
        {
            fd -= c [AFTi [p]]*AFTx [p] ;
        }

        /* -------------------------------------------------------------- */
        /* Evaluate break points as we search in direction of + deriv. */
        /* -------------------------------------------------------------- */

        nbrk = 0 ;
        fd0 = fd ;
        sd = D [row] ;
        st = fd/sd ;
        ineqindex = 0 ;

        fd = fabs (fd) ;

        if ( fd0 > ZERO ) /* lambda [row] increases */
        {
            if ( ir_row > 0 ) /* row could drop when lambda increases */
            {
                if ( W->shiftl_is_zero )
                {
                   t = -lambda [row] ;
                }
                else
                {
                   t = -(lambda [row] + shift_l [row]) ;
                }
                if ( t < st )
                {
                    st = t ;
                    ineqindex = ir_row ;
                }
            }
            st0 = st ;
 
            /* find all break points < st */
            q = ATp [row+1] ;
            for (p = ATp [row] ; p < q; p++)
            {
                j = ATi [p] ;
                if ( ib [j] )
                {
                    ax = ATx [p] ;
                    if ( ax  > ZERO )
                    {
                        if ( ib [j] < 0 )
                        {
                            t = -c [j]/ax ;
                            if ( t < st )
                            {
                                Br_value [nbrk] = t ;
                                Br_index [nbrk] = p ;
                                nbrk++ ;
                            }
                        }
                    }
                    else
                    {
                        if ( ib [j] > 0 )
                        {
                            t = -c [j]/ax ;
                            if ( t < st )
                            {
                                Br_value [nbrk] = t ;
                                Br_index [nbrk] = p ;
                                nbrk++ ;
                            }
                        }
                    }
                }
            }
        }
        else if ( fd0 < ZERO ) /* lambda [row] decreases */
        {
            if ( ir_row < 0 ) /* row could drop as lambda decreases */
            {
                if ( W->shiftl_is_zero )
                {
                   t = -lambda [row] ;
                }
                else
                {
                   t = -(lambda [row] + shift_l [row]) ;
                }
                if ( t > st )
                {
                    st = t ;
                    ineqindex = -ir_row ;
                }
            }
            st0 = st = -st ;
            q = ATp [row+1] ;
            for (p = ATp [row] ; p < q; p++)
            {
                j = ATi [p] ;
                if ( ib [j] )
                {
                    ax = ATx [p] ;
                    if ( ax  > ZERO )
                    {
                        if ( ib [j] > 0 )
                        {
                            t = c [j]/ax ;
                            if ( t < st )
                            {
                                Br_value [nbrk] = t ;
                                Br_index [nbrk] = p ;
                                nbrk++ ;
                            }
                        }
                    }
                    else
                    {
                        if ( ib [j] < 0 )
                        {
                            t = c [j]/ax ;
                            if ( t < st )
                            {
                                Br_value [nbrk] = t ;
                                Br_index [nbrk] = p ;
                                nbrk++ ;
                            }
                        }
                    }
                } /* bound variable, check for break */
            }     /* end of loop over the row */
        }         /* find breaks */

        if ( nbrk > 0 )
        {
            pproj_xminsort (Br_order, Br_value, iwork, nbrk) ;

#ifndef NDEBUG
            if ( debug > 1 )
            {
                check_order (Br_value, Br_order, nbrk,
                             "in coordinate ascent in DASA") ;
            }
#endif

            st = 0 ;
            for (brk = 0; brk < nbrk; brk++)
            {
                kk = Br_order [brk] ;
                Br = MAX (ZERO, Br_value [kk]) ;
                fn = fd - sd * (Br - st) ;
                p = Br_index [kk] ;
                s = ATx [p] ;
                j = ATi [p] ;

                if ( fn <= ZERO )
                {
                    if ( fd != fn )
                    {
                        st += (fd/(fd-fn))*(Br - st);
                        fd = ZERO ;
                    }
                    break ;
                }
                else
                {
                    fd = fn ;
                }

                /* x_j changes from bound to free */
                p1 = Ap [j+1] ;
                for (p = Ap [j]; p < p1; p++)
                {
                    i = Ai [p] ;
                    if ( ir [i] <= ni )
                    {
                        ax = Ax [p] ;
                        D [i] += ax*ax ;
                        l = AFTp [i] + AFTnz [i]++ ;
                        AFTx [l] = ax ;
                        AFTi [l] = j ;
                    }
                }
                sd += s*s ;

#ifndef NDEBUG
                if ( PrintLevel > 2 )
                {
                    if ( ib [j] > 0 )
                    {
                        PRINTF("row: %i free: %i hi: %e\n",
                             row, j, hi [j]);
                    }
                    else
                    {
                        PRINTF("row: %i free: %i lo: %e\n",
                             row, j, lo [j]);
                    }
                }
#endif

                ib [j] = 0 ;
                F [nf++] = j ;
                st = Br ;
                k = ColmodFlag [j] ;
                if ( k != EMPTY )
                {
                    l = ColmodList [ncol-ncoldel] ;
                    ColmodList [k] = l ;
                    ColmodFlag [l] = k ;
                    ColmodFlag [j] = EMPTY ;
                    ncoldel-- ;
                }
                else
                {
                    ColmodList [ncoladd] = j ;
                    ColmodFlag [j] = ncoladd ;
                    ncoladd++ ;
                }
            }

            st += fd/sd ;

            if ( st >= st0 )
            {
                st = st0 ;
            }
            else
            {
                ineqindex = 0 ; /* the row did not drop */
            }
        }

        if ( fd0 < 0 )
        {
            st = -st ;
        }

        if ( ineqindex ) /* drop row */
        {
            Stat->coor_ascent_drop++ ;
            k = RLinkDn [row] ;
            j = RLinkUp [k] = RLinkUp [row] ;
            RLinkDn [j] = k ;
            ASSERT (ir [row] <= ni) ;
            k = RowmodFlag [ineqindex] ;
            if ( k != EMPTY ) /* the row was set to be added, remove it */
            {
                AFTnz [row] = 0 ;
                /* move last added row to replace new dropped row */
                l = RowmodList [ni-nrowadd] ;
                RowmodList [k] = l ;
                RowmodFlag [l] = k ;
                RowmodFlag [ineqindex] = EMPTY ;
                nrowadd-- ; /* one less row to add */
            }
            else /* add row to the delete list */
            {
                RowmodList [nrowdel] = ineqindex ;
                RowmodFlag [ineqindex] = nrowdel ;
                nrowdel++ ;
            }

            if ( ir [row] > 0 ) /* remove from uLink */
            {
                k = uLinkDn [ineqindex] ;
                j = uLinkUp [ineqindex] ;
                uLinkUp [k] = j ;
                uLinkDn [j] = k ;
                b [row] -= bu [ineqindex] ;
                if ( Parm->cholmod == TRUE )
                {
                    blk = sol_to_blk [ineqindex] ;
                    if ( ineqindex == ustart [blk] )
                    {
                        ustart [blk] = j ;
                    }
                }
            }
            else                /* remove from lLink */
            {
                k = lLinkDn [ineqindex] ;
                j = lLinkUp [ineqindex] ;
                lLinkUp [k] = j ;
                lLinkDn [j] = k ;
                b [row] -= bl [ineqindex] ;
                if ( Parm->cholmod == TRUE )
                {
                    blk = sol_to_blk [ineqindex] ;
                    if ( ineqindex == lstart [blk] )
                    {
                        lstart [blk] = j ;
                    }
                }
            }
            ir [row] = ineqindex + ni ;  /* Drop row */
            if ( W->shiftl_is_zero )
            {
               lambda [row] = ZERO ;
            }
            else
            {
               lambda [row] = -shift_l [row] ;
            }

#ifndef NDEBUG
            if ( PrintLevel > 2 )
            {
                PRINTF ("dasa coordinate ascent drop row: %i\n", row);
            }
#endif
        }                /* end of row drop */
        else
        {
            lambda [row] += st ;
            Rl = row ;
        }
        if ( st != ZERO )
        {
            b [row] -= sigma*st ;
            q = ATp [row+1] ;
            for (p = ATp [row] ; p < q; p++)
            {
                c [ATi [p]] += ATx [p]*st ;
            }
        }
    } /* end of coordinate ascent iteration */
#ifndef NDEBUG
    check_dual (I, NULL, "end of dasa coordinate ascent", TRUE, TRUE) ;
#endif

    W->ncoladd = ncoladd ;
    W->ncoldel = ncoldel ;
    W->nrowadd = nrowadd ;
    W->nrowdel = nrowdel ;
    k = nf - W->nf ;
    W->nf = nf ;
    Stat->coor_ascent_free += k ;
    W->chg_coor = k + (Stat->coor_ascent_drop - coor_ascent_drop) ;
    Stat->coor_ascent_its++ ;
    Stat->coor_ascent += pproj_timer () - tic ;
}
