/* =========================================================================
   ============================== pproj_phase1 =============================
   =========================================================================
    A starting point is generated by first requiring that lambda lies in
    the domain of the dual function. We then alternate between a dual
    coordinate ascent iteration and an SSOR preconditioned gradient ascent
    iteration. Throughout phase1, the current estimate for lambda is the
    sum of the lambda variable in the code and shift_l. A detailed
    description for the SSOR preconditioned gradient ascent iteration is
    provided at the start of the ssor0 code.
   ========================================================================== */

#include "pproj.h"

int pproj_phase1
(
    PPcom *I
)
{
    int     debug, flag, ibj, PrintLevel, prevblk, *ib, *sol_to_blk ;
    INT     phase1_its, ir_row, i, j, k, first, last,
            l, m, kk, p, p0, p1, q, col, row, nrow, ncol, ni, ni1, ni2,
            maxrow, ntot, brk, nbrk, nup,
            it, nr, nf, nf_old, ineqflag, ineqindex, Rl, Ul, Ll,
           *Ai, *Ap, *ATp, *ATi, *AFTi, *AFTnz, *AFTp,
           *Br_index, *Br_order, *Heap, *F, *R, *RLinkDn, *RLinkUp,
           *uLinkUp, *uLinkDn, *lLinkUp, *lLinkDn,
           *ir, *ns, *ineq_row, *worki, *iwork ;

    double  Br, ax, cj, sigma, xbound, dk_row,
            s, t, u, st, st0, fd, fd0, fn, sd, dknew, snew, errdual,
           *Ax, *ATx, *AFTx, *b, *bl, *bu, *c, *hi, *lo, *lambda, *shift_l,
           *Br_value0, *Br_value1, *D, *dk, *pA, *gk, *y, *workd, *x, *absAx ;

    PPprob *Prob ;
    PPstat *Stat ;
    PPwork    *W ;
    PPparm *Parm ;

#ifndef NDEBUG
    char   *where ;
    I->Check->location = 1 ; /* code operates in phase1 */
#endif
    /* extract the problem, statistics, and work structures from I */
    Prob = I->Prob ;
    Stat = I->Stat ;
    W = I->Work ;
    Parm = I->Parm ;
    debug = Parm->debug ;
    PrintLevel = Parm->PrintLevel ;

    /* Problem matrix */
    Ap   = Prob->Ap ;
    Ai   = Prob->Ai ;
    Ax   = Prob->Ax ;
    y    = Prob->y ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    ni1 = ni + 1 ;
    ni2 = ni + 2 ;
    ntot = ni + ncol + 1 ;
    ineq_row = Prob->ineq_row ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    lo = Prob->lo ;
    hi = Prob->hi ;

    /* Transpose of A */
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;

    /* Transpose of AF */
    AFTp = W->AFTp ;
    AFTnz = W->AFTnz ;
    AFTi = W->AFTi ;
    AFTx = W->AFTx ;

    /* Links point to active rows */
    RLinkUp = W->RLinkUp ;
    RLinkDn = W->RLinkDn ;

    /* Links point to active singletons. lLinkUp points to the strict
       inequalities with b_i = bl_i while uLinkUp points to the strict
       inequalities with b_i = bu_i. Since only one of these can hold for
       each i, lLinkUp and uLinkUp can be stored in the same array */
    lLinkUp = W->SLinkUp ; 
    lLinkDn = W->SLinkDn ;
    uLinkUp = W->SLinkUp ;
    uLinkDn = W->SLinkDn ;

    /* working arrays */
    lambda = W->lambda ;
    shift_l = W->shift_l ;
    x = W->x ;
    ib = W->ib ;
    ir = W->ir ;
    b = W->b ; /* the part of grad L (lambda) associated with bound variables */
    c = W->c ; /* y + A'lambda */
    maxrow = Prob->maxrow ; /* max number of nonzeros in a row of A */
    F = W->F ;
    D = W->D ;
    ns = W->ns ;
    sigma = W->sigma ;

    /* initializations */
    /* ir [row] = 0         for an equality constraint
                =  strict # for an active inequality at upper bound
                = -strict # for an active inequality at lower bound
                =  strict # + ni for a dropped constraint */
    pproj_initI (ir, (INT) 0, nrow) ;
    pproj_initI (AFTnz, (INT) 0, nrow) ;
    pproj_copyx (c, Prob->y, ncol) ;
    pproj_initx (D, sigma, nrow) ;

    /* ======== start allocations ========
    -> work arrays in coordinate ascent step

       Br_index  - INT maxrow
       Br_order  - INT maxrow
       iwork     - INT maxrow
       Br_value0 - double maxrow (maxrow <= ncol)

    -> work arrays in ssor step:

       Heap      - INT    ntot, line search
       pA        - double ncol, A'lambda, share same space with Br_value0
       dk        - double nrow, search direction
       Br_value1 - double ncol + ni + 1
       gk        - double nrow, used in ssor iteration

    -> work arrays after ssor step in constraint updates:

       R         - INT    nrow, AFT bookkeeping after the line search
       pA        - double ncol */

    /* coordinate ascent step before ssor step */
    worki = W->arrayi ;
    Br_index  = worki ; worki += maxrow ;
    Br_order  = worki ; worki += maxrow ;
    iwork     = worki ; worki += maxrow ;

    workd = W->arrayd ;
    Br_value0 = workd ; workd += maxrow ;

    /* ssor step */
    worki = W->arrayi ;
    Heap      = worki ; worki += ntot ;

    workd = W->arrayd ;
    pA        = workd ; workd += ncol ; /* if pA is moved, tweek check_line */
    dk        = workd ; workd += nrow ;
    Br_value1 = workd ; workd += ncol + ni + 1 ;
    gk        = workd ; workd += nrow ; /* grad L (lambda) */

    /* constraint update */
    worki = W->arrayi ;
    R         = worki ; worki += nrow ;

    workd = W->arrayd ;
    pA        = workd ; workd += ncol ;

    /* ======== end of allocations ======== */

    /* ---------------------------------------------------------------------- */
    /* Setup for the dual coordinate ascent and SSOR iterations */
    /* 1. Check that starting guess is feasible.
       2. Set b.
       3. Set lLink and uLink (inequalities active at lower/upper bounds) */
    /* ---------------------------------------------------------------------- */
    Ll = ni1 ;
    Ul = ni2 ;
    Rl = nrow ;
    if ( Parm->use_lambda ) /* lambda = 0 and shift_l contains lambda */
    {
        W->shiftl_is_zero = FALSE ;
        for (i = 1; i <= ni; i++)
        {
            row = ineq_row [i] ;
            t = shift_l [row] ;
            if ( t > ZERO )
            {
                /* if lambda infeasible, set it to zero and drop the row */
                if ( bl [i] == -PPROJINF )
                {
                    shift_l [row] = ZERO ;
                }
                else
                {
                    b [row] = bl [i] ;
                    ir [row] = -i ;
                    lLinkDn [i] = Ll ;
                    lLinkUp [Ll] = i ;
                    Ll = i ;
                }
            }
            else if ( t < ZERO )
            {
                /* if lambda infeasible, set it to zero and drop the row */
                if ( bu [i] == PPROJINF )
                {
                    shift_l [row] = ZERO ;
                }
                else
                {
                    b [row] = bu [i] ;
                    ir [row] = i ;
                    uLinkDn [i] = Ul ;
                    uLinkUp [Ul] = i ;
                    Ul = i ;
                }
            }
            if ( shift_l [row] == ZERO )
            {
                ir [row] = i + ni ;
            }
        }
        for (i = 0; i < nrow; i++)
        {
            if ( ir [i] <= ni )
            {
                RLinkDn [i] = Rl ;
                RLinkUp [Rl] = i ;
                Rl = i ;
                t = shift_l [i] ;
                if ( t != ZERO )
                {
                    q = ATp [i+1] ;
                    for (p = ATp [i]; p < q; p++)
                    {
                        c [ATi [p]] += t*ATx [p] ;
                    }
                }
            }
        }
    }
    else  /* lambda = shift_l = 0, initially drop all the inequalities */
    {
        for (i = 1; i <= ni; i++)
        {
            ir [ineq_row [i]] = i + ni ;
        }
        for (i = 0; i < nrow; i++)
        {
            if ( ir [i] <= ni )
            {
                RLinkDn [i] = Rl ;
                RLinkUp [Rl] = i ;
                Rl = i ;
            }
        }
    }
    lLinkUp [Ll] = ni1 ;
    uLinkUp [Ul] = ni2 ;
    lLinkDn [ni1] = Ll ;
    uLinkDn [ni2] = Ul ;
    RLinkDn [nrow] = Rl ;
    RLinkUp [Rl] = nrow ;

    /* 1. evaluate b - A_B * x_B
       2. set ib [j] = +1 if x_j at upper bound
              ib [j] = -1 if x_j at lower bound
       3. evaluate AFT, diag of AF*AF', and F */
    nf = 0 ;
    q = 0 ;
    for (j = 0; j < ncol; j++)
    {
        p = q ;
        q = Ap [j+1] ;
        t = c [j] ;
        if ( t > hi [j] )
        {
            ib [j] = 1 ;
            t = hi [j] ;
#ifndef NDEBUG
            x [j] = t ;
#endif
            if ( t != ZERO )
            {
                for (; p < q; p++)
                {
                    b [Ai [p]] -= t*Ax [p] ;
                }
            }
        }
        else if ( t < lo [j] )
        {
            ib [j] =-1 ;
            t = lo [j] ;
#ifndef NDEBUG
            x [j] = t ;
#endif
            if ( t != ZERO )
            {
                for (; p < q; p++)
                {
                    b [Ai [p]] -= t*Ax [p] ;
                }
            }
        }
        else
        {
            ib [j] = 0 ;
            F [nf++] = j ;
            for (; p < q; p++)
            {
                i = Ai [p] ;
                k = AFTp [i] + AFTnz [i]++ ;
                AFTi [k] = j ;
                ax = Ax [p] ;
                AFTx [k] = ax ;
                D [i] += ax*ax ;
            }
        }
    }
    /* ---------------------------------------------------------------------- */
    /* Start the iterations */
    /* ---------------------------------------------------------------------- */

    if ( (Parm->phase1 >= 0) && (Parm->use_phase1 == TRUE) )
    {
        k = pow ((double) nrow, Parm->phase1) ;
        phase1_its = MAX (k, 5) ; /* number of iterations in phase 1 */
    }
    else
    {
        phase1_its = 0 ;
        /* all rows were initialized to be inactive, if we skip phase1,
           then we need to locate the active rows */
        Ll = ni1 ;
        Ul = ni2 ;
        Rl = nrow ;
        for (i = 0 ; i < nrow ; i++)
        {
            j = ir [i] ;
            if ( j == 0 ) /* equality row */
            {
                Rl = i ;
                if ( j < 0 )
                {
                    Ll = -j ;
                }
                else if ( j > 0 )
                {
                    Ul = j ;
                }
            }
            else          /* inequality row */
            {
                t = b [i] ;
                ASSERT (ir [i] > ni) ;
                p = AFTp [i] ;
                q = p + AFTnz [i] ;
                for (; p < q; p++)
                {
                    t -= c [AFTi [p]]*AFTx [p] ;
                }
                ineqindex = j - ni ;
                s = t + bl [ineqindex] ;
                u = sigma*lambda [i] ;
                /* Include the proximal term when deciding whether activating
                   the row will improve the objective. */
                if ( s - u  > ZERO ) /* lambda_i > 0 increases L */
                {
                    if ( PrintLevel > 1 )
                    {
                        printf ("add row: %i (lower bind)\n", i) ;
                    }
                    /* add row, lower bound */
                    l = RLinkUp [Rl]  ;
                    RLinkUp [i] = l ;
                    RLinkDn [l] = i ;
                    RLinkUp [Rl] = i ;
                    RLinkDn [i] = Rl ;
                    Rl = i ;
                    l = lLinkUp [Ll] ;
                    lLinkUp [ineqindex] = l ;
                    lLinkDn [l] = ineqindex ;
                    lLinkUp [Ll] = ineqindex ;
                    lLinkDn [ineqindex] = Ll ;
                    Ll = ineqindex ;
                    b [i] += bl [ineqindex] ;
                    ir [i] = -ineqindex ;     /* row is active at lower bound */
                }
                else
                {
                    s = t + bu [ineqindex] ;
                    if ( s - u < ZERO ) /* lambda_i < 0 improves L */
                    {
                        if ( PrintLevel > 1 )
                        {
                            printf ("add row: %i (upper bind)\n", i) ;
                        }
                        /* add row, upper bound */
                        l = RLinkUp [Rl]  ;
                        RLinkUp [i] = l ;
                        RLinkDn [l] = i ;
                        RLinkUp [Rl] = i ;
                        RLinkDn [i] = Rl ;
                        Rl = i ;
                        l = uLinkUp [Ul] ;
                        uLinkUp [ineqindex] = l ;
                        uLinkDn [l] = ineqindex ;
                        uLinkUp [Ul] = ineqindex ;
                        uLinkDn [ineqindex] = Ul ;
                        Ul = ineqindex ;
                        b [i] += bu [ineqindex] ;
                        ir [i] = ineqindex ;  /* row is active at upper bound */
                    }
                }
            }
        }
    }
#ifndef NDEBUG
    where = "at end of initialization in phase 1" ;
    checkc (I, where) ;
    /* save dual objective but do not check for increase */
    check_dual (I, NULL, where, TRUE, FALSE) ;
#endif

    for (it = 0; it < phase1_its ; )
    {
        /* printf ("================ it %d\n", it) ; */
        nf_old = nf ;

        /* ------------------------------------------------------------------ */
        /* for each active row, do: */
        /* ------------------------------------------------------------------ */

#ifndef NDEBUG
        checkb (I, "start of phase1 iteration") ;
#endif
        for (row = RLinkUp [nrow]; row < nrow; row = RLinkUp [row])
        {
            fd = b [row] - sigma*lambda [row] ;
            ASSERT (ir [row] <= ni) ;
            ir_row = ir [row] ;

            p = AFTp [row] ;
            p1 = p + AFTnz [row] ;
            for (; p < p1; p++)
            {
                fd -= c [AFTi [p]]*AFTx [p] ;
            }

            /* -------------------------------------------------------------- */
            /* Evaluate break points as we search in direction of + deriv. */
            /* -------------------------------------------------------------- */

            nbrk = 0 ;
            fd0 = fd ;
            sd = D [row] ;
            st = fd/sd ;
            ineqflag = 0 ;
            fd = fabs (fd) ;
            if ( PrintLevel > 2 )
            {
                printf ("active row: %i initial st: %e fd0: %e\n",
                         row, st, fd0) ;
            }

            if ( fd0 > ZERO ) /* lambda [row] increases */
            {
                if ( ir_row <= ni )/* row currently active */
                {
                    if ( ir_row > 0 ) /* row could drop when lambda increases */
                    {
                        if ( W->shiftl_is_zero )
                        {
                            t = -lambda [row] ;
                        }
                        else
                        {
                            t = -(lambda [row] + shift_l [row]) ;
                        }
                        if ( t < st )
                        {
                            st = t ;
                            ineqflag = ir_row ;
                        }
                    }
                }
                st0 = st ;

                q = ATp [row+1] ;
                for (p = ATp [row] ; p < q; p++)
                {
                    j = ATi [p] ;
                    if ( ib [j] )
                    {
                        ax = ATx [p] ;
                        if ( ax  > ZERO )
                        {
                            if ( ib [j] < 0 )
                            {
                                t = (lo [j] - c [j])/ax ;
                                if ( t < st )
                                {
                                    Br_value0 [nbrk] = t ;
                                    Br_index [nbrk] = p ;
                                    nbrk++ ;
                                }
                            }
                        }
                        else
                        {
                            if ( ib [j] > 0 )
                            {
                                t = (hi [j] - c [j])/ax ;
                                if ( t < st )
                                {
                                    Br_value0 [nbrk] = t ;
                                    Br_index [nbrk] = p ;
                                    nbrk++ ;
                                }
                            }
                        }
                    }
                }
            }
            else if ( fd0 < 0 ) /* lambda [row] decreases */
            {
                if ( ir_row <= ni )/* row currently active */
                {
                    if ( ir_row < 0 ) /* row could drop as lambda decreases */
                    {
                        if ( W->shiftl_is_zero )
                        {
                            t = -lambda [row] ;
                        }
                        else
                        {
                            t = -(lambda [row] + shift_l [row]) ;
                        }
                        if ( t > st )
                        {
                            st = t ;
                            ineqflag = -ir_row ;
                        }
                    }
                }
                st0 = st = -st ;
                q = ATp [row+1] ;
                for (p = ATp [row] ; p < q; p++)
                {
                    j = ATi [p] ;
                    if ( ib [j] )
                    {
                        ax = ATx [p] ;
                        if ( ax  > ZERO )
                        {
                            if ( ib [j] > 0 )
                            {
                                t = (c [j] - hi [j])/ax ;
                                if ( t < st )
                                {
                                    Br_value0 [nbrk] = t ;
                                    Br_index [nbrk] = p ;
                                    nbrk++ ;
                                }
                            }
                        }
                        else
                        {
                            if ( ib [j] < 0 )
                            {
                                t = (c [j] - lo [j])/ax ;
                                if ( t < st )
                                {
                                    Br_value0 [nbrk] = t ;
                                    Br_index [nbrk] = p ;
                                    nbrk++ ;
                                }
                            }
                        }
                    } /* bound variable, check for break */
                }     /* end of loop over the row */
            }         /* find breaks */

            if ( nbrk > 0 )
            {
                pproj_xminsort (Br_order, Br_value0, iwork, nbrk) ;

#ifndef NDEBUG
                if ( debug > 1 )
                {
                    check_order (Br_value0, Br_order, nbrk,
                                 "phase1, coordinate ascent") ;
                }
#endif

                st = 0 ;
                for (brk = 0; brk < nbrk; brk++)
                {
                    kk = Br_order [brk] ;
                    Br = MAX (ZERO, Br_value0 [kk]) ;
                    fn = fd - sd * (Br - st) ;
                    p = Br_index [kk] ;
                    s = ATx [p] ;
                    j = ATi [p] ;

                    if ( fn <= ZERO )
                    {
                        if ( fd != fn )
                        {
                            st += (fd/(fd-fn))*(Br - st);
                            fd = ZERO ;
                        }
                        break ;
                    }
                    else
                    {
                        fd = fn ;
                    }

                    /* x_j changes from bound to free */
#ifndef NDEBUG
                    x [j] = ZERO ;
#endif
                    if ( ib [j] > 0 ) t = hi [j] ;
                    else              t = lo [j] ;
                    p1 = Ap [j+1] ;
                    for (p = Ap [j]; p < p1; p++)
                    {
                        i = Ai [p] ;
                        b [i] += t*(ax = Ax [p]) ;
                        D [i] += ax*ax ;
                        l = AFTp [i] + AFTnz [i]++ ;
                        AFTx [l] = ax ;
                        AFTi [l] = j ;
                    }
                    sd += s*s ;

#ifndef NDEBUG
                    if ( PrintLevel > 2 )
                    {
                        if ( ib [j] > 0 )
                        {
                            PRINTF("row: %i free: %i hi: %e\n",
                                 row, j, hi [j]);
                        }
                        else
                        {
                            PRINTF("row: %i free: %i lo: %e\n",
                                 row, j, lo [j]);
                        }
                    }
#endif

                    ib [j] = 0 ;
                    F [nf++] = j ;
                    st = Br ;
                }

                st += fd/sd ;

                if ( st >= st0 )
                {
                    st = st0 ;
                }
                else
                {
                    ineqflag = 0 ; /* the row did not drop */
                }
            }

            if ( fd0 < 0 )
            {
                st = -st ;
            }

            if ( ineqflag )
            {
                lambda [row] = -shift_l [row] ;
                /* drop row */
                k = RLinkDn [row] ;
                j = RLinkUp [k] = RLinkUp [row] ;
                RLinkDn [j] = k ;
                ASSERT (ir [row] <= ni) ;
                if ( ir [row] > 0 ) /* remove from uLink */
                {
                    k = uLinkDn [ineqflag] ;
                    j = uLinkUp [k] = uLinkUp [ineqflag] ;
                    uLinkDn [j] = k ;
                    b [row] -= bu [ineqflag] ;
                }
                else                /* remove from lLink */
                {
                    k = lLinkDn [ineqflag] ;
                    j = lLinkUp [k] = lLinkUp [ineqflag] ;
                    lLinkDn [j] = k ;
                    b [row] -= bl [ineqflag] ;
                }
                ir [row] = ineqflag + ni ;  /* Drop row */

#ifndef NDEBUG
                if ( PrintLevel > 2 )
                {
                    PRINTF ("phase1 coordinate ascent drop row: %i\n", row);
                }
#endif
            }                /* end of row drop */
            else
            {
                Rl = row ;
                lambda [row] += st ;
            }
            if ( st != ZERO )
            {
                q = ATp [row+1] ;
                for (p = ATp [row] ; p < q; p++)
                {
                    c [ATi [p]] += ATx [p] * st ;
                }
            }
            if ( PrintLevel > 2 )
            {
                printf ("row: %i final st: %e\n", row, st) ;
            }

#ifndef NDEBUG
            if ( debug > 1 )
            {
                where = "at end of coordinate ascent iteration in phase 1" ;
                check_link (I, (int *) NULL, 0, where) ;
                /* check that dual objective increases */
                check_dual (I, NULL, where, TRUE, TRUE) ;
                checkb (I, where) ;
                checkc (I, where) ;
            }
#endif

        } /* end of coordinate ascent iteration */

        /* ------------------------------------------------------------------ */
        /* right side for ssor step */
        /* ------------------------------------------------------------------ */

        first = RLinkUp [nrow] ;
        /* b = part of grad L (lambda) associated with bound variables.
           gk = grad L (lambda) for active rows + proximal term */

        if ( first < nrow )
        {
            for (i = first; i < nrow; i = RLinkUp [i])
            {
                t = b [i] ;
                ASSERT (ir [i] <= ni) ;
                p = AFTp [i] ;
                q = p + AFTnz [i] ;
                for ( ; p < q ; p++)
                {
                    t -= AFTx [p]*c [AFTi [p]] ;
                }
                gk [i] = t - sigma*lambda [i] ;
            }

            /* -------------------------------------------------------------- */
            /* SSOR step */
            /* -------------------------------------------------------------- */

            for (k = 0; k < nf; k++)
            {
                pA [F [k]] = ZERO ;
            }

            i = first ;
            t = dk [i] = gk [i]/D [i] ;

            last = RLinkDn [nrow] ;

            while (i < last)
            {

                ASSERT (ir [i] <= ni) ;

                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for ( ; p < p1 ; p++)
                {
                    pA [AFTi [p]] += AFTx [p]*t ;
                }

                i = RLinkUp [i] ;

                ASSERT (ir [i] <= ni) ;

                t = 0 ;

                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for ( ; p < p1 ; p++)
                {
                    t += AFTx [p]*pA [AFTi [p]] ;
                }
                t = dk [i] = (gk [i] - t)/D [i] ;
            }

            pproj_initx (pA, ZERO, ncol) ;
            fd = t*gk [i] ;
            sd = t*t ;

            for (; i > first;)
            {
                p1 = ATp [i+1] ;
                for (p = ATp [i] ; p < p1 ; p++)
                {
                    pA [ATi [p]] += ATx [p]*t ;
                }

                i = RLinkDn [i] ;

                ASSERT (ir [i] <= ni) ;

                t = ZERO ;

                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for ( ; p < p1 ; p++)
                {
                    t += AFTx [p]*pA [AFTi [p]] ;
                }
                t = dk [i] = dk [i] - t/D [i] ;
                fd += t*gk [i] ;
                sd += t*t ;
            }

            p1 = ATp [i+1] ;
            for (p = ATp [i] ; p < p1 ; p++)
            {
                pA [ATi [p]] += ATx [p]*t ;
            }

            sd *= sigma ;
            for (k = 0; k < nf; k++)
            {
                j = F [k] ;
                sd += pA [j]*pA [j] ;
            }

            if ( sd > ZERO )
            {
                st0 = st = fd/sd ;
            }
            else
            {
                st = ZERO ;
            }
        }
        else /* all equations have been dropped */
        {
            st = ZERO ;
            if ( it == 0 )
            {
                pproj_initx (pA, ZERO, ncol) ;
            }
        }
        if ( PrintLevel > 1 )
        {
            printf ("initial ssor st: %25.15e\n", st) ;
        }

        nup = 0 ;
        if ( st != ZERO )
        {
            nbrk = 0 ;
            for (j = 0; j < ncol; j++)
            {
                ibj = ib [j] ;
                if ( ibj == 0 )
                {
                    continue ;
                }
                else if ( ibj < 0 )
                {
                    if ( pA [j] > 0 )
                    {
                        s = (lo [j] - c [j])/pA [j] ;
                        if ( s < st )
                        {
                            Br_value1 [j] = s ;
                            nbrk++ ;
                            Heap [nbrk] = j ;        /* add to heap  */
                        }
                    }
                }
                else
                {
                    if ( pA [j] < 0 )
                    {
                        s = (hi [j] - c [j])/pA [j] ;
                        if ( s < st )
                        {
                            Br_value1 [j] = s ;
                            nbrk++ ;
                            Heap [nbrk] = j ;        /* add to heap */
                        }
                    }
                }
            }

            if ( W->shiftl_is_zero )
            {
                for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dk [row]) < ZERO )
                    {
                        /* lambda > 0 for lLinkUp */
                        s = -lambda [row]/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value1 [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
                for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dk [row]) > ZERO )
                    {
                        /* lambda < 0 for uLinkUp */
                        s = -lambda [row]/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value1 [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
            }
            else /* shift_l != 0 */
            {
                for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dk [row]) < ZERO )
                    {
                        /* lambda > 0 for lLinkUp */
                        s = -(lambda [row]+shift_l [row])/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value1 [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
                for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dk [row]) > ZERO )
                    {
                        /* lambda < 0 for uLinkUp */
                        s = -(lambda [row]+shift_l [row])/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value1 [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
            }

            /* ============================================================== */
            /* sort break points, do a line search */
            /* ============================================================== */

            flag = 0 ;
            if ( nbrk > 0 )
            {
                pproj_minheap_build (Heap, Br_value1, nbrk) ;
                for (k = 1; k <= nbrk; k++)
                {
                    ns [Heap [k]] = k ;
                }

#ifndef NDEBUG
                check_minheap (Heap, Br_value1, ns, nbrk, ntot,
                    "build Heap, phase 1, ssor step") ;
#endif

                st = ZERO ;
                while ( nbrk > 0 )
                {
                    nup++ ;
                    if ( sd > ZERO )
                    {
                        col = Heap [1] ;
                        pproj_minheap_delete (Heap, ns, Br_value1, &nbrk, 1) ;
                        Br = MAX (Br_value1 [col], ZERO) ;
                        fn = fd - sd * (Br - st) ;

#ifndef NDEBUG
                        check_minheap (Heap, Br_value1, ns, nbrk, ntot,
                                "Heap delete, phase 1") ;
                        if ( PrintLevel > 2 )
                        {
                            PRINTF("    brk: %i col: %i fn: %9.3e fd: %9.3e"
                                   " sd: %9.3e st: %9.3e Br_value: %9.3e\n",
                                nbrk, col, fn, fd, sd, st, Br) ;
                        }
#endif

                        if ( fn <= 0 )
                        {
                            if ( fd != fn )
                            {
                                st += (fd/(fd-fn))*(Br - st);
                            }
                            fd = 0 ;
                            break ;
                        }
                        else
                        {
                            fd = fn ;
                        }
                    }
                    else
                    {
                        break ;
                    }

                    st = Br ;

                    if ( col < ncol ) /* free column */
                    {
#ifndef NDEBUG
                        x [col] = ZERO ;
#endif
                        s = pA [col] ;
                        sd += s*s ;
                        F [nf++] = col ;
                        if ( ib [col] > 0 ) t = hi [col] ;
                        else                t = lo [col] ;
                        q = Ap [col+1] ;
                        for (p = Ap [col]; p < q; p++)
                        {
                            i = Ai [p] ;
                            ax = Ax [p] ;
                            b [i] += t*ax ;
                            D [i] += ax*ax ;
                            l = AFTp [i] + AFTnz [i]++ ;
                            AFTx [l] = ax ;
                            AFTi [l] = col ;
                        }
                        ib [col] = 0 ;
                    }
                    else              /* drop row */
                    {
                        ineqindex = col - ncol ;
                        row = ineq_row [ineqindex] ;

#ifndef NDEBUG
                        if ( PrintLevel > 2 )
                        {
                            PRINTF("    jsing: %i row: %i ir: %i\n",
                                    col - ncol, row, ir [row]) ;
                        }
                        if ( ir [row] > ni )
                        {
                            PRINTF ("row: % ir: %i was already deleted "
                                    "in phase 1\n", row, ir [row]) ;
                            pproj_error (-1, __FILE__, __LINE__, "stop") ;
                        }
#endif

                        /* previously bound was subtracted from b, now restore*/
                        ASSERT ((ir [row] != 0) && ( ir [row] <= ni)) ;
                        /* row dropped due to singleton ineqindex */ 
                        ir [row] = ineqindex + ni ;
                        i = RLinkDn [row] ;
                        k = RLinkUp [row] ;
                        RLinkUp [i] = k ;
                        RLinkDn [k] = i ;
                        if ( W->shiftl_is_zero )
                        {
                            dknew = -lambda [row] ;
                            lambda [row] = ZERO ;
                        }
                        else
                        {
                            dknew = -(lambda [row]+shift_l [row]) ;
                            lambda [row] = -shift_l [row] ;
                        }
                        dk_row = dk [row] ;
                        fd -= dk_row*b [row] ;
                        /* fd += dk_row*(Dsigma*dknew-b [row]) ;*/
                        sd -= sigma*dk_row*dk_row ;
                        if ( dk [row] > ZERO ) /* drop upper bound bu */
                        {
                            b [row] -= bu [ineqindex] ;
                            m = uLinkUp [ineqindex] ;
                            l = uLinkDn [ineqindex] ;
                            uLinkUp [l] = m ;
                            uLinkDn [m] = l ;
                        }
                        else /* drop lower bound bl */
                        {
                            b [row] -= bl [ineqindex] ;
                            m = lLinkUp [ineqindex] ;
                            l = lLinkDn [ineqindex] ;
                            lLinkUp [l] = m ;
                            lLinkDn [m] = l ;
                        }

                        q = ATp [row+1] ;
                        for (p = ATp [row] ; p < q ; p++)
                        {
                            j = ATi [p] ;
                            ax = ATx [p] ;
                            s = pA [j] ;
                            /* cj = c after taking the step */
                            cj = c [j] + st*s ;
                            /* modify c to account for later update */
                            c [j] += dknew*ax ;
                            t = dk_row*ax ;
                            snew = pA [j] = s - t ;
                            /* update c, pA, sd, fd */
                            if ( !ib [j] )
                            {
                                fd += cj*t ;
                                sd -= t*(s + snew) ;
                            }
                            else
                            {
                                if ( ib [j] < 0 )
                                {
                                    if ( snew > ZERO )
                                    {
                                        Br_value1 [j] = 
                                            st + (lo [j]-cj)/snew ;
                                        if ( ns [j] != EMPTY )
                                        {
                                            pproj_minheap_update (Heap,
                                               ns, Br_value1, nbrk, ns [j]) ;
                                        }
                                        else
                                        {
                                           pproj_minheap_add (j, Heap,
                                               ns, Br_value1, &nbrk) ;
                                        }
                                    }
                                    else
                                    {
                                        if ( ns [j] != EMPTY )
                                        {
                                            pproj_minheap_delete (Heap,
                                               ns, Br_value1, &nbrk, ns [j]);
                                        }
                                    }
                                }
                                else
                                {
                                    if ( snew < ZERO )
                                    {
                                        Br_value1 [j] =
                                        st + (hi [j]-cj)/snew ;
                                        if ( ns [j] != EMPTY )
                                        {
                                            pproj_minheap_update (Heap,
                                               ns, Br_value1, nbrk, ns [j]) ;
                                        }
                                        else
                                        {
                                            pproj_minheap_add (j, Heap,
                                                ns, Br_value1, &nbrk) ;
                                        }
                                    }
                                    else
                                    {
                                        if ( ns [j] != EMPTY )
                                        {
                                            pproj_minheap_delete (Heap,
                                               ns, Br_value1, &nbrk, ns [j]);
                                        }
                                    }
                                }
                            }
                        }
                    }

#ifndef NDEBUG
                    if ( PrintLevel > 2 )
                    {
                        if ( col < ncol ) PRINTF ("    free: %i\n", col);
                        else              PRINTF ("    drop row: %i\n", row) ;
                    }
                    check_minheap (Heap, Br_value1, ns, nbrk, ntot,
                        "line search of ssor, phase 1") ;
#endif

                    if ( fd <= ZERO )
                    {

#ifndef NDEBUG
                        if ( PrintLevel > 2 )
                        {
                            PRINTF("    premature break from line search\n") ;
                        }
#endif

                        flag = 1 ;
                        fd = ZERO ;
                        break ;
                    }
                }
                if ( sd > ZERO )
                {
                    st += fd/sd ;
                }
                if ( st > st0 )
                {
                    st = st0 ;
                    flag = -1 ;
                }

                for (k = 1; k <= nbrk; k++)
                {
                    ns [Heap [k]] = EMPTY ;
                }
            }
        }
#ifndef NDEBUG
        where = "end of phase1 line search" ;
        /* check_phase1line (I, dk, st, where) ;*/
        check_line (I, flag, W->blks - 1, nup, dk, st) ;
        checkb (I, where) ;
        check_const (NULL, ZERO, ns, EMPTY, ntot, where) ;
#endif

        /* ------------------------------------------------------------------ */
        /* finish c update and check for binding columns */
        /* ------------------------------------------------------------------ */

        it++ ;

        xbound = ZERO ;
        nf = nr = 0 ;
        for (j = 0; j < ncol; j++)
        {
            c [j] += st*pA [j] ;
            cj = c [j] ;
            ibj = ib [j] ;
            if ( ibj == 0 )
            {
                if ( cj < lo [j] )
                {
                    if ( PrintLevel > 1 )
                    {
                        printf ("lower bind %i\n", j) ;
                    }
                    ib [j] = ibj = -1 ;
                    t = lo [j] ;
                }
                else if ( cj > hi [j] )
                {
                    if ( PrintLevel > 1 )
                    {
                        printf ("upper bind %i\n", j) ;
                    }
                    ib [j] = ibj = +1 ;
                    t = hi [j] ;
                }
                else
                {
                    if ( PrintLevel > 2 )
                    {
                        printf ("%i continues free\n", j) ;
                    }
                    F [nf++] = j ;
                }
                if ( ibj )
                {
#ifndef NDEBUG
                    x [j] = t ;
#endif
                    q = Ap [j+1] ;
                    if ( t != ZERO )
                    {
                        xbound += fabs (t) ;
                        for (p = Ap [j]; p < q; p++)
                        {
                            i = Ai [p] ;
                            ax = Ax [p] ;
                            b [i] -= t*ax ;
                            D [i] -= ax*ax ;
                            /* save rows associated with new bound columns */
                            if ( ns [i] == EMPTY )
                            {
                                ns [i] = 1 ;
                                R [nr++] = i ;
                            }
                        }
                    }
                    else /* update D, record rows where free columns bound */
                    {
                        for (p = Ap [j]; p < q; p++)
                        {
                            i = Ai [p] ;
                            D [i] -= Ax [p]*Ax [p] ;
                            if ( ns [i] == EMPTY )
                            {
                                ns [i] = 1 ;
                                R [nr++] = i ;
                            }
                        }
                    }
                }
                else /* x_j remains free */
                {
                     xbound += fabs (cj) ;
                }
            }
            else /* variable is bound */
            {
                if ( ibj > 0 )
                {
                     xbound += fabs (hi [j]) ;
                }
                else
                {
                     xbound += fabs (lo [j]) ;
                }
            }
        }
#ifndef NDEBUG
        where = "end of variable rebind in phase 1" ;
        checkb (I, where) ;
#endif

        /* ------------------------------------------------------------------ */
        /* update AFT by removing the bound columns */
        /* ------------------------------------------------------------------ */
        for (k = 0; k < nr; k++)
        {
            i = R [k] ;
            ns [i] = EMPTY ;
            D [i] = MAX (D [i], sigma) ;
            p0 = l = p = AFTp [i] ;
            q = p + AFTnz [i] ;
            for (; p < q; p++)
            {
                if ( !ib [AFTi [p]] )
                {
                    AFTi [l] = AFTi [p] ;
                    AFTx [l] = AFTx [p] ;
                    l++ ;
                }
            }
            AFTnz [i] = l - p0 ;
        }

        /* ------------------------------------------------------------------ */
        /* Look for rows to add. Also, augment the grad L (lambda) associated
           with the active row (computed at start of SSOR step), with
           grad L (lambda) for the inactive (dropped) rows. */
        /* ------------------------------------------------------------------ */

        if ( PrintLevel > 1 )
        {
            printf ("final ssor st: %25.15e\n", st) ;
        }
        Ll = ni1 ;
        Ul = ni2 ;
        Rl = nrow ;
        for (i = 0 ; i < nrow ; i++)
        {
            j = ir [i] ;
            if ( j <= ni )
            {
                /* row i is active */
                Rl = i ;
                lambda [i] += st*dk [i] ;
                if ( j < 0 )
                {
                    Ll = -j ;
                }
                else if ( j > 0 )
                {
                    Ul = j ;
                }
            }
            else
            {
                /* row i is inactive */
                t = b [i] ;
                ASSERT (ir [i] > ni) ;
                p = AFTp [i] ;
                q = p + AFTnz [i] ;
                for (; p < q; p++)
                {
                    t -= c [AFTi [p]]*AFTx [p] ;
                }
                ineqindex = j - ni ;
                s = t + bl [ineqindex] ;
                u = sigma*lambda [i] ;
                /* Include the proximal term when deciding whether activating
                   the row will improve the objective. */
                if ( s - u  > ZERO ) /* lambda_i > 0 increases L */
                {
                    if ( PrintLevel > 1 )
                    {
                        printf ("add row: %i (lower bind)\n", i) ;
                    }
                    /* add row, lower bound */
                    l = RLinkUp [Rl]  ;
                    RLinkUp [i] = l ;
                    RLinkDn [l] = i ;
                    RLinkUp [Rl] = i ;
                    RLinkDn [i] = Rl ;
                    Rl = i ;
                    l = lLinkUp [Ll] ;
                    lLinkUp [ineqindex] = l ;
                    lLinkDn [l] = ineqindex ;
                    lLinkUp [Ll] = ineqindex ;
                    lLinkDn [ineqindex] = Ll ;
                    Ll = ineqindex ;
                    b [i] += bl [ineqindex] ;
                    ir [i] = -ineqindex ;     /* row is active at lower bound */
                }
                else
                {
                    s = t + bu [ineqindex] ;
                    if ( s - u < ZERO ) /* lambda_i < 0 improves L */
                    {
                        if ( PrintLevel > 1 )
                        {
                            printf ("add row: %i (upper bind)\n", i) ;
                        }
                        /* add row, upper bound */
                        l = RLinkUp [Rl]  ;
                        RLinkUp [i] = l ;
                        RLinkDn [l] = i ;
                        RLinkUp [Rl] = i ;
                        RLinkDn [i] = Rl ;
                        Rl = i ;
                        l = uLinkUp [Ul] ;
                        uLinkUp [ineqindex] = l ;
                        uLinkDn [l] = ineqindex ;
                        uLinkUp [Ul] = ineqindex ;
                        uLinkDn [ineqindex] = Ul ;
                        Ul = ineqindex ;
                        b [i] += bu [ineqindex] ;
                        ir [i] = ineqindex ;  /* row is active at upper bound */
                    }
                }
            }
        }

#ifndef NDEBUG
        where = "at tail of phase 1" ;
        /* FALSE: also check deleted rows */
        check_AFT (I, FALSE, where) ;
        checkb (I, where) ;
        checkc (I, where) ;
        /* check that dual objective increases */
        check_dual (I, NULL, where, TRUE, TRUE) ;
        check_link (I, (int *) NULL, 0, where) ;
#endif
    }

#ifndef NDEBUG
    if ( PrintLevel > 0 )
    {
        PRINTF("phase1 iterations complete\n") ;
    }
#endif
    W->nf = nf ;             /* return number of free indices */
    Stat->phase1_its  = it ; /* record number of iterations */

    /* Check for convergence and perform the first part of the change
       of variables described in pproj. Recompute b from scratch.
       Current estimate of the projection is x. */
    pproj_initx (W->dlambda, ZERO, nrow) ;
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] )
        {
            k = ir [i] ;
            if ( k > 0 )
            {
                if ( k <= ni )
                {
                    b [i] = Prob->bu [k] ;
                }
                else
                {
                    b [i] = ZERO ;
                }
            }
            else
            {
                b [i] = Prob->bl [-k] ;
            }
        }
        else
        {
            b [i] = Prob->b [i] ;
        }
    }
    p = 0 ;
    s = ZERO ;
    absAx = W->arrayd ;
    pproj_initx (absAx, ZERO, nrow) ;
    for (j = 0; j < ncol; j++)
    {
        if ( ib [j] )         /* variable is bound */
        {
            if ( ib [j] > 0 ) /* variable at upper bound */
            {
                t = hi [j] ;
                W->lo [j] = lo [j] - t ;
                W->hi [j] = ZERO ;
            }
            else              /* variable at lower bound */
            {
                t = lo [j] ;
                W->hi [j] = hi [j] - t ;
                W->lo [j] = ZERO ;
            }
            x [j]  = t ;
            c [j] -= t ;
        }
        else                 /* variable is free */
        {
            t = c [j] ;
            c [j] = ZERO ;
            W->lo [j] = lo [j] - t ;
            W->hi [j] = hi [j] - t ;
            x [j] = t ;
        }
        s += fabs (t) ;
        if ( t != ZERO )
        {
            q = Ap [j+1] ;
            for (p = Ap [j]; p < q; p++)
            {
                INT ai ;
                ai = Ai [p] ;
                u = t*Ax [p] ;
                b [ai] -= u ;
                absAx [ai] += fabs (u) ;
            }
        }
    }
#ifndef NDEBUG
    where = "end of phase1" ;
    checkb (I, where) ;
    check_dual (I, NULL, where, TRUE, TRUE) ;
#endif

    W->normx = s ; /* L1 norm of x */
    /* Since the dropped rows all have 0 in their subdifferential set,
       errdual is the absolute max of b [i] for the active rows */
    errdual = ZERO ;
    s = ZERO ;
    for (i = RLinkUp [nrow]; i < nrow; i = RLinkUp [i])
    {
        if ( errdual < fabs (b [i]) )
        {
            errdual = fabs (b [i]) ;
        }
        if ( s < absAx [i] )
        {
            s = absAx [i] ;
        }
    }
    if ( s == ZERO )
    {
        s = ONE ;
    }
    /*recompute absAx when ||x||_1 changes by factor >= 1.5*/
    W->absAx = s ;

    /* in SpaRSA, only monitor undecided index set when gradient <= grad0 */
    W->grad0 = errdual*Parm->grad_decay ;

    if ( !Parm->stopabs )
    {
        errdual /= s ;
    }
    if ( PrintLevel > 0 )
    {
        printf ("phase1 errdual: %e absAx: %e\n", errdual, s) ;
    }
    if ( errdual <= Prob->grad_tol )
    {
        Stat->errdual = errdual ;
        return (PPROJ_SOLUTION_FOUND) ;
    }

    /* for dasa, replace b by b - sigma*lambda */
    pproj_saxpy (b, lambda, -sigma, nrow) ;

    if ( !Parm->cholmod ) /* return if purely iterative method is used */
    {
        pproj_copyx (W->lambda_tot, lambda, nrow) ;
        if ( !W->shiftl_is_zero )
        {
            pproj_saxpy (W->lambda_tot, shift_l, ONE, nrow) ;
        }
        return (PPROJ_TOLERANCE_NOT_MET) ;
    }

    /* otherwise update/downdate is used */
    {
        int blk, blks ;
        INT nactive, nextrow, Annz, ATnz, Lnnz,
            *Anz, *lstart, *ustart, *col_to_blk, *row_start, *DropP ;
        cholmod_common *cmm ;

        cmm = W->cmm ;

        col_to_blk = W->col_to_blk ;
        row_start = W->row_start ;
        blks = W->blks ;
        lstart = W->lstart ;
        ustart = W->ustart ;

        /* lstart and ustart are associated with the inequalities that are
           strict.  For each block in the multilevel decomposition,
           lstart points to the first singleton at its lower bound while ustart
           points to the first singleton at it upper bound */
        for (k = 0; k < blks; k++)
        {
            lstart [k] = ni1 ;
            ustart [k] = ni2 ;
        }

        sol_to_blk = W->sol_to_blk ;
        prevblk = EMPTY ;
        for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
        {
            blk = sol_to_blk [j] ;
            if ( blk > prevblk )
            {
                lstart [blk] = j ;
                prevblk = blk ;
            }
        }

        prevblk = EMPTY ;
        for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
        {
            blk = sol_to_blk [j] ;
            if ( blk > prevblk )
            {
                ustart [blk] = j ;
                prevblk = blk ;
            }
        }

        blk = 0 ;     /* next block is 0 */
        nextrow = 0 ; /* block starts at row 0 */
        for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
        {
            row = ineq_row [j] ;   /* row associated with strict inequality */
            if ( row >= nextrow ) /* find block containing row */
            {
                /* need to find nextrow for the block just past row */
                while ( row_start [++blk] <= row ) ;
                nextrow = row_start [blk] ;
                lstart [blk-1] = j ;
            }
        }

        blk = 0 ;     /* next block is 0 */
        nextrow = 0 ; /* block starts at row 0 */
        for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
        {
            row = ineq_row [j] ;   /* row associated with strict inequality */
            if ( row >= nextrow ) /* find block containing row */
            {
                /* need to find nextrow for the block just past row */
                while ( row_start [++blk] <= row ) ;
                nextrow = row_start [blk] ;
                ustart [blk-1] = j ;
            }
        }

        DropP = W->arrayi ;
        Anz = Prob->Anz ;
        for (j = 0; j < ncol; j++)
        {
            Anz [j] = 0 ;          /* number of active rows in column */
            DropP [j] = Ap [j+1] ; /* points to first dropped row in column */
        }
        Annz = 0 ; /* number of nonzeros in active rows and free columns of A */
        ATnz = 0 ; /* number of nonzeros in active rows of full matrix */
        nactive = 0 ; /* number of active rows */
        p = 0 ;
        for (i = 0; i < nrow; i++)
        {
            q = ATp [i+1] ;
            if ( ir [i] <= ni ) /* active row */
            {
                nactive++ ;
                ATnz += q - p ;
                Annz += AFTnz [i] ;
                for (; p < q; p++)
                {
                    j = ATi [p] ;
                    k = Ap [j] + Anz [j] ;
                    Anz [j]++ ;
                    Ax [k] = ATx [p] ;
                    Ai [k] = i ;
                }
            }
            else                      /* inactive (dropped) row */
            {
                AFTnz [i] = 0 ;       /* phase1 (full AFT) ascent (active AFT)*/
                for (; p < q; p++)
                {
                    j = ATi [p] ;
                    k = --DropP [j] ; /* put index at bottom of column */
                    Ax [k] = ATx [p] ;
                    Ai [k] = i ;
                }
            }
        }
        W->ATnz = ATnz ;
        W->Annz = Annz ;
        W->nactive = nactive ;
        W->A->packed = FALSE ; /* just changed the structure to unpacked */

        /* If SSOR is used in conjunction with the update and downdate,
           then we evaluate nnz (chol (AF*AF')) since this determines
           how many SSOR iterations will be performed */
        if ( (Parm->use_ssor0 == TRUE) || (Parm->use_ssor1 == TRUE) )
        {
            INT *Iwork, *Parent, *Post, *First, *Level, *ColCount ;
            cholmod_sparse AFT ;

            pproj_cholmod_sparse (&AFT, ncol, nrow, AFTp [nrow], AFTp, AFTnz,
                                   AFTi, AFTx, TRUE, FALSE, CHOLMOD_REAL);
            cholmod_allocate_work (nrow, 7*nrow+ncol, 0, W->cmm) ;
            Iwork = cmm->Iwork ;
            Parent   = Iwork + 2*nrow+ncol ;
            Post     = Iwork + 3*nrow+ncol ;
            First    = Iwork + 4*nrow+ncol ;
            Level    = Iwork + 5*nrow+ncol ;
            ColCount = Iwork + 6*nrow+ncol ;

            /* compute the etree of A(:,F)*A(:,F)' */
            /* uses Iwork (0..nrow+ncol-1) */
            cholmod_etree (&AFT, Parent, cmm) ;

            /* postorder the etree */
            /* uses Iwork (0..2*nrow-1) */
            cholmod_postorder (Parent, nrow, NULL, Post, cmm) ;

            /* compute the row/col counts, first, and level */
            /* uses Iwork (0..2*nrow+ncol-1) */
            cholmod_rowcolcounts (W->A, F, nf,
                Parent, Post, NULL, ColCount, First, Level, cmm) ;

            /* sum up the ColCounts for just the active rows */
            Lnnz = 0 ;
            for (row = RLinkUp [nrow]; row < nrow; row = RLinkUp [row])
            {
                /* ColCount includes the diagonal */
                Lnnz += (ColCount [row] - 1) ;
            }
            W->Lnnz = Lnnz ;
        }
    }
    return (PPROJ_TOLERANCE_NOT_MET) ;
}
