/* ========================================================================= */
/* === pproj_ssor1 ========================================================= */
/* ========================================================================= */

/*
    Apply SSOR Gauss-Seidel preconditioned conjugate gradients to
    the dual problem. Periodically, we check to see whether a variable
    should be freed or a row should be dropped at the current iterate.
    If the answer is no, then the CG iteration continues. If there is
    a change in the active set, then we perform a line search along the
    direction from the starting CG iterate to the current point.
    As the iterations proceed, we also save up to ssormem of the
    previous CG search directions. After the line search and before
    restarting the CG iteration, we minimize the objective corresponding
    to the new active set over the space spanned by the directions in memory.
    We can efficiently solve this subspace problem since the CG method
    is diagonalizing the Hessian relative to the search directions,
    and we only need to consider the effect of the rank-1 changes to
    the Hessian associated with the freeing of variables and the rank-2
    changes to the Hessian associated with the dropping of rows.

    The difference between ssor0 and ssor1 is that the goal in ssor0
    is to free as many variables and drop as many rows as possible by
    performing a gradient ascent iteration. In ssor1, we try to solve
    the dasa subproblems more accurately by performing an SSOR
    Gauss-Seidel preconditioned conjugate gradient iteration. A few
    iterations of the CG algorithm generates an approximation to the
    solution to the linear system solved in dasa using a Cholesky
    factorization.  The routine pproj_iterquery uses the number of nnz's in
    the Cholesky factor and the number of nnz's in the reduced A matrix
    to determine how many ssor iterations should be performed.
    The ssor and coordinate ascent routines can be turned on or off
    using the parameters Parm->coorquery, Parm->ssor0query, and
    Parm->ssor1query.

    Consider an optimization problem of the form
 
    max -.5x'Qx + b'x

    The SSOR Gauss-Seidel preconditioner P =  inv (S)' inv (S),
    where S = (D+L) inv (sqrt(D)) and L+D+L' = Q, was introduced
    in the preconditioned gradient ascent code ssor0. The corresponding
    SSOR preconditioned conjugate gradient iteration is the following:

    beta_0 = 0, alpha_0 = 0, g_0 = b - Qx_0, d_0 = Pg_0

    for k = 0, 1, 2, ...

        x_k+1 = x_k + alpha_k d_k,   alpha_k = g_k'd_k/(d_k'Qd_k)

        g_k+1 = g_k - alpha_k Qd_k,

        d_k+1 = Pg_k+1 + beta_k d_k, beta_k = gk+1' P gk+1 / (gk' P gk)

    end

    Here we are using the Fletcher-Reeves formula, and the gradients are
    updated in each iteration for stability, as explained in [1].
    Note that g_k'Pg_k and g_k'd_k are equal, so we only need to compute
    one of them.

    According to the theory for the conjugate gradient method, the
    directions d0, ... , dk are Q conjugate. Hence, we have

    [d0 | d1 | ... | dk]'Q[d0 | d1 | ... | dk] =

        diag([d0'Qd0, d1'Qd1, ... , dk'Qdk])

    As the CG iteration proceeds, we save the ssormem most recent d_k.
    When we reach an iterate where there is a change in the active set
    (either a variable frees or a row drops), we maximize the dual function
    along the line segment [lambda0, lambdak]. If lambda_bar is the optimum,
    then we maximize the dual function over lambda_bar plus the linear
    space spanned by [dj, ... dk], the directions stored in the limited
    memory. The effect to the new free variables and the new dropped rows
    is to add to the current Hessian Q the sum of say m rank-1 matrices
    r_i r_i', 1 <= i <= m. The quadratic for this limited memory
    optimization problem is

    [dj | ... | dk] (Q + sum_{i=1}^m r_i r_i') [dj | ... | dk] =

    diag([dj'Qdj, ... , dk'Qdk]) + sum_{i=1}^m v_i v_i' where

    v_i' =   r_i' [dj | ... | dk]. Hence, the effect of the change in
    the active set is to perturb the diagonal matrix by the sum of m
    rank-1 matrices. We can either form this "ssormem by ssormem" matrix
    and Cholesky factor it, or we can update its factorization, starting from
    the diagonal matrix.

    If we directly implemented the preconditioned CG algorithm as
    described above, each iteration requires 5 multiplications between
    A_F and a vector. By reorganizing the computations, this cost can
    be reduced to 4 multiplications. We introduce the following new
    variables:

    pk = inv (D+L) Q dk
    qk = inv (D+L) gk
    sk = (D+L)' dk

    Observe that

    pk = inv (D+L) (D + L + L') dk
       = inv (D+L) (D + L + L') inv (D+L)' sk
       = inv (D+L) (I + L inv (D+L)') sk
       = inv (D+L) (sk + L dk)

    Also note that the numerator of alpha_k satisfies the following relations:

    gk' dk = qk' sk = gk' P gk = qk' D qk

    In terms of the new variables, the preconditioned CG iteration
    can be expressed as follows:

    q_0 = inv (D+L) g0

    s0 = D q0

    d0 = inv (D+L)' s0

    x1 = x0 + alpha_0 d0,   alpha_0 = q0' D q0/(d0' Q d0)

    for k = 0, 1, 2, ...

        pk = inv (D+L) (sk + Ldk)

        qk+1 = qk - alpha_k pk,       alpha_k = (qk' D qk)/(dk' Q dk)

        sk+1 = D qk+1 + beta_k sk,    beta_k = (qk+1' D qk+1) / (qk' D qk)
 
        dk+1 = inv (D+L)' sk+1

        xk+2 = xk+1 + alpha_k+1 dk+1, alpha_k+1 = (qk+1' D qk+1) /(dk+1' Q dk+1)

    end

    Comparing the new structure to the old, note that computation of x1
    was moved to the initialization step, and then in the loop, the computation
    of x is moved to the end of the loop. With this organization, at the
    end of each iteration, we have both a new direction vector and the new x
    associated with the new direction.  If Q = A * A', then A'*dk can
    be computed at the same time that dk = inv (D+L)' sk is evaluated. There
    are two multiplies when dk+1 is evaluated and two when we solve for pk.

    [1] William W. Hager and Hongchao Zhang, The Limited Memory Conjugate
        Gradient Method, SIAM Journal on Optimization, 23 (2013), pp. 2150-2168. 
*/

#include "pproj.h"

int pproj_ssor1 /* return status:
                       (0) PPROJ_SOLUTION_FOUND
                       (4) PPROJ_SSOR_NONASCENT
                       (5) PPROJ_SSOR_MAX_ITS
                       (6) PPROJ_SWITCH_TO_UPDOWN
                       (7) PPROJ_ALL_ROWS_DROPPED
                       (9) PPROJ_STATUS_OK */
(
    PPcom *I
)
{
    int     blk, debug, flag, ibj, PrintLevel, ssormem, sub_it, sub_inc,
            status, memp, n, n1, n2, lineflag, nlinesearch,
           *ib, *sol_to_blk ;
    INT     it, it0, i, j, k, first, last, nup, chg,
            l, m, p, p1, q, q1, col, row, nrow, ncol, ni, ni1, ni2,
            ntot, nbrk, ssor1_drop, nb, nf, nf0, ineqindex,
            nrowdel, nrowadd, ncoladd, ncoldel, rowp, nr,
           *Ai, *Ap, *ATp, *ATi, *AFTi, *AFTnz, *AFTp,
           *Heap, *F, *bound, *rowlist, *RLinkDn, *RLinkUp,
           *uLinkUp, *uLinkDn, *lLinkUp, *lLinkDn, *lstart, *ustart,
           *RowmodFlag, *RowmodList, *ColmodFlag, *ColmodList,
           *ir, *ns, *ineq_row, *worki ;

    double  Br, ax, cj, sigma, dk_row, alpha, beta, beta_denom, errdual, errls,
            s, t, u, y, z, st, st0, fd, fn, sd, tic, dknew, snew,
           *Ax, *ATx, *AFTx, *b, *bl, *bu, *c, *hi, *lo, *lambda, *shift_l,
           *Br_value, *D, *L, *Lj, *Lk, *diag, *e, *v, *w,
           *dk, *dk_start, *pk, *qk, *sk, *dl, *pA, *gk, *workd, *x ;

    PPprob *Prob ;
    PPstat *Stat ;
    PPwork    *W ;
    PPparm *Parm ;

#ifndef NDEBUG
    char   *where ;
    I->Check->location = 4 ; /* code operates in ssor1 */
#endif

    tic = pproj_timer () ;

    /* extract the problem, statistics, and work structures from I */
    Prob = I->Prob ;
    Stat = I->Stat ;
    ssor1_drop = Stat->ssor1_drop ;
    W = I->Work ;
    Parm = I->Parm ;
    debug = Parm->debug ;
    PrintLevel = Parm->PrintLevel ;

    /* Problem matrix */
    Ap   = Prob->Ap ;
    Ai   = Prob->Ai ;
    Ax   = Prob->Ax ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    ni1 = ni + 1 ;
    ni2 = ni + 2 ;
    ntot = ni + ncol + 1 ;
    ineq_row = Prob->ineq_row ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    lo = Prob->lo ;
    hi = Prob->hi ;

    /* Transpose of A */
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;

    /* Transpose of AF */
    AFTp = W->AFTp ;
    AFTnz = W->AFTnz ;
    AFTi = W->AFTi ;
    AFTx = W->AFTx ;

    /* Links point to active rows */
    RLinkUp = W->RLinkUp ;
    RLinkDn = W->RLinkDn ;

    /* Links point to active singletons. lLinkUp points to the strict
       inequalities with b_i = bl_i while uLinkUp points to the strict
       inequalities with b_i = bu_i. Since only one of these can hold for
       each i, lLinkUp and uLinkUp can be stored in the same array */
    lLinkUp = W->SLinkUp ; 
    lLinkDn = W->SLinkDn ;
    uLinkUp = W->SLinkUp ;
    uLinkDn = W->SLinkDn ;

    /* working arrays */
    lambda = W->lambda ;
    shift_l = W->shift_l ;
    x = W->x ;
    ib = W->ib ;
    ir = W->ir ;
    b = W->b ; /* the part of grad L (lambda) associated with bound variables */
    c = W->c ; /* y + A'lambda */
    F = W->F ;
    D = W->D ;
    ns = W->ns ;
    ssormem = W->ssormem ;

    nf = W->nf ;
    nrowdel = W->nrowdel ;
    nrowadd = W->nrowadd ;
    ncoladd = W->ncoladd ;
    ncoldel = W->ncoldel ;
    RowmodFlag = W->RowmodFlag ;
    RowmodList = W->RowmodList ;
    ColmodFlag = W->ColmodFlag ;
    ColmodList = W->ColmodList ;

    lstart = W->lstart ;
    ustart = W->ustart ;
    sol_to_blk = W->sol_to_blk ;

    /* ======== start allocations ========
    -> work arrays in ssor step:

       Heap     - INT    ntot, line search
       bound    - INT    ncol, line search
       rowlist  - INT    nrow, active rows associated with update
       pA       - double ncol, A'lambda
       Br_value - double ntot, line search
       gk       - double nrow, starting gradient in CG iteration
       pk       - double nrow, used in CG iteration = inv (D+L) Q dk
       qk       - double nrow, used in CG iteration = inv (D+L) gk
       sk       - double nrow, used in CG iteration = (D+L)' dk
       dk       - double nrow, search direction in a CG iteration (ssor memory)
       dl       - double nrow, search direction from initial CG iterate
       L        - double ssormem*(ssormem-1)/2, Cholesky factor below diagonal
       diag     - double ssormem, diagonal of Cholesky factory
       e        - double ssormem, used in updates
       v        - double ssormem, used in updates
       w        - double ssormem, used in updates */

    worki = W->arrayi ;
    Heap    = worki ; worki += ntot ;
    bound   = worki ; worki += ncol ;
    rowlist = worki ; worki += nrow ;

    workd = W->arrayd ;
    pA       = workd ; workd += ncol ;
    Br_value = workd ; workd += ntot ;
    gk       = workd ; workd += nrow ;
    pk       = workd ; workd += nrow ;
    qk       = workd ; workd += nrow ;
    sk       = workd ; workd += nrow ;
    dk_start = workd ; workd += ssormem*nrow ;
    dl       = workd ; workd += nrow ;
    L        = workd ; workd += (ssormem*(ssormem-1))/2 ;
    diag     = workd ; workd += ssormem ;
    e        = workd ; workd += ssormem ;
    v        = workd ; workd += ssormem ;
    w        = workd ; workd += ssormem ;
    /* ======== end of allocations ======== */

    nb = 0 ;
    /* all bound indices are stored in bound array */
    for (j = 0; j < ncol; j++)
    {
        if ( ib [j] )
        {
            bound [nb] = j ;
            nb++ ;
        }
    }
    sigma = W->sigma ;
    it = 0 ;
    chg = 0 ;  /* count the number of variables freed and rows dropped */
    /* flag = -1 means restart the CG iteration
       flag =  0 means no change in active constraints
       flag =  1 means all rows became inactive
       flag =  2 means the CG iteration generated a nonascent point */
    flag = -1 ;

#ifndef NDEBUG
    where = "at very start of ssor1" ;
    checkF (I, where) ;
    checkc (I, where) ;
    Heap [ntot-1] = EMPTY ;
    checkb (I, where) ;
    check_const (NULL, 0, ns, EMPTY, ntot, where) ;
    checkD (I, where) ;
    check_AT (I, where) ;
    check_AFT (I, TRUE, where) ;
    check_link (I, (int *) NULL, 0, where) ;
    check_dual (I, NULL, where, TRUE, TRUE) ;
#endif

    while ( 1 )
    {
        if ( flag < 0 ) /* restart CG iteration */
        {
            /* record the starting iteration */
            it0 = it ;
            it++ ;
            /* number of CG iterations before checking for active constraints */
            sub_inc = 2 ;
            first = RLinkUp [nrow] ;
            if ( first == nrow ) /* all equations have been dropped */
            {
                status = PPROJ_ALL_ROWS_DROPPED ;
                break ;
            }
            /* b = part of grad L (lambda) associated with bound variables
                   (includes proximal term -sigma*lambda)
               gk = grad L_R (lambda) for active rows */

            /* compute objective gradient at the current iterate */
            for (i = first; i < nrow; i = RLinkUp [i])
            {
                t = b [i] ; /* includes -sigma*lambda [i] */
                ASSERT (ir [i] <= ni) ;
                p = AFTp [i] ;
                q = p + AFTnz [i] ;
                for (; p < q; p++)
                {
                    t -= AFTx [p]*c [AFTi [p]] ;
                }
                gk [i] = t ;
            }

            /* compute q0 = inv (D+L) * g0,  s0 = D * q0, and
                       alpha_num = q0'Dq0.  (initializations above) */

            pproj_initFx (pA, ZERO, F, nf) ; /* set pA [F [k]] = 0 */

            i = first ;
            s = gk [i] ;
            sk [i] = s ;
            t = s/D [i] ;
            fd = s*t ; /* = q0' * D * q0 = d0'g0 = s0*q0 */
            qk [i] = t ;

            /* compute s0, q0, and fd */
            last = RLinkDn [nrow] ;
            while (i < last)
            {

                ASSERT (ir [i] <= ni) ;

                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    pA [AFTi [p]] += AFTx [p]*t ;
                }

                i = RLinkUp [i] ;

                ASSERT (ir [i] <= ni) ;

                t = ZERO ;

                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    t += AFTx [p]*pA [AFTi [p]] ;
                }
                s = gk [i] - t ;
                sk [i] = s ;
                t = s/D [i] ;
                fd += s*t ;
                qk [i] = t ;
            }

            dk = dk_start ;
            memp = 0 ; /* pointer to current memory location */

            /* compute d0 = inv (D+L)' * s0, pA = A_F' * d0, and sd = q0 D q0 */
            pproj_initFx (pA, ZERO, F, nf) ; /* set pA [F [k]] = 0 */
            sd = t*t ; /* sd initially stores d'd */
            dk [i] = t ;

            while ( i > first )
            {
                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1 ; p++)
                {
                    pA [AFTi [p]] += AFTx [p]*t ;
                }

                i = RLinkDn [i] ;

                ASSERT (ir [i] <= ni) ;

                t = ZERO ;

                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    t += AFTx [p]*pA [AFTi [p]] ;
                }
                t = (sk [i] - t)/D [i] ;
                dk [i] = t ;
                sd += t*t ;
            }

            p = AFTp [i] ;
            p1 = p + AFTnz [i] ;
            for (; p < p1; p++)
            {
                pA [AFTi [p]] += AFTx [p]*t ;
            }
            /* end computation of dk and pA = A'*dk */

            sd *= sigma ; /* sd stored sigma * dk' * dk */
            for (k = 0; k < nf; k++)
            {
                j = F [k] ;
                sd += pA [j]*pA [j] ; /* sd stores  dk'(AF*AF' + sigma I)dk */
            }
            if ( sd == ZERO ) /* gradient vanishes at current iterate */
            {
                status = PPROJ_STATUS_OK ; /* relaxed dual maximized */
                break ;
            }

            /* numerator of alpha is denominator of beta */
            beta_denom = fd ;
            alpha = fd/sd ;
            diag [memp] = sd ;
            /* dl stores the total movement from the initial iterate */
            for (i = first; i < nrow; i = RLinkUp [i])
            {
                dl [i] = alpha*dk [i] ;
            }
        } /* CG startup complete */

        /* perform CG iterations */
        for (sub_it = 0; sub_it < sub_inc; sub_it++)
        {
            /* evaluate pk, qk, and fd = numerator of next beta
                                       = numerator of next alpha */
            pproj_initFx (pA, ZERO, F, nf) ; /* set pA [F [k]] = 0 */
            i = first ;
            s = sk [i]/ D [i] ;
            pk [i] = s ;
            t = dk [i] - s ;
            s = qk [i] - alpha*s ;
            qk [i] = s ;
            fd = s*s*D [i] ;
            while (i < last)
            {
                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    pA [AFTi [p]] += AFTx [p]*t ;
                }

                i = RLinkUp [i] ;

                t = ZERO ;
                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    t += AFTx [p]*pA [AFTi [p]] ;
                }
                t = (sk [i] + t)/D [i] ;
                pk [i] = t ;
                s = qk [i] - alpha*t ;
                qk [i] = s ;
                fd += s*s*D [i] ;
                t = dk [i] - t ;
            }
            beta = fd/beta_denom ;
            beta_denom = fd ;        /* fd = denominator of next beta */

            memp++ ;
            if ( memp == ssormem )
            {
                memp = 0 ;
                dk = dk_start ;
            }
            else
            {
                dk = dk_start+(memp*nrow) ;
            }

            /* evaluate sk, dk, and sd */
            pproj_initFx (pA, ZERO, F, nf) ; /* set pA [F [k]] = 0 */
            t = D [i]*qk [i] + beta*sk [i] ; /* = new sk [i] */
            sk [i] = t ;
            t /= D [i] ;
            sd = t*t ;  /* sd stores dk' * dk */
            dk [i] = t ;
            while (i > first)
            {
                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    pA [AFTi [p]] += AFTx [p]*t ;
                }
                i = RLinkDn [i] ;

                t = ZERO ;
                p = AFTp [i] ;
                p1 = p + AFTnz [i] ;
                for (; p < p1; p++)
                {
                    t += AFTx [p]*pA [AFTi [p]] ;
                }
                s = D [i]*qk [i] + beta*sk [i] ;
                sk [i] = s ;
                t = (s - t)/D [i] ;
                dk [i] = t ;
                sd += t*t ;
            }

            p = AFTp [i] ;
            p1 = p + AFTnz [i] ;
            for (; p < p1; p++)
            {
                pA [AFTi [p]] += AFTx [p]*t ;
            }

            sd *= sigma ; /* sd stored sigma * dk' * dk */
            for (k = 0; k < nf; k++)
            {
                j = F [k] ;
                sd += pA [j]*pA [j] ; /* sd stores  dk'(AF*AF' + sigma I)dk */
            }
            if ( sd == ZERO ) /* gradient vanishes at current iterate */
            {
                status = PPROJ_STATUS_OK ; /* relaxed dual maximized */
                pproj_initx (pA, ZERO, ncol) ;
                for (i = first; i < nrow; i = RLinkUp [i])
                {
                    t = dl [i] ;
                    lambda [i] += t ;
                    b [i] -= sigma*t ;
                    q = ATp [i+1] ;
                    for (p = ATp [i] ; p < q ; p++)
                    {
                        pA [ATi [p]] += t*ATx [p] ;
                    }
                }
                pproj_saxpy (c, pA, ONE, ncol) ;
                break ;
            }
            diag [memp] = sd ;
            alpha = fd/sd ;
            /* dl stores the total movement from the initial iterate */
            for (i = first; i < nrow; i = RLinkUp [i])
            {
                dl [i] += alpha*dk [i] ;
            }
        }

        it += sub_it ;
        nlinesearch = 0 ;
        /* Minimize over the directions in memory until no change in the
           active constraints. If in the first iteration there is no
           change in the active constraints, then continue the CG iteration.
           If there is a change in the active constraints, then we continue
           to minimize over the search directions in memory until there is
           no change in the active constraints. */
        while ( 1 )
        {
            nlinesearch++ ;
            pproj_initx (pA, ZERO, ncol) ;
            fd = ZERO ;
            sd = ZERO ;

            for (i = first; i < nrow; i = RLinkUp [i])
            {
                t = dl [i] ;
                q = ATp [i+1] ;
                for (p = ATp [i] ; p < q ; p++)
                {
                    pA [ATi [p]] += t*ATx [p] ;
                }
                fd += t*gk [i] ;
                sd += t*t ;
            }
            if ( fd < ZERO )
            {
                flag = 2 ; /* not an ascent direction */
                break ;
            }
            sd *= sigma ;
            for (k = 0; k < nf; k++)
            {
                sd += pA [F [k]]*pA [F [k]] ;
            }
            nbrk = 0 ;
            st0 = st = ONE ;
            for (k = 0; k < nb; k++)
            {
                j = bound [k] ;
                ibj = ib [j] ;
                if ( ibj < 0 )
                {
                    if ( pA [j] > ZERO )
                    {
                        s = -c [j]/pA [j] ;
                        if ( s < st )
                        {
                            Br_value [j] = s ;
                            nbrk++ ;
                            Heap [nbrk] = j ;        /* add to heap  */
                        }
                    }
                }
                else if ( ibj > 0 )
                {
                    if ( pA [j] < ZERO )
                    {
                        s = -c [j]/pA [j] ;
                        if ( s < st )
                        {
                            Br_value [j] = s ;
                            nbrk++ ;
                            Heap [nbrk] = j ;        /* add to heap */
                        }
                    }
                }
            }

            if ( W->shiftl_is_zero )
            {
                for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dl [row]) < ZERO )
                    {
                        /* lambda > 0 for lLinkUp */
                        s = -lambda [row]/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
                for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dl [row]) > ZERO )
                    {
                        /* lambda < 0 for uLinkUp */
                        s = -lambda [row]/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
            }
            else /* shift_l != 0 */
            {
                for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dl [row]) < ZERO )
                    {
                        /* lambda > 0 for lLinkUp */
                        s = -(lambda [row]+shift_l [row])/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
                for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
                {
                    row = ineq_row [j] ;
                    if ( (t = dl [row]) > ZERO )
                    {
                        /* lambda < 0 for uLinkUp */
                        s = -(lambda [row]+shift_l [row])/t ;
                        if ( s < st )
                        {
                            k = j + ncol ;
                            Br_value [k] = s ;
                            nbrk++ ;
                            Heap [nbrk] = k ; /* add to heap  */
                        }
                    }
                }
            }

            /* if no constraints activated, then return to the CG iteration */
            if ( nbrk == 0 )
            {
                break ;
            }

            /* ============================================================== */
            /* sort break points, do a line search */
            /* ============================================================== */

            pproj_minheap_build (Heap, Br_value, nbrk) ;
            for (k = 1; k <= nbrk; k++)
            {
                ns [Heap [k]] = k ;
            }

#ifndef NDEBUG
            check_minheap (Heap, Br_value, ns, nbrk, ntot,
                "build Heap, ssor1") ;
#endif

            lineflag = 0 ;
            nf0 = nf ;
            rowp = ntot ;
            st = ZERO ;
            nup = 0 ;
            while ( nbrk > 0 )
            {
                if ( sd > ZERO )
                {
                    col = Heap [1] ;
                    pproj_minheap_delete (Heap, ns, Br_value, &nbrk, 1) ;
                    Br = MAX (Br_value [col], ZERO) ;
                    fn = fd - sd * (Br - st) ;

#ifndef NDEBUG
                    check_minheap (Heap, Br_value, ns, nbrk, ntot,
                            "Heap delete, ssor1") ;
                    if ( PrintLevel > 2 )
                    {
                        PRINTF("    brk: %i col: %i fn: %9.3e fd: %9.3e"
                               " sd: %9.3e st: %9.3e Br_value: %9.3e\n",
                            nbrk, col, fn, fd, sd, st, Br) ;
                    }
#endif

                    if ( fn <= ZERO )
                    {
                        if ( fd != fn )
                        {
                            st += (fd/(fd-fn))*(Br - st);
                        }
                        fd = ZERO ;
                        break ;
                    }
                    else
                    {
                        fd = fn ;
                    }
                    nup++ ;
                }
                else
                {
                    break ;
                }

                st = Br ;

                if ( col < ncol ) /* free column */
                {
                    F [nf++] = col ;
                    s = pA [col] ;
                    sd += s*s ;
                    ib [col] = 0 ;
                    k = ColmodFlag [col] ;
                    if ( k != EMPTY )
                    {
                        l = ColmodList [ncol-ncoldel] ;
                        ColmodList [k] = l ;
                        ColmodFlag [l] = k ;
                        ColmodFlag [col] = EMPTY ;
                        ncoldel-- ;
                    }
                    else
                    {
                        ColmodList [ncoladd] = col ;
                        ColmodFlag [col] = ncoladd ;
                        ncoladd++ ;
                    }
                    q = Ap [col+1] ;
                    for (p = Ap [col]; p < q; p++)
                    {
                        i = Ai [p] ;
                        if ( ir [i] <= ni )
                        {
                            ax = Ax [p] ;
                            D [i] += ax*ax ;
                            l = AFTp [i] + AFTnz [i]++ ;
                            AFTx [l] = ax ;
                            AFTi [l] = col ;
                        }
                    }
                }
                else              /* drop row */
                {
                    Stat->ssor1_drop++ ;
                    ineqindex = col - ncol ;
                    Heap [--rowp] = ineqindex ;
                    row = ineq_row [ineqindex] ;

#ifndef NDEBUG
                    if ( PrintLevel > 2 )
                    {
                        PRINTF("    jsing: %i row: %i ir: %i\n",
                                col - ncol, row, ir [row]) ;
                    }
                    if ( ir [row] > ni )
                    {
                        PRINTF ("row: % ir: %i was already deleted "
                                "in ssor1\n", row, ir [row]) ;
                        pproj_error (-1, __FILE__, __LINE__, "stop") ;
                    }
#endif
                    ASSERT ((ir [row] != 0) && (ir [row] <= ni)) ;
                    /* set to ni+ineqindex after updates, keep track of
                       nonzeros in row until after the update associated
                       with rank 1 changes to the matrix */
                    /* ir [row] = ni + ineqindex ;*/
                    i = RLinkDn [row] ;
                    k = RLinkUp [row] ;
                    RLinkUp [i] = k ;
                    RLinkDn [k] = i ;

                    k = RowmodFlag [ineqindex] ;
                    if ( k != EMPTY ) /* row set to be added, remove it */
                    {
                        AFTnz [row] = 0 ;
                        /* move last added row to replace new dropped row */
                        l = RowmodList [ni-nrowadd] ;
                        RowmodList [k] = l ;
                        RowmodFlag [l] = k ;
                        RowmodFlag [ineqindex] = EMPTY ;
                        nrowadd-- ; /* one less row to add */
                    }
                    else /* add row to the delete list */
                    {
                        RowmodList [nrowdel] = ineqindex ;
                        RowmodFlag [ineqindex] = nrowdel ;
                        nrowdel++ ;
                    }

                    if ( W->shiftl_is_zero )
                    {
                        dknew = -lambda [row] ;
                        lambda [row] = ZERO ;
                    }
                    else
                    {
                        dknew = -(lambda [row]+shift_l [row]) ;
                        lambda [row] = -shift_l [row] ;
                    }
                    b [row] -= sigma*dknew ;
                    dk_row = dl [row] ;
                    /* fd += dk_row*(Dsigma*dknew-b [row]) ;*/
                    fd -= dk_row*b [row] ;
                    sd -= sigma*dk_row*dk_row ;
                    if ( dl [row] > ZERO ) /* drop upper bound bu */
                    {
                        b [row] -= bu [ineqindex] ;
                        m = uLinkUp [ineqindex] ;
                        l = uLinkDn [ineqindex] ;
                        uLinkUp [l] = m ;
                        uLinkDn [m] = l ;
                        if ( Parm->cholmod == TRUE )
                        {
                            blk = sol_to_blk [ineqindex] ;
                            if ( ineqindex == ustart [blk] )
                            {
                                ustart [blk] = m ;
                            }
                        }
                    }
                    else /* drop lower bound bl */
                    {
                        b [row] -= bl [ineqindex] ;
                        m = lLinkUp [ineqindex] ;
                        l = lLinkDn [ineqindex] ;
                        lLinkUp [l] = m ;
                        lLinkDn [m] = l ;
                        if ( Parm->cholmod == TRUE )
                        {
                            blk = sol_to_blk [ineqindex] ;
                            if ( ineqindex == lstart [blk] )
                            {
                                lstart [blk] = m ;
                            }
                        }
                    }

                    q = ATp [row+1] ;
                    for (p = ATp [row] ; p < q ; p++)
                    {
                        j = ATi [p] ;
                        ax = ATx [p] ;
                        s = pA [j] ;
                        /* cj = c after taking the step */
                        cj = c [j] + st*s ;
                        /* modify c to account for later update */
                        c [j] += dknew*ax ;
                        t = dk_row*ax ;
                        snew = pA [j] = s - t ;
                        /* update c, pA, sd, fd */
                        if ( !ib [j] )
                        {
                            fd += cj*t ;
                            sd -= t*(s + snew) ;
                        }
                        else
                        {
                            if ( ib [j] < 0 )
                            {
                                if ( snew > ZERO )
                                {
                                    Br_value [j] = st - cj/snew ;
                                    if ( ns [j] != EMPTY )
                                    {
                                        pproj_minheap_update (Heap,
                                           ns, Br_value, nbrk, ns [j]) ;
                                    }
                                    else
                                    {
                                       pproj_minheap_add (j, Heap,
                                           ns, Br_value, &nbrk) ;
                                    }
                                }
                                else
                                {
                                    if ( ns [j] != EMPTY )
                                    {
                                        pproj_minheap_delete (Heap,
                                           ns, Br_value, &nbrk, ns [j]);
                                    }
                                }
                            }
                            else
                            {
                                if ( snew < ZERO )
                                {
                                    Br_value [j] = st - cj/snew ;
                                    if ( ns [j] != EMPTY )
                                    {
                                        pproj_minheap_update (Heap,
                                           ns, Br_value, nbrk, ns [j]) ;
                                    }
                                    else
                                    {
                                        pproj_minheap_add (j, Heap,
                                            ns, Br_value, &nbrk) ;
                                    }
                                }
                                else
                                {
                                    if ( ns [j] != EMPTY )
                                    {
                                        pproj_minheap_delete (Heap,
                                           ns, Br_value, &nbrk, ns [j]);
                                    }
                                }
                            }
                        }
                    }
                }

#ifndef NDEBUG
                if ( PrintLevel > 2 )
                {
                    if ( col < ncol ) PRINTF ("    free: %i\n", col);
                    else              PRINTF ("    drop row: %i\n", row) ;
                }
                check_minheap (Heap, Br_value, ns, nbrk, ntot,
                    "line search of ssor1") ;
#endif

                if ( fd <= ZERO )
                {

#ifndef NDEBUG
                    if ( PrintLevel > 2 )
                    {
                        PRINTF("    premature break from line search\n") ;
                    }
#endif

                    lineflag = 1 ;
                    fd = ZERO ;
                    break ;
                }
            }
            if ( sd > ZERO )
            {
                st += fd/sd ;
            }
            if ( st > st0 )
            {
                st = st0 ;
                lineflag = -1 ;
            }

            for (k = 1; k <= nbrk; k++)
            {
                ns [Heap [k]] = EMPTY ;
            }
            if ( st > ZERO )
            {
                for (row = RLinkUp [nrow]; row < nrow; row = RLinkUp [row])
                {
                    t = st*dl [row] ;
                    lambda [row] += t ;
                    b [row] -= sigma*t ;
                }
                pproj_saxpy (c, pA, st, ncol) ;
            }
#ifndef NDEBUG
            where = "end of ssor1 line search" ;
            check_dual (I, NULL, where, TRUE, TRUE) ;
            check_line (I, lineflag, W->blks - 1, nup, dl, st) ;
            Heap [rowp-1] = EMPTY ;
            checkb (I, where) ;
            check_const (NULL, ZERO, ns, EMPTY, ntot, where) ;
            checkD (I, where) ;
            checkc (I, where) ;
#endif
            chg += nup ;
            /* subspace minimization */
            if ( it - it0 >= ssormem )
            {
                memp = ssormem - 1 ; /* we can use all the memory */
            }
            if ( nlinesearch == 1 ) /*initialization for subspace minimization*/
            {
                n = memp + 1 ; /* diag [0, ... , memp] */
                n1 = n - 1 ;
                n2 = n - 2 ;
                pproj_initx (L, ZERO, (n*(n-1))/2) ; /* initialize L = 0*/
            }

            /* When nlinesearch > 1, we must do update/downdate, however,
               after the first line search, it may be more efficient to
               simply form the ssormem by ssormem matrix by scratch and
               LDL' factor it */
            if ( (nlinesearch > 1) ||
                 (nf-nf0 + 2*(ntot-rowp) < 0.16*n) ) /* update */
            {

#ifndef NDEBUG
                if ( PrintLevel > 2 )
                {
                    printf("update the matrix, nf0: %i nf: %i n: %i\n",
                            nf0, nf, n) ;
                }
#endif

                for (k = nf0; k < nf; k++)
                {
                    j = F [k] ;
                    dk = dk_start ;
                    q = Ap [j+1] ;
                    for (l = 0; l < n; l++)
                    {
                        s = ZERO ;
                        for (p = Ap [j] ; p < q ; p++)
                        {
                            i = Ai [p] ;
                            if ( ir [i] <= ni )
                            {
                                s += dk [i]*Ax [p] ;
                            }
                        }
                        v [l] = s ;
                        dk = dk+nrow ;
                    }

                    /* update the factorization */

                    pproj_updown_dense (/* I, */ L, diag, v, +1, n) ;
                }

                for (m = rowp; m < ntot; m++)
                {
                    ineqindex = Heap [m] ;
                    row = ineq_row [ineqindex] ;
                    /* form AF times row of AF  */
                    p = AFTp [row] ;
                    q = p + AFTnz [row] ;
                    ASSERT (ir [row] <= ni) ;
                    nr = 0 ;
                    for ( ; p < q ; p++)
                    {
                        j = AFTi [p] ;
                        q1 = Ap [j+1] ;
                        t = AFTx [p] ;
                        for (l = Ap [j]; l < q1; l++)
                        {
                            i = Ai [l] ;
                            if ( ir [i] <= ni )
                            {
                                if ( ns [i] == EMPTY )
                                {
                                    ns [i] = 1 ;
                                    rowlist [nr++] = i ;
                                    pk [i] = ZERO ;
                                }
                                pk [i] += Ax [l]*t ;
                            }
                        }
                    }

                    s = .5*(pk [row] - sigma) ;
                    pproj_initx (e, ZERO, n) ;

                    for (l = 0; l < nr; l++)
                    {
                        i = rowlist [l] ;
                        ns [i] = EMPTY ;
                        t = pk [i] ;
                        dk = dk_start+i ;
                        if ( ir [i] <= ni )
                        {
                            for (k = 0; k < n; k++)
                            {
                                e [k] -= t*(*dk) ;
                                dk = dk+nrow ;
                            }
                        }
                    }

                    dk = dk_start+row ;
                    u = ZERO ;
                    for (k = 0; k < n; k++)
                    {
                        t = v [k] = *dk ;
                        u += t*t ;
                        e [k] += s*t ;
                        dk = dk+nrow ;
                    }
                    ir [row] = ni + ineqindex ; /* drop row */
                    AFTnz [row] = 0 ;

                    t = ZERO ;
                    for (k = 0; k < n; k++)
                    {
                        t += e [k]*e [k] ;
                    }

                    if ( (t > ZERO) && (u > ZERO) )
                    {
                        t = sqrt (t) ;
                        u = sqrt (u) ;
                        s = sqrt (.5*t*u) ;
                        t = s/t ;
                        u = s/u ;
                        for (k = 0; k < n; k++)
                        {
                            y = u*v [k] ;
                            z = t*e [k] ;
                            v [k] = y + z ;
                            e [k] = y - z ;
                        }

                        /* update the factorization */
                        pproj_updown_dense (L, diag, v, +1, n) ;

                        /* downdate the factorization */
                        pproj_updown_dense (L, diag, e, -1, n) ;
                    }
                }
            }
            else /* construct the matrix from scratch and factor it */
            {
#ifndef NDEBUG
                if ( PrintLevel > 2 )
                {
                    PRINTF("build and factor matrix, nf0: %i nf: %i n: %i\n",
                            nf0, nf, n) ;
                }
#endif
                /* ------------------------------------------------------ */
                /* add the columns */
                /* ------------------------------------------------------ */

                for (k = nf0; k < nf; k++)
                {
                    j = F [k] ;
                    dk = dk_start ;
                    q = Ap [j+1] ;
                    for (l = 0; l < n; l++)
                    {
                        s = 0. ;
                        for (p = Ap [j]; p < q; p++)
                        {
                            i = Ai [p] ;
                            if ( ir [i] <= ni )
                            {
                                s += dk [i]*Ax [p] ;
                            }
                        }
                        v [l] = s ;
                        dk = dk+nrow ;
                    }

                    Lj = L-1 ;
                    for (j = 0; j < n; j++)
                    {
                        t = v [j] ;
                        diag [j] += t*t ;
                        for (i = j+1; i < n; i++)
                        {
                            Lj [i] += t*v [i] ;
                        }
                        Lj = Lj+(n2-j) ;
                    }
                }

                /* ------------------------------------------------------ */
                /* remove the rows */
                /* ------------------------------------------------------ */

                for (m = rowp; m < ntot; m++)
                {
                    ineqindex = Heap [m] ;
                    row = ineq_row [ineqindex] ;
                    /* form AF times row of AF  */
                    p = AFTp [row] ;
                    q = p + AFTnz [row] ;
                    ASSERT (ir [row] <= ni) ;
                    nr = 0 ;
                    for (; p < q; p++)
                    {
                        j = AFTi [p] ;
                        q1 = Ap [j+1] ;
                        t = AFTx [p] ;
                        for (l = Ap [j]; l < q1; l++)
                        {
                            i = Ai [l] ;
                            if ( ir [i] <= ni )
                            {
                                if ( ns [i] == EMPTY )
                                {
                                    ns [i] = 1 ;
                                    rowlist [nr++] = i ;
                                    pk [i] = ZERO ;
                                }
                                pk [i] += Ax [l]*t ;
                            }
                        }
                    }

                    s = .5*(pk [row] - sigma) ;
                    pproj_initx (e, ZERO, n) ;

                    for (l = 0; l < nr; l++)
                    {
                        i = rowlist [l] ;
                        ns [i] = EMPTY ;
                        t = pk [i] ;
                        dk = dk_start+i ;
                        if ( ir [i] <= ni )
                        {
                            for (k = 0; k < n; k++)
                            {
                                e [k] -= t*(*dk) ;
                                dk = dk+nrow ;
                            }
                        }
                    }

                    dk = dk_start+row ;
                    for (k = 0; k < n; k++)
                    {
                        t = v [k] = *dk ;
                        e [k] += s*t ;
                        dk = dk+nrow ;
                    }
                    ir [row] = ni + ineqindex ; /* drop the row */
                    AFTnz [row] = 0 ;
 
                    Lj = L-1 ;
                    for (j = 0; j < n; j++)
                    {
                        t = v [j] ;
                        s = e [j] ;
                        diag [j] += 2.*s*t ;
                        for (i = j+1; i < n; i++)
                        {
                            Lj [i] += t*e [i] + s*v [i] ;
                        }
                        Lj = Lj+(n2-j) ;
                    }
                }
 
                /* ------------------------------------------------------ */
                /* factor the matrix by columns */
                /* ------------------------------------------------------ */

                Lj = L-1 ;
                for (j = 0; j < n; j++)
                {
                    /* apply previous columns to column j */
                    Lk = L-1 ;
                    s = 0. ;
                    for (k = 0; k < j; k++)
                    {
                        t = Lk [j] ;
                        z = diag [k] * t ;
                        s += t*z ;
                        for (i = j+1; i < n; i++)
                        {
                            Lj [i] -= Lk [i]*z ;
                        }
                        Lk = Lk+(n2-k) ;
                    }
                    s = diag [j] = diag [j] - s ;
                    for (k = j+1; k < n; k++)
                    {
                        Lj [k] /= s ;
                    }
                    Lj = Lj+(n2-j) ;
                }
            }

#ifndef NDEBUG
            where = "ssor1 after update" ;
            W->nrowdel = nrowdel ;
            W->nrowadd = nrowadd ;
            W->ncoladd = ncoladd ;
            W->ncoldel = ncoldel ;
            check_link (I, (int *) NULL, 0, where) ;
            check_modlist (I, where) ;
#endif
            /* ---------------------------------------------------------- */
            /* compute residual at current point */
            /* ---------------------------------------------------------- */

            first = RLinkUp [nrow] ;
            for (i = first; i < nrow; i = RLinkUp [i])
            {
                t = b [i] ;
                p = AFTp [i] ;
                q = p + AFTnz [i] ;
                ASSERT (ir [i] <= ni) ;
                for (; p < q; p++)
                {
                    t -= AFTx [p]*c [AFTi [p]] ;
                }
                gk [i] = t ;
            }
            dk = dk_start ;
            for (l = 0; l < n; l++)
            {
                t = ZERO ;
                for (i = first; i < nrow; i = RLinkUp [i])
                {
                    t += gk [i]*dk [i] ;
                }
                w [l] = t ;
                dk = dk+nrow ;
            }

            /* ---------------------------------------------------------- */
            /* forward solve the factored system */
            /* ---------------------------------------------------------- */

            Lj = L-1 ;
            for (j = 0; j < n1; j++)
            {
                t = w [j] ;
                for (i = j+1; i < n; i++)
                {
                    w [i] -= t * Lj [i] ;
                }
                Lj = Lj+(n2-j) ;
            }

            /* ---------------------------------------------------------- */
            /* back solve the factored system */
            /* ---------------------------------------------------------- */

            for (i = n1; i >= 0; i--)
            {
                t = w [i] / diag [i] ;
                for (j = i+1; j < n; j++)
                {
                    t -= w [j]*Lj [j] ;
                }
                w [i] = t ;
                Lj = Lj-(n1-i) ;
            }

            for (l = 0; l < n; l++)
            {
                t = w [l] ;
                if ( l > 0 )
                {
                    for (i = first; i < nrow; i = RLinkUp [i])
                    {
                        dl [i] += dk [i]*t ;
                    }
                    dk = dk+nrow ;
                }
                else
                {
                    dk = dk_start ;
                    for (i = first; i < nrow; i = RLinkUp [i])
                    {
                        dl [i] = dk [i]*t ;
                    }
                    dk = dk+nrow ;
                }
            }
        }   /* return to perform a new line search in the direction dl */

        /* flag = -1 means reinitialize the CG iteration
           flag =  0 means no change in active constraints
           flag =  1 means all rows became inactive
           flag =  2 means the CG iteration generated a nonascent point */
        if ( flag == 1 )
        {
            status = PPROJ_ALL_ROWS_DROPPED ;
            break ; /* all rows became inactive */
        }

        /* check to see if the iteration limit was reached */
        if ( it >= Parm->ssormaxits )
        {
            status = PPROJ_SSOR_MAX_ITS ;
            break ;
        }
        /* If CG generated a nonascent direction, and nlinesearch = 1,
           then terminate since CG should always generate a descent direction.
           If nlinesearch > 1, this poor direction could be due to the
           fact that the direction vectors are not perfectly conjugate.
           In this case, do not take a step in the search direction and
           restart CG. */
        if ( (flag == 2) && (nlinesearch == 1) )
        {
            status = PPROJ_SSOR_NONASCENT ;
            break ;
        }
        else if ( flag == 2 ) /* nlinesearch > 1 */
        {
            flag = -1 ; /* restart CG */
            continue ;
        }

        /* update variables */
        for (row = first; row < nrow; row = RLinkUp [row])
        {
            t = dl [row] ;
            lambda [row] += t ;
            b [row] -= sigma*t ;
        }
        pproj_saxpy (c, pA, ONE, ncol) ;

        if ( nlinesearch > 1 ) /* restart CG */
        {
            flag = -1 ;
            continue ;
        }

        /* if no variables are freeing or rows are dropping, then switch
           to update/downdate */
        if ( (nlinesearch == 1) && (it - it0 >= W->ssor1_its) )
        {
            status = PPROJ_SWITCH_TO_UPDOWN ;
            break ;
        }

        /* check error and continue CG */
        flag = 0 ;
        sub_inc = MIN (sub_inc+2, 10) ;
        for (k = 0; k < nf; k++)
        {
            j = F [k] ;
            t = c [j] ;
            if      ( t < lo [j] ) t = lo [j] ;
            else if ( t > hi [j] ) t = hi [j] ;
            pA [j] = t ; /* projection of c on the box */
        }
        errls = ZERO ;
        errdual = ZERO ;
        for (i = first; i < nrow; i = RLinkUp [i])
        {
            t = b [i] ;                /* includes -sigma*lambda [i] */
            s = t + sigma*lambda [i] ; /* remove -sigma*lambda [i] term */
            dl [i] = ZERO ;
            ASSERT (ir [i] <= ni) ;
            p = AFTp [i] ;
            q = p + AFTnz [i] ;
            for (; p < q; p++)
            {
                j = AFTi [p] ;
                ax = AFTx [p] ;
                t -= ax*c [j] ;
                s -= ax*pA [j] ; /* pA stores projection of c on the box */
            }
            gk [i] = t ;
            if ( fabs (s) > errdual ) errdual = fabs (s) ;
            if ( fabs (t) > errls   ) errls   = fabs (t) ;
        }
        if ( !Parm->stopabs )
        {
            errdual /= W->absAx ;
            errls   /= W->absAx ;
        }
        if ( PrintLevel > 1 )
        {
            printf ("it: %i errdual: %e errls: %e\n", it, errdual, errls) ;
        }

        if ( errdual <= 0.8*Prob->grad_tol )
        {
            status = PPROJ_SOLUTION_FOUND ;
            break ;
        }
        if ( errls <= Parm->ssordecay*errdual )
        {
            status = PPROJ_STATUS_OK ;
            break ;
        }
    }
    W->ncoladd = ncoladd ;
    W->ncoldel = ncoldel ;
    W->nrowdel = nrowdel ;
    W->nrowadd = nrowadd ;
    Stat->ssor1_free += nf - W->nf ;
    W->chg_ssor1 = chg ;
    W->nf = nf ;
    Stat->ssor1 += pproj_timer () - tic ;
    Stat->ssor1_its += it ;
    if ( PrintLevel > 1 )
    {
        PRINTF ("SSOR1 complete, it: %i\n", it) ;
    }
    return (status) ;
}

/* ========================================================================= */
/* === pproj_updown_dense ================================================== */
/* ========================================================================= */

void pproj_updown_dense
(
    double    *L, /* lower triangle of factorization (beneath diagonal) */
    double *diag, /* D in the LDL' factorization */
    double    *w, /* update/downdate matrix */
    int     info, /* +1 = update, -1 = downdate */
    int        n  /* rank of the matrix */
)
{
    int j, n2, p ;
    double *Lj, a, abar, dj, gamma, t, wj ;

/* update the factorization stored in L */

    a = 1. ;
    n2 = n - 2 ;
    Lj = L - 1 ;
    if ( info > 0 ) /* update */
    {
        for (j = 0; j < n; j++)
        {
            dj = diag [j] ;
            wj = w [j] ;
            abar = a +  wj*wj / dj ;
            dj *= abar ;
            gamma = wj / dj ;
            dj /= a ;
            diag [j] = dj ;
            a = abar ;
            for (p = j+1; p < n; p++)
            {
                t = w [p] = w [p] - wj*Lj [p] ;
                Lj [p] += gamma * t ;
            }
            Lj = Lj+(n2-j) ;
        }
    }
    else            /* downdate */
    {
        for (j = 0; j < n; j++)
        {
            dj = diag [j] ;
            wj = w [j] ;
            abar = a -  wj*wj / dj ;
            dj *= abar ;
            gamma = wj / dj ;
            dj /= a ;
            diag [j] = dj ;
            a = abar ;
            for (p = j+1; p < n; p++)
            {
                t = w [p] = w [p] - wj*Lj [p] ;
                Lj [p] -= gamma * t ;
            }
            Lj = Lj+(n2-j) ;
        }
    }
}
